package ch.psi.daq.queryrest.config;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.EnumMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;

import org.apache.commons.lang3.tuple.Pair;
import org.msgpack.jackson.dataformat.MessagePackFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Import;
import org.springframework.context.annotation.Lazy;
import org.springframework.context.annotation.PropertySource;
import org.springframework.context.annotation.Scope;
import org.springframework.core.env.Environment;
import org.springframework.util.StringUtils;
import org.springframework.validation.Validator;

import com.fasterxml.jackson.annotation.JsonInclude.Include;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.smile.SmileFactory;

import ch.psi.daq.common.statistic.Statistics;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.backend.BackendType;
import ch.psi.daq.domain.backend.DomainBackendType;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.config.DomainConfigCORS;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.json.ServerAddress;
import ch.psi.daq.domain.json.mixin.HistoricChannelConfigurationPropertyFilterMixin;
import ch.psi.daq.domain.json.mixin.PropertyFilterMixin;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.backend.analyzer.BackendQueryAnalyzer;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.ConfigField;
import ch.psi.daq.domain.query.operation.EventField;
import ch.psi.daq.domain.query.operation.aggregation.extrema.AbstractExtremaMeta;
import ch.psi.daq.domain.query.response.Response;
import ch.psi.daq.domain.request.validate.RequestProviderValidator;
import ch.psi.daq.domain.rest.RestHelper;
import ch.psi.daq.query.analyzer.BackendQueryAnalyzerImpl;
import ch.psi.daq.query.config.QueryConfig;
import ch.psi.daq.queryrest.controller.validator.ConfigQueryValidator;
import ch.psi.daq.queryrest.controller.validator.EventQueryValidator;
import ch.psi.daq.queryrest.query.QueryManager;
import ch.psi.daq.queryrest.query.QueryManagerImpl;
import ch.psi.daq.queryrest.query.QueryManagerRemote;
import ch.psi.daq.queryrest.response.PolymorphicResponseMixIn;
import ch.psi.daq.queryrest.response.csv.CSVResponseStreamWriter;
import ch.psi.daq.queryrest.response.formatter.AnyResponseFormatter;
import ch.psi.daq.queryrest.response.formatter.DAQConfigQueryResponseFormatter;
import ch.psi.daq.queryrest.response.formatter.DAQQueriesResponseFormatter;
import ch.psi.daq.queryrest.response.json.JSONResponseStreamWriter;
import ch.psi.daq.queryrest.response.msgpack.MsgPackResponseStreamWriter;
import ch.psi.daq.queryrest.response.raw.event.RawEventResponseFormatter;
import ch.psi.daq.queryrest.response.raw.event.RawEventResponseStreamWriter;
import ch.psi.daq.queryrest.response.smile.SmileResponseStreamWriter;

@Configuration
@Import(value = DomainConfigCORS.class)
@PropertySource(value = {"classpath:queryrest.properties"})
@PropertySource(value = {
      "classpath:queryrest-${daq.config.environment}.properties",
      "file:${user.home}/.config/daq/queryrest.properties"}, ignoreResourceNotFound = true)
public class QueryRestConfig { // extends WebMvcConfigurerAdapter {

   private static final String QUERYREST_RESPONSE_FIELDS_EVENT_QUERY = "queryrest.response.fields.event.query";
   private static final String QUERYREST_RESPONSE_FIELDS_EVENT_QUERY_AGGREGATIONS =
         "queryrest.response.fields.event.query.aggregations";

   private static final String QUERYREST_RESPONSE_FIELDS_CONFIG_QUERY = "queryrest.response.fields.config.query";
   private static final String QUERYREST_RESPONSE_FIELDS_CONFIG_HISTORIC = "queryrest.response.fields.config.historic";

   // a nested configuration
   // this guarantees that the ordering of the properties file is as expected
   // see:
   // https://jira.spring.io/browse/SPR-10409?focusedCommentId=101393&page=com.atlassian.jira.plugin.system.issuetabpanels:comment-tabpanel#comment-101393
   @Configuration
   @Import({QueryConfig.class})
   static class InnerConfiguration {
   }

   // somehow needed to make sure @Import elements will get initialized and afterPropertiesSet will
   // get called (ensuring BackendAcces gets initialized according hierarchy)
   @Resource
   private QueryConfig queryConfig;

   private static final Logger LOGGER = LoggerFactory.getLogger(QueryRestConfig.class);

   public static final String BEAN_NAME_QUERY_MANAGER = "queryManager";
   public static final String BEAN_NAME_QUERY_MANAGER_LOCAL = "queryManagerLocal";
   public static final String BEAN_NAME_QUERY_MANAGER_REMOTE = "queryManagerRemote";
   public static final String BEAN_NAME_QUERY_ANALIZER_FACTORY = "queryAnalizerFactory";
   public static final String BEAN_NAME_EVENT_QUERY_VALIDATOR = "eventQueryValidator";
   public static final String BEAN_NAME_CONFIG_QUERY_VALIDATOR = "configQueryValidator";
   public static final String BEAN_NAME_REQUEST_PROVIDER_VALIDATOR = "requestProviderValidator";
   public static final String BEAN_NAME_JSON_FACTORY = "jsonFactory";
   public static final String BEAN_NAME_MSG_PACK_FACTORY = "msgPackFactory";
   public static final String BEAN_NAME_SMILE_FACTORY = "smileFactory";
   public static final String BEAN_NAME_DEFAULT_EVENT_RESPONSE_FIELDS = "defaultEventResponseFields";
   public static final String BEAN_NAME_DEFAULT_EVENT_RESPONSE_AGGREGATIONS = "defaultEventResponseAggregations";
   public static final String BEAN_NAME_CONFIG_RESPONSE_FIELDS_QUERY = "configResponseFieldsQuery";
   public static final String BEAN_NAME_CONFIG_RESPONSE_FIELDS_HISTORIC = "configResponseFieldsHistoric";
   public static final String BEAN_NAME_FORMATTER_DAQ_QUERIES = "formatterDAQQueries";
   public static final String BEAN_NAME_FORMATTER_DAQ_CONFIG_QUERY = "formatterDAQConfigQuery";
   public static final String BEAN_NAME_FORMATTER_ANY = "formatterAny";
   public static final String BEAN_NAME_FORMATTER_HISTORIC_CHANNEL_CONFIGURATION =
         "formatterHistoricChannelConfiguration";
   public static final String BEAN_NAME_FORMATTER_RAW_EVENT = "formatterRawEvent";
   public static final String BEAN_NAME_QUERY_SERVER_ADDRESSES = "queryServerAddesses";
   public static final String BEAN_NAME_QUERY_SERVER_BACKENDS = "queryServerBackends";

   private static final String QUERY_SERVER_TYPE = "query.server.type";
   private static final String QUERY_SERVER_ADDRESSES = "query.server.addresses";

   @Resource
   private ApplicationContext context;
   private Environment env;
   private ObjectMapper objectMapper;

   @SuppressWarnings("unchecked")
   @PostConstruct
   public void afterPropertiesSet() {
      env = context.getEnvironment();
      objectMapper = context.getBean(DomainConfig.BEAN_NAME_OBJECT_MAPPER, ObjectMapper.class);

      initObjectMapper(objectMapper);

      // init BackendAccesses
      final List<Backend> backends = context.getBean(DomainConfig.BEAN_NAME_BACKENDS, List.class);
      for (final Backend backend : backends) {
         // backends instance is different to objectMapper above
         initObjectMapper(
               backend.getApplicationContext().getBean(DomainConfig.BEAN_NAME_OBJECT_MAPPER, ObjectMapper.class));
      }
   }

   private void initObjectMapper(ObjectMapper objectMapper) {
      // only include non-null values
      objectMapper.setSerializationInclusion(Include.NON_NULL);
      // Mixin which is used dynamically to filter out which properties get serialised and which
      // won't. This way, the user can specify which columns are to be received.
      objectMapper.addMixIn(DataEvent.class, PropertyFilterMixin.class);
      objectMapper.addMixIn(Statistics.class, PropertyFilterMixin.class);
      objectMapper.addMixIn(AbstractExtremaMeta.class, PropertyFilterMixin.class);
      objectMapper.addMixIn(EnumMap.class, PropertyFilterMixin.class);

      objectMapper.addMixIn(ChannelConfiguration.class, HistoricChannelConfigurationPropertyFilterMixin.class);

      objectMapper.addMixIn(Response.class, PolymorphicResponseMixIn.class);
   }

   // @Override
   // public void configureMessageConverters(List<HttpMessageConverter<?>> converters) {
   // final MappingJackson2HttpMessageConverter converter = new
   // MappingJackson2HttpMessageConverter();
   // /**
   // * This is necessary so that the message conversion uses the configured object mapper.
   // * Otherwise, a separate object mapper is instantiated for Springs message conversion.
   // */
   // converter.setObjectMapper(objectMapper);
   // converters.add(converter);
   // super.configureMessageConverters(converters);
   // }

   // does this work for json comming from web-requests
   // -> use WebMvcConfigurationSupport?
   // https://stackoverflow.com/questions/26639475/how-to-set-context-param-in-spring-boot
   // @Bean
   // @Lazy
   // public MappingJackson2HttpMessageConverter mappingJackson2HttpMessageConverter() {
   // final MappingJackson2HttpMessageConverter jsonConverter = new
   // MappingJackson2HttpMessageConverter();
   // jsonConverter.setObjectMapper(objectMapper);
   // return jsonConverter;
   // }

   @Bean(name = BEAN_NAME_JSON_FACTORY)
   @Lazy
   public JsonFactory jsonFactory() {
      return new JsonFactory();
   }

   @Bean(name = BEAN_NAME_MSG_PACK_FACTORY)
   @Lazy
   public MessagePackFactory messagePackFactory() {
      return new MessagePackFactory();
   }

   @Bean(name = BEAN_NAME_SMILE_FACTORY)
   @Lazy
   public SmileFactory smileFactory() {
      return new SmileFactory();
   }

   @Bean(name = BEAN_NAME_QUERY_ANALIZER_FACTORY)
   @Lazy
   public Function<BackendQuery, BackendQueryAnalyzer> queryAnalizerFactory() {
      return (query) -> new BackendQueryAnalyzerImpl(query);
   }

   @Bean
   @Lazy
   public JSONResponseStreamWriter jsonResponseStreamWriter() {
      return new JSONResponseStreamWriter();
   }

   @Bean
   @Lazy
   public MsgPackResponseStreamWriter msgPackResponseStreamWriter() {
      return new MsgPackResponseStreamWriter();
   }

   @Bean
   @Lazy
   public SmileResponseStreamWriter smileResponseStreamWriter() {
      return new SmileResponseStreamWriter();
   }

   @Bean
   @Lazy
   public CSVResponseStreamWriter csvResponseStreamWriter() {
      return new CSVResponseStreamWriter();
   }

   @Bean
   @Lazy
   public RawEventResponseStreamWriter rawEventResponseStreamWriter() {
      return new RawEventResponseStreamWriter();
   }

   @Bean(name = BEAN_NAME_QUERY_MANAGER)
   @Lazy
   public QueryManager queryManager() {
      final String value = env.getProperty(QUERY_SERVER_TYPE, "local");
      LOGGER.debug("Load '{}={}'", QUERY_SERVER_TYPE, value);

      if (QueryManagerImpl.QUERY_SERVER_TYPE.equals(value)) {
         return context.getBean(BEAN_NAME_QUERY_MANAGER_LOCAL, QueryManager.class);
      } else if (QueryManagerRemote.QUERY_SERVER_TYPE.equals(value)) {
         return context.getBean(BEAN_NAME_QUERY_MANAGER_REMOTE, QueryManager.class);
      } else {
         final String message = String.format("Unknown value for '%s=%s'", QUERY_SERVER_TYPE, value);
         LOGGER.error(message);
         throw new IllegalStateException(message);
      }

   }

   @Bean(name = BEAN_NAME_QUERY_MANAGER_LOCAL)
   @Lazy
   public QueryManagerImpl queryManagerLocal() {
      return new QueryManagerImpl();
   }

   @Bean(name = BEAN_NAME_QUERY_MANAGER_REMOTE)
   @Lazy
   public QueryManagerRemote queryManagerRemote() {
      return new QueryManagerRemote();
   }

   @Bean(name = BEAN_NAME_DEFAULT_EVENT_RESPONSE_FIELDS)
   @Lazy
   public Set<EventField> defaultEventResponseFields() {
      String[] responseFields =
            StringUtils.commaDelimitedListToStringArray(env.getProperty(QUERYREST_RESPONSE_FIELDS_EVENT_QUERY));
      LOGGER.debug("Load '{}={}'", QUERYREST_RESPONSE_FIELDS_EVENT_QUERY, Arrays.toString(responseFields));

      // preserve order
      LinkedHashSet<EventField> defaultResponseFields = new LinkedHashSet<>(responseFields.length);
      for (String field : responseFields) {
         try {
            defaultResponseFields.add(EventField.valueOf(field));
         } catch (Exception e) {
            LOGGER.error("Field '{}' in '{}' is invalid.", field, QUERYREST_RESPONSE_FIELDS_EVENT_QUERY, e);
         }
      }

      return defaultResponseFields;
   }

   @Bean(name = BEAN_NAME_CONFIG_RESPONSE_FIELDS_QUERY)
   @Lazy
   public Set<ConfigField> configResponseFieldsQuery() {
      String[] responseFields =
            StringUtils.commaDelimitedListToStringArray(env.getProperty(QUERYREST_RESPONSE_FIELDS_CONFIG_QUERY));
      LOGGER.debug("Load '{}={}'", QUERYREST_RESPONSE_FIELDS_CONFIG_QUERY, Arrays.toString(responseFields));

      // preserve order
      LinkedHashSet<ConfigField> extractedResponseFields = new LinkedHashSet<>(responseFields.length);
      for (String field : responseFields) {
         try {
            extractedResponseFields.add(ConfigField.valueOf(field));
         } catch (Exception e) {
            LOGGER.error("Field '{}' in '{}' is invalid.", field, QUERYREST_RESPONSE_FIELDS_CONFIG_QUERY, e);
         }
      }

      return extractedResponseFields;
   }

   @Bean(name = BEAN_NAME_CONFIG_RESPONSE_FIELDS_HISTORIC)
   @Lazy
   public Set<ConfigField> configResponseFieldsHistoric() {
      String[] responseFields =
            StringUtils.commaDelimitedListToStringArray(env.getProperty(QUERYREST_RESPONSE_FIELDS_CONFIG_HISTORIC));
      LOGGER.debug("Load '{}={}'", QUERYREST_RESPONSE_FIELDS_CONFIG_HISTORIC, Arrays.toString(responseFields));

      // preserve order
      LinkedHashSet<ConfigField> extractedResponseFields = new LinkedHashSet<>(responseFields.length);
      for (String field : responseFields) {
         try {
            extractedResponseFields.add(ConfigField.valueOf(field));
         } catch (Exception e) {
            LOGGER.error("Field '{}' in '{}' is invalid.", field, QUERYREST_RESPONSE_FIELDS_CONFIG_HISTORIC, e);
         }
      }

      return extractedResponseFields;
   }

   @Bean(name = BEAN_NAME_DEFAULT_EVENT_RESPONSE_AGGREGATIONS)
   @Lazy
   public Set<Aggregation> defaultResponseAggregations() {
      String[] responseAggregations =
            StringUtils
                  .commaDelimitedListToStringArray(env.getProperty(QUERYREST_RESPONSE_FIELDS_EVENT_QUERY_AGGREGATIONS));
      LOGGER.debug("Load '{}={}'", QUERYREST_RESPONSE_FIELDS_EVENT_QUERY_AGGREGATIONS,
            Arrays.toString(responseAggregations));

      // preserve order
      LinkedHashSet<Aggregation> defaultResponseAggregations = new LinkedHashSet<>(responseAggregations.length);
      for (String aggregation : responseAggregations) {
         try {
            defaultResponseAggregations.add(Aggregation.valueOf(aggregation));
         } catch (Exception e) {
            LOGGER.error("Aggregation '{}' in '{}' is invalid.", aggregation,
                  QUERYREST_RESPONSE_FIELDS_EVENT_QUERY_AGGREGATIONS,
                  e);
         }
      }

      return defaultResponseAggregations;
   }

   @Bean(name = BEAN_NAME_EVENT_QUERY_VALIDATOR)
   @Lazy
   public Validator eventQueryValidator() {
      return new EventQueryValidator();
   }

   @Bean(name = BEAN_NAME_CONFIG_QUERY_VALIDATOR)
   @Lazy
   public Validator configQueryValidator() {
      return new ConfigQueryValidator();
   }

   @Bean(name = BEAN_NAME_REQUEST_PROVIDER_VALIDATOR)
   @Lazy
   public Validator requestProviderValidator() {
      return new RequestProviderValidator();
   }

   @Bean(name = BEAN_NAME_FORMATTER_DAQ_QUERIES)
   @Lazy
   public DAQQueriesResponseFormatter daqQueriesFormatter() {
      return new DAQQueriesResponseFormatter();
   }

   @Bean(name = BEAN_NAME_FORMATTER_DAQ_CONFIG_QUERY)
   @Lazy
   public DAQConfigQueryResponseFormatter daqConfigQueryFormatter() {
      return new DAQConfigQueryResponseFormatter();
   }

   @Bean(name = BEAN_NAME_FORMATTER_ANY)
   @Lazy
   public AnyResponseFormatter anyFormatter() {
      return new AnyResponseFormatter(
            BEAN_NAME_DEFAULT_EVENT_RESPONSE_FIELDS,
            BEAN_NAME_CONFIG_RESPONSE_FIELDS_HISTORIC);
   }

   @Bean(name = BEAN_NAME_FORMATTER_HISTORIC_CHANNEL_CONFIGURATION)
   @Lazy
   public AnyResponseFormatter historicChannelConfigurationFormatter() {
      return new AnyResponseFormatter(
            BEAN_NAME_DEFAULT_EVENT_RESPONSE_FIELDS,
            BEAN_NAME_CONFIG_RESPONSE_FIELDS_HISTORIC);
   }

   @Bean(name = BEAN_NAME_FORMATTER_RAW_EVENT)
   @Scope(BeanDefinition.SCOPE_PROTOTYPE)
   @Lazy
   public RawEventResponseFormatter idRawResponseFormatter() {
      return new RawEventResponseFormatter();
   }

   @Bean(name = BEAN_NAME_QUERY_SERVER_ADDRESSES)
   @Lazy
   public List<String> queryServerAddresses() {
      final String value = env.getProperty(QUERY_SERVER_ADDRESSES, "[{}]");
      LOGGER.debug("Load '{}={}'", QUERY_SERVER_ADDRESSES, value);

      List<ServerAddress> queryBackends;
      try {
         queryBackends = objectMapper.readValue(value, new TypeReference<ArrayList<ServerAddress>>() {});
      } catch (final Exception e) {
         LOGGER.warn("Could not load '{}={}'. Use default", QUERY_SERVER_ADDRESSES, value, e);
         ServerAddress defaultAddress = new ServerAddress();
         queryBackends = Arrays.asList(defaultAddress);
      }

      final String newDefaultProtocol = "https";
      final String newDefaultHost = "data-api.psi.ch";


      LOGGER.warn("\nSet back to 443!\n#####");
      final int newDefaultPort = 8080;
      // this should be overwritten
      final String newDefaultPath = "";
      final List<String> queryBackendStrs = new ArrayList<>(queryBackends.size());
      for (final ServerAddress serverAddress : queryBackends) {
         serverAddress.replaceDefaults(
               newDefaultProtocol,
               newDefaultHost,
               newDefaultPort,
               newDefaultPath);
         queryBackendStrs.add(serverAddress.getAddress());
      }

      return queryBackendStrs;
   }

   @Bean(name = BEAN_NAME_QUERY_SERVER_BACKENDS)
   @Lazy
   public Map<Backend, String> queryServerBackends() {
      final BiFunction<String, Pair<String, String>, BackendType> typeCreator =
            (name, serverAddresses) -> new DomainBackendType(
                  name,
                  serverAddresses.getLeft(),
                  serverAddresses.getRight());
      @SuppressWarnings("unchecked")
      final List<String> queryServerAddresses =
            context.getBean(QueryRestConfig.BEAN_NAME_QUERY_SERVER_ADDRESSES, List.class);
      return RestHelper.getBackends(
            context,
            queryServerAddresses.stream()
                  .map(queryServer -> Pair.of(queryServer, null)),
            serverAddresses -> serverAddresses.getLeft(),
            DomainConfig.PATH_BACKENDS_BY_ID,
            typeCreator)
            .entrySet()
            .stream()
            .collect(Collectors.toMap(
                  Entry::getKey,
                  entry -> entry.getValue().getLeft()));
   }
}
