package ch.psi.daq.queryrest.controller;

import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;

import org.apache.commons.lang3.tuple.Triple;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.validation.DirectFieldBindingResult;
import org.springframework.validation.Errors;
import org.springframework.validation.ObjectError;
import org.springframework.validation.Validator;
import org.springframework.web.bind.WebDataBinder;
import org.springframework.web.bind.annotation.InitBinder;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.RestController;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.collect.Lists;

import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.common.string.StringUtils;
import ch.psi.daq.common.tuple.Quadruple;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQConfigQuery;
import ch.psi.daq.domain.query.DAQConfigQueryElement;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.channels.ChannelConfigurationsRequest;
import ch.psi.daq.domain.query.channels.ChannelsRequest;
import ch.psi.daq.domain.query.channels.LongHash;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.AggregationType;
import ch.psi.daq.domain.query.operation.Compression;
import ch.psi.daq.domain.query.operation.ConfigField;
import ch.psi.daq.domain.query.operation.EventField;
import ch.psi.daq.domain.query.response.Response;
import ch.psi.daq.domain.query.response.ResponseFormat;
import ch.psi.daq.domain.query.transform.image.color.ColorModelType;
import ch.psi.daq.domain.query.transform.image.resize.ValueAggregation;
import ch.psi.daq.queryrest.config.QueryRestConfig;
import ch.psi.daq.queryrest.query.QueryManager;
import ch.psi.daq.queryrest.response.AbstractHTTPResponse;
import ch.psi.daq.queryrest.response.PolymorphicResponseMixIn;
import ch.psi.daq.queryrest.response.json.JSONHTTPResponse;

@RestController
public class QueryRestController implements ApplicationContextAware {

   private static final Logger LOGGER = LoggerFactory.getLogger(QueryRestController.class);

   private ApplicationContext context;
   private ObjectMapper objectMapper;
   private QueryManager queryManager;
   private Validator eventQueryValidator;
   private Validator configQueryValidator;
   private Validator requestProviderValidator;
   private Response defaultResponse = new JSONHTTPResponse();

   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      final Backend backend = context.getBean(DomainConfig.BEAN_NAME_BACKEND_DEFAULT, Backend.class);
      context = backend.getApplicationContext();
      this.context = context;

      objectMapper = context.getBean(DomainConfig.BEAN_NAME_OBJECT_MAPPER, ObjectMapper.class);
      queryManager = context.getBean(QueryRestConfig.BEAN_NAME_QUERY_MANAGER, QueryManager.class);
      eventQueryValidator = context.getBean(QueryRestConfig.BEAN_NAME_EVENT_QUERY_VALIDATOR, Validator.class);
      configQueryValidator = context.getBean(QueryRestConfig.BEAN_NAME_CONFIG_QUERY_VALIDATOR, Validator.class);
      requestProviderValidator = context.getBean(QueryRestConfig.BEAN_NAME_REQUEST_PROVIDER_VALIDATOR, Validator.class);
   }

   @InitBinder
   protected void initBinder(WebDataBinder binder) {
      if (binder.getTarget() != null) {
         if (requestProviderValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(requestProviderValidator);
         }
         if (eventQueryValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(eventQueryValidator);
         }
         if (configQueryValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(configQueryValidator);
         }
      }
   }

   @RequestMapping(
         value = DomainConfig.PATH_CHANNELS_HASH,
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public LongHash getChannelsHash() throws Throwable {
      return queryManager.getChannelsHash();
   }

   @RequestMapping(
         value = DomainConfig.PATH_CHANNELS,
         method = {RequestMethod.GET, RequestMethod.POST},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public void getChannels(
         @RequestBody(required = false) ChannelsRequest request,
         final HttpServletRequest httpRequest,
         final HttpServletResponse httpResponse)
         throws Throwable {
      // in case not specified use defaults (e.g. GET)
      if (request == null) {
         request = new ChannelsRequest();
      }

      ((AbstractHTTPResponse) defaultResponse).respond(
            context,
            httpResponse,
            request,
            queryManager.getChannels(request));
   }

   @RequestMapping(
         value = DomainConfig.PATH_CHANNELS + "/{channel}",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public void getChannels(
         @PathVariable(value = "channel") final String channelName,
         final HttpServletRequest httpRequest,
         final HttpServletResponse httpResponse)
         throws Throwable {
      getChannels(new ChannelsRequest(channelName), httpRequest, httpResponse);
   }

   @RequestMapping(
         value = DomainConfig.PATH_CHANNELS_CONFIG_HASH,
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public LongHash getChannelConfigurationsHash() throws Throwable {
      return queryManager.getChannelConfigurationsHash();
   }

   @RequestMapping(
         value = DomainConfig.PATH_CHANNELS_CONFIG,
         method = {RequestMethod.GET, RequestMethod.POST},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public void getChannelConfigurations(
         @RequestBody(required = false) ChannelConfigurationsRequest request,
         final HttpServletRequest httpRequest,
         final HttpServletResponse httpResponse) throws Throwable {
      // in case not specified use defaults (e.g. GET)
      if (request == null) {
         request = new ChannelConfigurationsRequest();
      }

      ((AbstractHTTPResponse) defaultResponse).respond(
            context,
            httpResponse,
            request,
            queryManager.getChannelConfigurations(request));
   }

   @RequestMapping(
         value = DomainConfig.PATH_CHANNELS_CONFIG + "/{channel}",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public void getChannelConfigurations(
         @PathVariable(value = "channel") final String channelName,
         final HttpServletRequest httpRequest,
         final HttpServletResponse httpResponse)
         throws Throwable {
      getChannelConfigurations(new ChannelConfigurationsRequest(channelName), httpRequest, httpResponse);
   }

   @RequestMapping(
         value = DomainConfig.PATH_CHANNEL_CONFIG,
         method = {RequestMethod.POST},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public void getChannelConfiguration(
         @RequestBody final ChannelName channelName,
         final HttpServletRequest httpRequest,
         final HttpServletResponse httpResponse)
         throws Throwable {
      try {
         final ChannelConfiguration config = queryManager.getChannelConfiguration(channelName);

         if (config != null) {
            ((AbstractHTTPResponse) defaultResponse).respond(
                  context,
                  httpResponse,
                  channelName,
                  config);
         } else {
            httpResponse.setStatus(HttpStatus.NOT_FOUND.value());
         }
      } catch (Throwable t) {
         LOGGER.warn("Failed to get channel configuration.", t);
         throw t;
      }
   }

   @RequestMapping(
         value = DomainConfig.PATH_CHANNEL_CONFIG + "/{channel}",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public void getChannelConfiguration(
         @PathVariable(value = "channel") final String channelName,
         final HttpServletRequest httpRequest,
         final HttpServletResponse httpResponse)
         throws Throwable {
      getChannelConfiguration(new ChannelName(channelName), httpRequest, httpResponse);
   }

   @RequestMapping(
         value = DomainConfig.PATH_QUERY_CONFIG,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public void executeDAQConfigQuery(
         @RequestBody @Valid final DAQConfigQuery query,
         final HttpServletRequest httpRequest,
         final HttpServletResponse httpResponse)
         throws Throwable {
      final Response response = query.getResponseOrDefault(defaultResponse);
      if (response.isAllowRedirect()) {
         // Do a redirection if only one backend is requested
         //
         final Supplier<Collection<ChannelName>> channelsSupplier = () -> query.getChannels();
         final String redirect = queryManager.getRedirection(channelsSupplier);
         if (redirect != null) {
            final String redirectURL = 
                  redirect
                  + httpRequest.getRequestURI()
                  + ((httpRequest.getQueryString() != null && StringUtils.NULL_STRING.equals(httpRequest.getQueryString())) ? StringUtils.QUESTION_MARK + httpRequest.getQueryString()
                        : StringUtils.EMPTY);

            LOGGER.info("Send redirect to '{}'.", redirectURL);
            httpResponse.setHeader(HttpHeaders.LOCATION, redirectURL);
            // use 307 - works for POST too
            httpResponse.setStatus(HttpServletResponse.SC_TEMPORARY_REDIRECT);
            return;
         }
      }

      try {
         LOGGER.debug("Executing query '{}'", query);

         if (response instanceof AbstractHTTPResponse) {
            LOGGER.debug("Executing config query '{}'", query);
            final AbstractHTTPResponse httpRes = ((AbstractHTTPResponse) response);

            httpRes.validateQuery(query);
            // execute query
            final Entry<DAQConfigQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>> result =
                  queryManager.queryConfigs(query);

            httpRes.respond(
                  context,
                  httpResponse,
                  query,
                  result);
         } else {
            final String message =
                  String.format(
                        "Expecting Response of type '%s' but received '%s'. Check JSON deserialization defined in '%s'",
                        AbstractHTTPResponse.class.getName(), response.getClass().getName(),
                        PolymorphicResponseMixIn.class.getName());
            LOGGER.error(message);
            throw new IllegalArgumentException(message);
         }
      } catch (Exception e) {
         LOGGER.error("Failed to execute config query '{}'.", query, e);
         throw e;
      }
   }

   /**
    * Accepts the properties for the {@link DAQQuery} instance as stringified JSON query string.
    * 
    * @param jsonBody The {@link DAQQuery} properties sent as a JSON string, i.e. this is the
    *        stringified body of the POST request method
    * @param httpRequest the {@link HttpServletRequest} instance associated with this request
    * @param httpResponse the {@link HttpServletResponse} instance associated with this request
    * @throws Exception if reading the JSON string fails or if the subsequent call fails
    */
   @RequestMapping(
         value = DomainConfig.PATH_QUERY,
         method = RequestMethod.GET)
   public void executeDAQQueryBodyAsString(
         @RequestParam final String jsonBody,
         final HttpServletRequest httpRequest,
         final HttpServletResponse httpResponse) throws Exception {
      DAQQuery query;
      try {
         query = objectMapper.readValue(jsonBody, DAQQuery.class);
      } catch (Exception e) {
         // somehow only needed for our test mokup environment (@RequestParam seems to do the
         // decoding on production server).
         LOGGER.info("Could not parse '{}' due to '{}'. Retry url decoded.", jsonBody, e.getMessage());
         query = objectMapper.readValue(URLDecoder.decode(jsonBody, StandardCharsets.UTF_8.name()), DAQQuery.class);
      }

      final Errors errors = new DirectFieldBindingResult(query, query.getClass().getName());
      if (requestProviderValidator.supports(query.getClass())) {
         requestProviderValidator.validate(query, errors);
      }
      if (eventQueryValidator.supports(query.getClass())) {
         eventQueryValidator.validate(query, errors);
      }

      final List<ObjectError> allErrors = errors.getAllErrors();
      if (allErrors.isEmpty()) {
         executeDAQQuery(query, httpRequest, httpResponse);
      } else {
         final String message = String.format("Could not parse '%s' due to '%s'.", jsonBody, errors.toString());
         LOGGER.error(message);
         throw new IllegalStateException(message);
      }
   }

   /**
    * Executes a single query.
    * 
    * @param query the {@link DAQQuery}
    * @param httpRequest the {@link HttpServletRequest} instance associated with this request
    * @param httpResponse the {@link HttpServletResponse} instance associated with this request
    * @throws Exception thrown if writing to the output stream fails
    */
   @RequestMapping(
         value = DomainConfig.PATH_QUERY,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public void executeDAQQuery(
         @RequestBody @Valid final DAQQuery query,
         final HttpServletRequest httpRequest,
         final HttpServletResponse httpResponse) throws Exception {
      executeDAQQueries(new DAQQueries(query), httpRequest, httpResponse);
   }

   /**
    * Accepts the properties for the {@link DAQQueries} instance as stringified JSON query string.
    * 
    * @param jsonBody The {@link DAQQueries} properties sent as a JSON string, i.e. this is the
    *        stringified body of the POST request method
    * @param httpRequest the {@link HttpServletRequest} instance associated with this request
    * @param httpResponse the {@link HttpServletResponse} instance associated with this request
    * @throws Exception if reading the JSON string fails or if the subsequent call fails
    */
   @RequestMapping(
         value = DomainConfig.PATH_QUERIES,
         method = RequestMethod.GET)
   public void executeDAQQueriesBodyAsString(
         @RequestParam final String jsonBody,
         final HttpServletRequest httpRequest,
         final HttpServletResponse httpResponse) throws Exception {
      DAQQueries queries;
      try {
         queries = objectMapper.readValue(jsonBody, DAQQueries.class);
      } catch (Exception e) {
         // somehow only needed for our test mokup environment (@RequestParam seems to do the
         // decoding on production server).
         LOGGER.info("Could not parse '{}' due to '{}'. Retry url decoded.", jsonBody, e.getMessage());
         queries = objectMapper.readValue(URLDecoder.decode(jsonBody, StandardCharsets.UTF_8.name()), DAQQueries.class);
      }

      final Errors errors = new DirectFieldBindingResult(queries, queries.getClass().getName());
      if (requestProviderValidator.supports(queries.getClass())) {
         requestProviderValidator.validate(queries, errors);
      }
      if (eventQueryValidator.supports(queries.getClass())) {
         eventQueryValidator.validate(queries, errors);
      }

      final List<ObjectError> allErrors = errors.getAllErrors();
      if (allErrors.isEmpty()) {
         executeDAQQueries(queries, httpRequest, httpResponse);
      } else {
         final String message = String.format("Could not parse '%s' due to '%s'.", jsonBody, errors.toString());
         LOGGER.error(message);
         throw new IllegalStateException(message);
      }
   }

   /**
    * Catch-all query method for getting data from the backend for both JSON and CSV requests.
    * <p>
    * The {@link DAQQueries} object will contain the concrete subclass based on the combination of
    * fields defined in the user's query. The AttributeBasedDeserializer decides which class to
    * deserialize the information into and has been configured (see
    * QueryRestConfig#afterPropertiesSet) accordingly.
    * 
    * @param queries the {@link DAQQueries}
    * @param httpRequest the {@link HttpServletRequest} instance associated with this request
    * @param httpResponse the {@link HttpServletResponse} instance associated with this request
    * @throws Exception thrown if writing to the output stream fails
    */
   @RequestMapping(
         value = DomainConfig.PATH_QUERIES,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public void executeDAQQueries(
         @RequestBody @Valid final DAQQueries queries,
         final HttpServletRequest httpRequest,
         final HttpServletResponse httpResponse) throws Exception {
      final Response response = queries.getResponseOrDefault(defaultResponse);
      if (response.isAllowRedirect()) {
         // Do a redirection if only one backend is requested
         //
         final Supplier<Collection<ChannelName>> channelsSupplier = () -> queries.getQueries().stream()
               .flatMap(daqQuery -> daqQuery.getChannels().stream())
               .collect(Collectors.toList());
         final String redirect = queryManager.getRedirection(channelsSupplier);
         if (redirect != null) {
            final String redirectURL = 
                  redirect
                  + httpRequest.getRequestURI()
                  + ((httpRequest.getQueryString() != null && StringUtils.NULL_STRING.equals(httpRequest.getQueryString())) ? StringUtils.QUESTION_MARK + httpRequest.getQueryString()
                        : StringUtils.EMPTY);

            LOGGER.info("Send redirect to '{}'.", redirectURL);
            httpResponse.setHeader(HttpHeaders.LOCATION, redirectURL);
            // use 307 - works for POST too
            httpResponse.setStatus(HttpServletResponse.SC_TEMPORARY_REDIRECT);
            return;
         }
      }

      try {
         LOGGER.debug("Executing queries '{}'", queries);

         if (response instanceof AbstractHTTPResponse) {
            LOGGER.debug("Executing query '{}'", queries);
            final AbstractHTTPResponse httpRes = ((AbstractHTTPResponse) response);

            httpRes.validateQuery(queries);
            // execute query
            final List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>> result =
                  queryManager.queryEvents(queries);

            httpRes.respond(
                  context,
                  httpResponse,
                  queries,
                  result);
         } else {
            final String message =
                  String.format(
                        "Expecting Response of type '%s' but received '%s'. Check JSON deserialization defined in '%s'",
                        AbstractHTTPResponse.class.getName(), response.getClass().getName(),
                        PolymorphicResponseMixIn.class.getName());
            LOGGER.error(message);
            throw new IllegalArgumentException(message);
         }
      } catch (Exception e) {
         LOGGER.error("Failed to execute query '{}'.", queries, e);
         throw e;
      }
   }

   /**
    * Returns the current list of {@link Ordering}s available.
    * 
    * @return list of {@link Ordering}s as String array
    */
   @RequestMapping(
         value = DomainConfig.PATH_PARAMETERS_ROOT + "/ordering",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Ordering> getOrderingValues() {
      return Lists.newArrayList(Ordering.values());
   }

   /**
    * Returns the current list of {@link ResponseFormat}s available.
    * 
    * @return list of {@link Ordering}s as String array
    */
   @RequestMapping(
         value = DomainConfig.PATH_PARAMETERS_ROOT + "/responseformat",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<ResponseFormat> getResponseFormatValues() {
      return Arrays.stream(ResponseFormat.values())
            .filter(format -> format.isPublish())
            .collect(Collectors.toList());
   }

   /**
    * Returns the current list of {@link ConfigField}s available.
    * 
    * @return list of {@link ConfigField}s as String array
    */
   @RequestMapping(
         value = DomainConfig.PATH_PARAMETERS_ROOT + "/configfields",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<ConfigField> getConfigFieldValues() {
      return Arrays.stream(ConfigField.values())
            .filter(queryField -> queryField.isPublish())
            .collect(Collectors.toList());
   }

   /**
    * Returns the current list of {@link EventField}s available.
    * 
    * @return list of {@link EventField}s as String array
    */
   @RequestMapping(
         value = DomainConfig.PATH_PARAMETERS_ROOT + "/eventfields",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<EventField> getEventFieldValues() {
      return Arrays.stream(EventField.values())
            .filter(queryField -> queryField.isPublish())
            .collect(Collectors.toList());
   }

   /**
    * Returns the current list of {@link EventField}s available.
    * 
    * @return list of {@link EventField}s as String array
    */
   @Deprecated
   @RequestMapping(
         value = DomainConfig.PATH_PARAMETERS_ROOT + "/queryfields",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<EventField> getQueryFieldValues() {
      return Arrays.stream(EventField.values())
            .filter(queryField -> queryField.isPublish())
            .collect(Collectors.toList());
   }

   /**
    * Returns the current list of {@link Aggregation}s available.
    * 
    * @return list of {@link Aggregation}s as String array
    */
   @RequestMapping(
         value = DomainConfig.PATH_PARAMETERS_ROOT + "/aggregations",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Aggregation> getAggregationValues() {
      return Lists.newArrayList(Aggregation.values());
   }

   /**
    * Returns the current list of {@link AggregationType}s available.
    * 
    * @return list of {@link AggregationType}s as String array
    */
   @RequestMapping(
         value = DomainConfig.PATH_PARAMETERS_ROOT + "/aggregationtypes",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<AggregationType> getAggregationTypeValues() {
      return Lists.newArrayList(AggregationType.values());
   }

   /**
    * Returns the current list of {@link Backend} values.
    * 
    * @return list of {@link Backend}s as String array
    */
   @RequestMapping(
         value = DomainConfig.PATH_BACKENDS,
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody Collection<Backend> getBackendValues() {
      return queryManager.getBackends()
            .collect(Collectors.toList());
   }

   /**
    * Returns the current list of {@link Backend} values.
    * 
    * @return list of {@link Backend}s as String array
    */
   @RequestMapping(
         value = DomainConfig.PATH_BACKENDS_BY_ID,
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody Map<Integer, Backend> getBackendValuesById() {
      return queryManager.getBackends()
            .collect(
                  Collectors.toMap(
                        Backend::getId,
                        Function.identity()));
   }

   /**
    * Returns the current list of {@link Compression}s available.
    * 
    * @return list of {@link Compression}s as String array
    */
   @RequestMapping(
         value = DomainConfig.PATH_PARAMETERS_ROOT + "/compression",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Compression> getCompressionValues() {
      return Lists.newArrayList(Compression.values());
   }

   /**
    * Returns the current list of {@link ValueAggregation}s available.
    * 
    * @return list of {@link ValueAggregation}s as String array
    */
   @RequestMapping(
         value = DomainConfig.PATH_PARAMETERS_ROOT + "/valueaggregations",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<ValueAggregation> getValueAggregations() {
      return Lists.newArrayList(ValueAggregation.values());
   }

   /**
    * Returns the current list of {@link ValueAggregation}s available.
    * 
    * @return list of {@link ValueAggregation}s as String array
    */
   @RequestMapping(
         value = DomainConfig.PATH_PARAMETERS_ROOT + "/colormodeltypes",
         method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<ColorModelType> getColorModelTypes() {
      return Lists.newArrayList(ColorModelType.values());
   }
}
