package ch.psi.daq.queryrest.controller.validator;

import java.util.LinkedHashSet;
import java.util.Set;

import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.validation.Errors;
import org.springframework.validation.Validator;

import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.common.time.Time;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.query.DAQConfigQuery;
import ch.psi.daq.domain.query.operation.ConfigField;
import ch.psi.daq.domain.request.range.RequestRange;
import ch.psi.daq.queryrest.config.QueryRestConfig;

public class ConfigQueryValidator implements Validator, ApplicationContextAware {
   private Set<ConfigField> queryResponseFields;

   private long maxPulseRange;
   private Time maxTimeRange;

   @SuppressWarnings("unchecked")
   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      final Backend backend = context.getBean(DomainConfig.BEAN_NAME_BACKEND_DEFAULT, Backend.class);
      context = backend.getApplicationContext();

      queryResponseFields = context.getBean(QueryRestConfig.BEAN_NAME_CONFIG_RESPONSE_FIELDS_QUERY, Set.class);

      maxPulseRange = context.getBean(DomainConfig.BEAN_NAME_QUERY_MAX_PULSE_RANGE, Long.class);
      maxTimeRange = context.getBean(DomainConfig.BEAN_NAME_QUERY_MAX_TIME_RANGE, Time.class);
   }

   @Override
   public boolean supports(final Class<?> clazz) {
      return DAQConfigQuery.class.isAssignableFrom(clazz);
   }

   @Override
   public void validate(final Object target, final Errors errors) {
      if (target instanceof DAQConfigQuery) {
         this.checkElement((DAQConfigQuery) target, errors);
      }
   }

   private void checkElement(final DAQConfigQuery query, final Errors errors) {
      // set default values (if not set)
      if (query.getConfigFields() == null || query.getConfigFields().isEmpty()) {
         query.setConfigFields(new LinkedHashSet<>(queryResponseFields));
      }

      RequestRange range = query.getRange();
      if (query.hasLimit()) {
         if (Ordering.asc.equals(query.getOrdering())) {
            if (range.isPulseIdRangeDefined() && range.getStartPulseId() < 0) {
               errors.reject("limit",
                     "Pulse based query with limit and asc order requires start pulse-id to be defined.");
            }
            if (range.isTimeRangeDefined() && range.getStartTime() == null) {
               errors.reject("limit",
                     "Time based query with limit and asc order requires start time to be defined.");
            }
         } else if (Ordering.desc.equals(query.getOrdering())) {
            if (range.isPulseIdRangeDefined() && range.getEndPulseId() <= 0) {
               errors.reject("limit",
                     "Pulse based query with limit and desc order requires end pulse-id to be defined.");
            }
            if (range.isTimeRangeDefined() && range.getEndTime() == null) {
               errors.reject("limit", "Time based query with limit and desc order requires end time to be defined.");
            }
         }
      }

      range.validate(maxPulseRange, maxTimeRange, query.getOrdering(), query.getLimit());
      query.setRange(range);
   }
}
