package ch.psi.daq.queryrest.controller.validator;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.Set;

import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.validation.Errors;
import org.springframework.validation.Validator;

import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.common.time.Time;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.mapping.Mapping;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.EventField;
import ch.psi.daq.domain.query.response.Response;
import ch.psi.daq.domain.query.response.ResponseFormat;
import ch.psi.daq.domain.query.transform.ExecutionEnvironment;
import ch.psi.daq.domain.query.transform.ValueTransformationSequence;
import ch.psi.daq.domain.request.Request;
import ch.psi.daq.domain.request.range.RequestRange;
import ch.psi.daq.queryrest.config.QueryRestConfig;

public class EventQueryValidator implements Validator, ApplicationContextAware {
   private Set<EventField> defaultResponseFields;
   private Set<Aggregation> defaultResponseAggregations;
   private long maxPulseRange;
   private Time maxTimeRange;

   @SuppressWarnings("unchecked")
   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      final Backend backend = context.getBean(DomainConfig.BEAN_NAME_BACKEND_DEFAULT, Backend.class);
      context = backend.getApplicationContext();

      defaultResponseFields = context.getBean(QueryRestConfig.BEAN_NAME_DEFAULT_EVENT_RESPONSE_FIELDS, Set.class);
      defaultResponseAggregations =
            context.getBean(QueryRestConfig.BEAN_NAME_DEFAULT_EVENT_RESPONSE_AGGREGATIONS, Set.class);

      maxPulseRange = context.getBean(DomainConfig.BEAN_NAME_QUERY_MAX_PULSE_RANGE, Long.class);
      maxTimeRange = context.getBean(DomainConfig.BEAN_NAME_QUERY_MAX_TIME_RANGE, Time.class);
   }

   @Override
   public boolean supports(final Class<?> clazz) {
      return DAQQuery.class.isAssignableFrom(clazz) || DAQQueries.class.isAssignableFrom(clazz);
   }

   @Override
   public void validate(final Object target, final Errors errors) {
      if (target instanceof DAQQuery) {
         this.checkElement(((DAQQuery) target).getResponse(), (DAQQuery) target, errors);
      } else if (target instanceof DAQQueries) {
         final DAQQueries queries = (DAQQueries) target;
         for (final DAQQueryElement daqQueryElement : queries) {
            this.checkElement(queries.getResponse(), daqQueryElement, errors);
         }
      }
   }

   private void checkElement(final Response response, final DAQQueryElement query, final Errors errors) {
      // set default values (if not set)
      if (query.getEventFields() == null || query.getEventFields().isEmpty()) {
         query.setEventFields(new LinkedHashSet<>(defaultResponseFields));
      }

      RequestRange range = query.getRange();
      if (query.hasLimit()) {
         if (Ordering.asc.equals(query.getOrdering())) {
            if (range.isPulseIdRangeDefined() && range.getStartPulseId() < 0) {
               errors.reject("limit",
                     "Pulse based query with limit and asc order requires start pulse-id to be defined.");
            }
            if (range.isTimeRangeDefined() && range.getStartTime() == null) {
               errors.reject("limit",
                     "Time based query with limit and asc order requires start time to be defined.");
            }
         } else if (Ordering.desc.equals(query.getOrdering())) {
            if (range.isPulseIdRangeDefined() && range.getEndPulseId() <= 0) {
               errors.reject("limit",
                     "Pulse based query with limit and desc order requires end pulse-id to be defined.");
            }
            if (range.isTimeRangeDefined() && range.getEndTime() == null) {
               errors.reject("limit", "Time based query with limit and desc order requires end time to be defined.");
            }
         }
      }

      if (range.isStartExpansion() && !range.isStartInclusive()) {
         errors.reject("startExpansion", "Start needs to be inclusive for start-expansion.");
      }

      range.validate(maxPulseRange, maxTimeRange, query.getOrdering(), query.getLimit());
      query.setRange(range);

      if (query.getAggregation() != null) {
         // check if only one binning element is defined
         final long durationPerBin = query.getAggregation().getDurationPerBin();
         final long pulsesPerBin = query.getAggregation().getPulsesPerBin();
         final int nrOfBins = query.getAggregation().getNrOfBins();
         if ((durationPerBin != Request.NOT_SET && (pulsesPerBin != Request.NOT_SET || nrOfBins != Request.NOT_SET))
               || (pulsesPerBin != Request.NOT_SET
                     && (durationPerBin != Request.NOT_SET || nrOfBins != Request.NOT_SET))
               || (nrOfBins != Request.NOT_SET
                     && (durationPerBin != Request.NOT_SET || pulsesPerBin != Request.NOT_SET))) {
            errors.reject("durationPerBin", "Only one binning element must be defined.");
            errors.reject("pulsesPerBin", "Only one binning element must be defined.");
            errors.reject("nrOfBins", "Only one binning element must be defined.");
         }

         if (query.getRange().isPulseIdRangeDefined() && durationPerBin != Request.NOT_SET) {
            errors.reject("durationPerBin", "Pulse range queries only support pulse based binning.");
         }
         if (query.getRange().isTimeRangeDefined() && pulsesPerBin != Request.NOT_SET) {
            errors.reject("pulsesPerBin", "Time range queries only support time based binning.");
         }

         // set default values (if not set)
         if (query.getAggregation().getAggregations() == null || query.getAggregation().getAggregations().isEmpty()) {
            query.getAggregation().setAggregations(new ArrayList<>(defaultResponseAggregations));
         }

         if (query.hasLimit()) {
            errors.reject("limit", "Aggregation with limit does not make sense.");
         }

         // without adding this field, user need to explicitly ask for value field when querying
         // aggregations.
         // if (!query.getFields().contains(QueryField.value)) {
         // // without this field, json will not contain Stats
         // query.addField(QueryField.value);
         // }
      }

      if (query.getValueTransformations() != null && !query.getValueTransformations().isEmpty()) {
         // without this field, json will not contain transformedValue
         query.addEventField(EventField.transformedValue);

         for (final ValueTransformationSequence transformationSequence : query.getValueTransformations()) {
            transformationSequence.setExecutionEnvironment(ExecutionEnvironment.QUERYING);
         }
      }

      if (response != null && response.getFormat() == ResponseFormat.CSV && query.getMapping() == null) {
         query.setMapping(new Mapping());
      }
   }
}
