package ch.psi.daq.queryrest.query;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.commons.lang3.tuple.Triple;

import ch.psi.daq.common.tuple.Quadruple;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQConfigQuery;
import ch.psi.daq.domain.query.DAQConfigQueryElement;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.backend.BackendQueryImpl;
import ch.psi.daq.domain.query.backend.analyzer.BackendQueryAnalyzer;
import ch.psi.daq.domain.query.channels.BackendsChannelConfigurationCache;
import ch.psi.daq.domain.query.channels.ChannelConfigurationsRequest;
import ch.psi.daq.domain.query.channels.ChannelConfigurationsResponse;
import ch.psi.daq.domain.query.channels.ChannelsRequest;
import ch.psi.daq.domain.query.channels.ChannelsResponse;
import ch.psi.daq.domain.query.channels.LongHash;
import ch.psi.daq.domain.query.processor.QueryProcessor;

public abstract class AbstractQueryManager implements QueryManager {
   private Set<Backend> activeBackends;
   private BackendsChannelConfigurationCache channelsCache;
   private Function<BackendQuery, BackendQueryAnalyzer> queryAnalizerFactory;

   protected void init(
         final Set<Backend> activeBackends,
         final BackendsChannelConfigurationCache channelsCache,
         final Function<BackendQuery, BackendQueryAnalyzer> queryAnalizerFactory) {
      this.activeBackends = activeBackends;
      this.channelsCache = channelsCache;
      this.queryAnalizerFactory = queryAnalizerFactory;
   }

   protected Set<Backend> getActiveBackends() {
      return activeBackends;
   }

   protected BackendsChannelConfigurationCache getConfigurationCache() {
      return channelsCache;
   }

   protected Function<BackendQuery, BackendQueryAnalyzer> getQueryAnalizerFactory() {
      return queryAnalizerFactory;
   }

   @Override
   public Stream<Backend> getBackends() {
      return Backend.getBackends().stream()
            .filter(backend -> activeBackends.contains(backend));
   }

   @Override
   public LongHash getChannelsHash() {
      return channelsCache.getChannelsHash();
   }

   @Override
   public Stream<ChannelsResponse> getChannels(
         final ChannelsRequest request) {
      return channelsCache.getChannels(request);
   }

   @Override
   public LongHash getChannelConfigurationsHash() {
      return channelsCache.getChannelConfigurationsHash();
   }

   @Override
   public Stream<ChannelConfigurationsResponse> getChannelConfigurations(
         final ChannelConfigurationsRequest request) {
      return channelsCache.getChannelConfigurations(request);
   }

   @Override
   public ChannelConfiguration getChannelConfiguration(
         final ChannelName channel) {
      return channelsCache.getChannelConfiguration(channel);
   }

   @Override
   public Entry<DAQConfigQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>> queryConfigs(
         final DAQConfigQuery daqQuery) {
      // set backends if not defined yet
      channelsCache.configureBackends(daqQuery.getChannels());

      Stream<Triple<BackendQuery, ChannelName, ?>> resultStreams =
            BackendQueryImpl
                  .getBackendQueries(daqQuery)
                  .stream()
                  .filter(
                        query -> query.getBackend().getBackendAccess().hasDataReader())
                  .flatMap(
                        query -> {
                           /* all the magic happens here */
                           final Map<String, Stream<? extends ChannelConfiguration>> channelToConfig =
                                 query.getChannelConfigurations();

                           return channelToConfig.entrySet().stream()
                                 .map(entry -> {
                                    return Triple.of(
                                          query,
                                          new ChannelName(entry.getKey(), query.getBackend()),
                                          entry.getValue());
                                 });
                        });

      return Pair.of(daqQuery, resultStreams);
   }

   @Override
   public List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>> queryEvents(
         final DAQQueries queries) {
      // set backends if not defined yet
      for (DAQQueryElement daqQuery : queries) {
         channelsCache.configureBackends(daqQuery.getChannels());
      }

      final List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>> results =
            new ArrayList<>(queries.getQueries().size());

      for (final DAQQueryElement queryElement : queries) {
         Stream<Quadruple<BackendQuery, ChannelName, ?, ?>> resultStreams =
               BackendQueryImpl
                     .getBackendQueries(queryElement)
                     .stream()
                     .filter(
                           query -> {
                              return query.getBackend().getBackendAccess().hasDataReader()
                                 && query.getBackend().getBackendAccess().hasQueryProcessor();})
                     .flatMap(
                           query -> {
                              final QueryProcessor processor =
                                    query.getBackend().getBackendAccess().getQueryProcessor();
                              final BackendQueryAnalyzer queryAnalizer = queryAnalizerFactory.apply(query);

                              // ChannelEvent query
                              /* all the magic happens here */
                              final Stream<Entry<ChannelName, Stream<? extends DataEvent>>> channelToDataEvents =
                                    processor.process(queryAnalizer);
                              /* do post-process */
                              final Stream<Entry<ChannelName, ?>> channelToData =
                                    queryAnalizer.postProcess(channelToDataEvents);

                              // ChannelConfig query
                              final BackendQuery configQuery =
                                    new BackendQueryImpl(query, queryElement.getConfigFields());
                              final Map<String, Stream<? extends ChannelConfiguration>> channelToConfig =
                                    configQuery.getChannelConfigurations();

                              return channelToData.map(entry -> {
                                 return Quadruple.of(
                                       query,
                                       entry.getKey(),
                                       channelToConfig.get(entry.getKey().getName()),
                                       entry.getValue());
                              });
                           });

         // Now we have a stream that loads elements sequential BackendQuery by BackendQuery.
         // By materializing the outer Stream the elements of all BackendQuery are loaded async
         // (speeds things up but requires also more memory - i.e. it relies on Backends not loading
         // all elements into memory at once)
         resultStreams = resultStreams.collect(Collectors.toList()).stream();

         results.add(Pair.of(queryElement, resultStreams));
      }

      return results;
   }

}
