package ch.psi.daq.queryrest.query;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.annotation.PreDestroy;

import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;

import ch.psi.daq.common.tuple.Quadruple;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.StreamEvent;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.backend.BackendQueryImpl;
import ch.psi.daq.domain.query.backend.analyzer.BackendQueryAnalyzer;
import ch.psi.daq.domain.query.channels.BackendsChannelConfigurationCache;
import ch.psi.daq.domain.query.channels.ChannelConfigurationLoader;
import ch.psi.daq.domain.query.mapping.IncompleteStrategy;
import ch.psi.daq.domain.query.response.Response;
import ch.psi.daq.domain.query.response.ResponseFormat;
import ch.psi.daq.domain.query.response.ResponseImpl;
import ch.psi.daq.domain.rest.RestHelper;
import ch.psi.daq.query.config.QueryConfig;
import ch.psi.daq.queryrest.config.QueryRestConfig;

import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

public class QueryManagerRemote extends AbstractQueryManager implements ApplicationContextAware {
   private static final Logger LOGGER = LoggerFactory.getLogger(QueryManagerRemote.class);

   public static final String QUERY_SERVER_TYPE = "remote";
   private Map<Backend, String> backendToServerAddresses;

   @SuppressWarnings("unchecked")
   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      backendToServerAddresses = context.getBean(QueryRestConfig.BEAN_NAME_QUERY_SERVER_BACKENDS, Map.class);

      final long reloadPeriodMillis = context.getBean(QueryConfig.BEAN_NAME_CHANNELS_CACHE_RELOAD_PERIOD, Long.class);
      final Function<Backend, ChannelConfigurationLoader> loaderProvider = backend -> {
         if (backend.getBackendAccess().hasHistoricChannelConfigurationLoader()) {
            return backend.getBackendAccess().getHistoricChannelConfigurationLoader();
         } else {
            return null;
         }
      };
      final BackendsChannelConfigurationCache channelsCache =
            new BackendsChannelConfigurationCache(loaderProvider, reloadPeriodMillis);
      channelsCache.init(
            context,
            backendToServerAddresses.keySet().iterator().next(),
            backendToServerAddresses.keySet());
      final Function<BackendQuery, BackendQueryAnalyzer> queryAnalizerFactory =
            context.getBean(QueryRestConfig.BEAN_NAME_QUERY_ANALIZER_FACTORY, Function.class);

      init(backendToServerAddresses.keySet(), channelsCache, queryAnalizerFactory);
   }

   @PreDestroy
   public void destroy() {}

   // @Override
   // public Stream<Backend> getBackends() {
   // return backendToServerAddresses.keySet().stream();
   // }

   @Override
   public String getRedirection(final Supplier<Collection<ChannelName>> channelsSupplier) {
      // set backends if not defined yet
      final Collection<ChannelName> channels = channelsSupplier.get();
      getConfigurationCache().configureBackends(channels);

      final Set<String> backendRoots = channels.stream()
            .map(channelName -> backendToServerAddresses.get(channelName.getBackend()))
            .collect(Collectors.toSet());

      if (backendRoots.size() == 1) {
         // request can be sent to one REST backend -> do a redirection
         return backendRoots.iterator().next();
      } else {
         return null;
      }
   }

   @Override
   public List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>> queryEvents(
         final DAQQueries queries) {
      // INFO: It is always an option to call super.queryEvents(queries);
      // The super call will use QueryRestStreamEventReader. However this will be slower because
      // more data is transfered (aggregation info for e.g. index/bin is not available on
      // QueryRestStreamEventReader level).

      // set backends if not defined yet
      for (DAQQueryElement daqQuery : queries) {
         getConfigurationCache().configureBackends(daqQuery.getChannels());
      }

      final List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>> results =
            new ArrayList<>(queries.getQueries().size());
      // TODO: consider rawevent -> ???
      final Response response = new ResponseImpl(ResponseFormat.JSON);

      for (final DAQQueryElement queryElement : queries) {
         Stream<Quadruple<BackendQuery, ChannelName, ?, ?>> resultStreams =
               BackendQueryImpl
                     .getBackendQueries(queryElement)
                     .stream()
                     .flatMap(
                           query -> {
                              final String queryServer = backendToServerAddresses.get(query.getBackend());

                              final Stream<Entry<ChannelName, Stream<? extends DataEvent>>> channelToDataEvents =
                                    query.getChannels()
                                          .stream()
                                          .map(channel -> {
                                             if (queryServer == null) {
                                                LOGGER.warn(
                                                      "There is no query server defined for '{}' of '{}'. Provide empty stream.",
                                                      channel, query.getBackend());
                                                return Pair.of(new ChannelName(channel, query.getBackend()),
                                                      Stream.empty());
                                             } else {
                                                final ChannelName channelName =
                                                      new ChannelName(channel, query.getBackend());
                                                final DAQQuery daqQuery =
                                                      new DAQQuery(queryElement, response, channelName);
                                                final Stream<StreamEvent> events;
                                                if (daqQuery.getMapping() != null) {
                                                   daqQuery.getMapping().setIncomplete(IncompleteStrategy.FILL_NULL);
                                                   events = RestHelper.queryEventsTableAsync(
                                                         query.getBackend().getApplicationContext(), queryServer,
                                                         daqQuery)
                                                         .onErrorResume(thrw -> {
                                                            LOGGER.warn(
                                                                  "Could not query '{}' for '{}' of '{}' as table. Provide empty stream.",
                                                                  queryServer, channel, query.getBackend(), thrw);
                                                            return Mono.empty();
                                                         })
                                                         .flatMapMany(eventTable -> Flux
                                                               .fromStream(eventTable.getEventsOfColumn(0)))
                                                         .toStream();
                                                } else {
                                                   events = RestHelper
                                                         .queryEventsAsync(query.getBackend().getApplicationContext(),
                                                               queryServer, daqQuery)
                                                         .onErrorResume(thrw -> {
                                                            LOGGER.warn(
                                                                  "Could not query '{}' for '{}' of '{}'. Provide empty stream.",
                                                                  queryServer, channel, query.getBackend(), thrw);
                                                            return Flux.empty();
                                                         })
                                                         .flatMap(channelEvents -> Flux
                                                               .fromStream(channelEvents.getEvents()))
                                                         .toStream();
                                                }

                                                return Pair.of(channelName, events);
                                             }
                                          });


                              final BackendQueryAnalyzer queryAnalizer = getQueryAnalizerFactory().apply(query);

                              /* do post-process */
                              final Stream<Entry<ChannelName, ?>> channelToData =
                                    queryAnalizer.postProcess(channelToDataEvents);

                              // ChannelConfig query
                              final BackendQuery configQuery =
                                    new BackendQueryImpl(query, queryElement.getConfigFields());
                              final Map<String, Stream<? extends ChannelConfiguration>> channelToConfig =
                                    configQuery.getChannelConfigurations();

                              return channelToData.map(entry -> {
                                 return Quadruple.of(
                                       query,
                                       entry.getKey(),
                                       channelToConfig.get(entry.getKey().getName()),
                                       entry.getValue());
                              });
                           });

         // Now we have a stream that loads elements sequential BackendQuery by BackendQuery.
         // By materializing the outer Stream the elements of all BackendQuery are loaded async
         // (speeds things up but requires also more memory - i.e. it relies on Backends not loading
         // all elements into memory at once)
         resultStreams = resultStreams.collect(Collectors.toList()).stream();

         results.add(Pair.of(queryElement, resultStreams));
      }

      return results;
   }
}
