package ch.psi.daq.queryrest.response.csv;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.function.ToLongFunction;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.servlet.ServletResponse;

import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVPrinter;
import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;

import ch.psi.daq.common.statistic.StorelessStatistics;
import ch.psi.daq.common.stream.StreamIterable;
import ch.psi.daq.common.stream.match.MapFiller;
import ch.psi.daq.common.stream.match.Padder;
import ch.psi.daq.common.stream.match.ReusingMapCreator;
import ch.psi.daq.common.stream.match.StreamMatcher;
import ch.psi.daq.common.string.stringifier.ObjectStringifier;
import ch.psi.daq.common.string.stringifier.Stringifier;
import ch.psi.daq.common.tuple.Quadruple;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.backend.analyzer.BackendQueryAnalyzer;
import ch.psi.daq.domain.query.bin.BinningStrategy;
import ch.psi.daq.domain.query.mapping.Alignment;
import ch.psi.daq.domain.query.mapping.IncompleteStrategy;
import ch.psi.daq.domain.query.mapping.Mapping;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.ConfigField;
import ch.psi.daq.domain.query.operation.EventField;
import ch.psi.daq.domain.query.operation.Extrema;
import ch.psi.daq.domain.query.response.Response;
import ch.psi.daq.domain.request.range.RequestRange;
import ch.psi.daq.query.bin.aggregate.BinnedValueCombiner;
import ch.psi.daq.queryrest.config.QueryRestConfig;
import ch.psi.daq.queryrest.response.AbstractHTTPResponse;
import ch.psi.daq.queryrest.response.ResponseStreamWriter;
import ch.psi.daq.queryrest.response.formatter.DAQQueriesResponseFormatter;

/**
 * Takes a Java 8 stream and writes it to the output stream provided by the {@link ServletResponse}
 * of the current request.
 */
public class CSVResponseStreamWriter implements ResponseStreamWriter, ApplicationContextAware {
   private static final Logger LOGGER = LoggerFactory.getLogger(CSVResponseStreamWriter.class);

   public static final Mapping DEFAULT_MAPPING = new Mapping(IncompleteStrategy.FILL_NULL);
   public static final char DELIMITER_CVS = ';';
   public static final String DELIMITER_ARRAY = ",";
   public static final char DELIMITER_CHANNELNAME_FIELDNAME = '.';
   public static final String EMPTY_VALUE = "";
   public static final String FIELDNAME_EXTREMA = "extrema";

   private ApplicationContext context;
   private Function<BackendQuery, BackendQueryAnalyzer> queryAnalizerFactory;

   @SuppressWarnings("unchecked")
   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      final Backend backend = context.getBean(DomainConfig.BEAN_NAME_BACKEND_DEFAULT, Backend.class);
      context = backend.getApplicationContext();
      this.context = context;

      queryAnalizerFactory = context.getBean(QueryRestConfig.BEAN_NAME_QUERY_ANALIZER_FACTORY, Function.class);
   }

   @SuppressWarnings("unchecked")
   @Override
   public <R> void respond(
         final Object query,
         final R result,
         final OutputStream out,
         final AbstractHTTPResponse response) throws Exception {
      if (query instanceof DAQQueries) {
         respond((List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>>) result,
               out, response);
      } else {
         final String message = String.format("'%s' has no response type for '%s'.", query);
         LOGGER.error(message);
         throw new IllegalStateException(message);
      }
   }

   @SuppressWarnings("unchecked")
   public void respond(final List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>> results,
         final OutputStream out, final Response response) throws Exception {
      if (results.size() > 1) {
         throw new IllegalStateException("CSV format does not allow for multiple queries.");
      }

      final AtomicReference<Exception> exception = new AtomicReference<>();

      final Map<ChannelName, List<ChannelConfiguration>> configLists = new LinkedHashMap<>(results.size());
      final List<String> configHeader = new ArrayList<>();
      final Collection<Pair<ChannelName, Function<ChannelConfiguration, String>>> configAccessors = new ArrayList<>();

      final Map<ChannelName, Stream<DataEvent>> eventStreams = new LinkedHashMap<>(results.size());
      final List<String> eventHeader = new ArrayList<>();
      final Collection<Pair<ChannelName, Function<DataEvent, String>>> eventAccessors = new ArrayList<>();
      final AtomicReference<DAQQueryElement> daqQueryRef = new AtomicReference<>();
      final AtomicReference<BackendQuery> backendQueryRef = new AtomicReference<>();

      /* get DataEvent stream of all sub-queries for later match */
      results.forEach(entry -> {
         final DAQQueryElement query = entry.getKey();
         daqQueryRef.compareAndSet(null, query);

         entry.getValue()
               .sequential()
               .forEach(quadruple -> {
                  // single BackendQuery instance for all
                  backendQueryRef.compareAndSet(null, quadruple.getFirst());

                  if (query.hasConfigFields() && quadruple.getThird() instanceof Stream) {
                     setupChannelConfigColumns(query, quadruple.getFirst(), quadruple.getSecond(), configHeader,
                           configAccessors);

                     final List<ChannelConfiguration> configList =
                           ((Stream<ChannelConfiguration>) quadruple.getThird())
                                 .collect(Collectors.toList());
                     configLists.put(quadruple.getSecond(), configList);
                  }

                  if (quadruple.getFourth() instanceof Stream) {
                     setupChannelEventColumns(query, quadruple.getFirst(), quadruple.getSecond(), eventHeader,
                           eventAccessors);

                     final Stream<DataEvent> eventStream = ((Stream<DataEvent>) quadruple.getFourth());
                     eventStreams.put(quadruple.getSecond(), eventStream);
                  } else {
                     final String message = String.format("Expect a DataEvent Stream for '%s'.", quadruple.getSecond());
                     LOGGER.warn(message);
                  }
               });
      });

      final DAQQueryElement daqQuery = daqQueryRef.get();
      final BackendQuery backendQuery = backendQueryRef.get();
      final RequestRange requestRange = backendQuery.getRequest().getRequestRange();
      BinningStrategy binningStrategy = backendQuery.getBinningStrategy();

      final Mapping mapping = daqQueryRef.get().getMappingOrDefault(DEFAULT_MAPPING);
      final Padder<ChannelName, DataEvent> padder = mapping.getIncomplete().getPadder(backendQueryRef.get());

      ToLongFunction<DataEvent> alignmentProvider =
            Alignment.getAlignmentProvider(daqQuery.getMapping().getAlignment(), binningStrategy);
      if (binningStrategy == null) {
         binningStrategy = Alignment.getBinningStrategy(daqQuery.getMapping().getAlignment(), requestRange);
      }
      BiFunction<DataEvent, DataEvent, DataEvent> valueCombiner = null;
      if (binningStrategy != null) {
         binningStrategy.setRequestRange(requestRange);
         valueCombiner = new BinnedValueCombiner(binningStrategy);
      }

      // online matching of the stream's content
      final StreamMatcher<ChannelName, DataEvent, Map<ChannelName, DataEvent>> streamMatcher =
            new StreamMatcher<>(
                  DAQQueriesResponseFormatter.KEY_PROVIDER,
                  alignmentProvider,
                  new ReusingMapCreator<>(),
                  new MapFiller<>(),
                  valueCombiner,
                  padder,
                  eventStreams.keySet(),
                  eventStreams.values());
      final Iterator<Map<ChannelName, DataEvent>> streamsMatchIter = streamMatcher.iterator();

      // prepare csv output
      final CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(DELIMITER_CVS);
      Writer writer = null;
      CSVPrinter csvFilePrinter = null;

      try {
         writer = new BufferedWriter(new OutputStreamWriter(out, StandardCharsets.UTF_8));
         // writer = new BufferedWriter(new OutputStreamWriter(out,
         // response.getCharacterEncoding()));
         csvFilePrinter = new CSVPrinter(writer, csvFormat);

         if (!configHeader.isEmpty()) {
            csvFilePrinter.printRecord(configHeader);

            // ensure correct order
            final Stream<String> rowStream = configAccessors.stream().sequential()
                  .map(accessorPair -> {
                     final List<ChannelConfiguration> configs = configLists.get(accessorPair.getKey());
                     if (!configs.isEmpty()) {
                        // print the last (current) config
                        return accessorPair.getValue().apply(configs.get(configs.size() - 1));
                     } else {
                        return EMPTY_VALUE;
                     }
                  });

            csvFilePrinter.printRecord(new StreamIterable<String>(rowStream));
            csvFilePrinter.println();
         }

         csvFilePrinter.printRecord(eventHeader);

         while (streamsMatchIter.hasNext()) {
            final Map<ChannelName, DataEvent> match = streamsMatchIter.next();

            // ensure correct order
            final Stream<String> rowStream = eventAccessors.stream().sequential()
                  .map(accessorPair -> {
                     DataEvent event = match.get(accessorPair.getKey());
                     if (event != null) {
                        return accessorPair.getValue().apply(event);
                     } else {
                        return EMPTY_VALUE;
                     }
                  });

            csvFilePrinter.printRecord(new StreamIterable<String>(rowStream));
         }
      } catch (IOException e) {
         LOGGER.error("Could not write CSV.", e);
         exception.compareAndSet(null, e);
      } finally {
         try {
            if (writer != null) {
               writer.flush();
               writer.close();
            }

            if (streamMatcher != null) {
               try {
                  streamMatcher.close();
               } catch (Throwable t) {
                  LOGGER.error("Something went wrong while closing stream matcher for csv response writer.", t);
               }
            }

            if (csvFilePrinter != null) {
               csvFilePrinter.close();
            }
         } catch (IOException e) {
            LOGGER.error("Could not close CSV writer.", e);
            exception.compareAndSet(null, e);
         }
      }

      if (exception.get() != null) {
         throw exception.get();
      }
   }


   private void setupChannelConfigColumns(
         final DAQQueryElement daqQuery,
         final BackendQuery backendQuery,
         final ChannelName channelName,
         final Collection<String> header,
         Collection<Pair<ChannelName, Function<ChannelConfiguration, String>>> accessors) {
      final Set<ConfigField> queryFields = daqQuery.getConfigFields();

      for (final ConfigField field : queryFields) {
         final StringBuilder buf = new StringBuilder(3)
               .append(channelName.getName())
               .append(DELIMITER_CHANNELNAME_FIELDNAME)
               .append(field.name());

         header.add(buf.toString());
         accessors.add(Pair.of(
               channelName,
               new ConfigFieldStringifyer(field.getAccessor(),
                     EMPTY_VALUE,
                     DELIMITER_ARRAY,
                     null)));
      }
   }

   @SuppressWarnings("unchecked")
   private void setupChannelEventColumns(
         final DAQQueryElement daqQuery,
         final BackendQuery backendQuery,
         final ChannelName channelName,
         final Collection<String> header, Collection<Pair<ChannelName, Function<DataEvent, String>>> accessors) {
      final Set<EventField> eventFields = daqQuery.getEventFields();
      final List<Aggregation> aggregations =
            daqQuery.getAggregation() != null ? daqQuery.getAggregation().getAggregations() : null;
      final List<Extrema> extrema = daqQuery.getAggregation() != null ? daqQuery.getAggregation().getExtrema() : null;
      final List<Aggregation> mappingAggregations =
            daqQuery.getMapping() != null ? daqQuery.getMapping().getAggregations() : null;

      final BackendQueryAnalyzer queryAnalyzer = queryAnalizerFactory.apply(backendQuery);

      for (final EventField field : eventFields) {
         if (!(EventField.value.equals(field) && queryAnalyzer.isAggregationEnabled())) {
            final StringBuilder buf = new StringBuilder(3)
                  .append(channelName.getName())
                  .append(DELIMITER_CHANNELNAME_FIELDNAME)
                  .append(field.name());

            // in case Mapping maps several events to the same bin -> values will be aggregated into
            // BinnedValueCombinedDataEvent or BinnedIndexCombinedDataEvent
            Collection<Aggregation> locMappingAggregations = null;
            if (aggregations == null) {
               // only when general aggregations is disabled
               // -> otherwise values will already be globally aggregated
               locMappingAggregations = mappingAggregations;
               if (locMappingAggregations == null) {
                  // use default -> should there be an extra field
                  locMappingAggregations =
                        context.getBean(QueryRestConfig.BEAN_NAME_DEFAULT_EVENT_RESPONSE_AGGREGATIONS, Set.class);
               }
            }

            Map<Class<?>, Stringifier> stringifiers = null;
            if (locMappingAggregations != null) {
               final Stringifier statsStringifier;
               if (locMappingAggregations.size() == 1) {
                  // if only one aggregation is active, write only this value
                  // -> makes it compatible with scalar/array layout
                  statsStringifier = new AggregationStringifyer(locMappingAggregations.iterator().next(), EMPTY_VALUE);
               } else {
                  final Map<String, Stringifier> aggregationsStringifiers =
                        new LinkedHashMap<>(locMappingAggregations.size());
                  for (final Aggregation aggregation : locMappingAggregations) {
                     aggregationsStringifiers.put(aggregation.name(),
                           new AggregationStringifyer(aggregation, EMPTY_VALUE));
                  }
                  statsStringifier = new ObjectStringifier(aggregationsStringifiers);
               }
               stringifiers = new HashMap<>(2);
               stringifiers.put(StorelessStatistics.class, statsStringifier);
               stringifiers.put(StorelessStatistics[].class, statsStringifier);
            }
            header.add(buf.toString());
            accessors.add(Pair.of(
                  channelName,
                  new EventFieldStringifyer(
                        field.getAccessor(),
                        EMPTY_VALUE,
                        DELIMITER_ARRAY,
                        stringifiers)));
         }
      }

      if (aggregations != null && queryAnalyzer.isAggregationEnabled()) {
         for (final Aggregation aggregation : aggregations) {
            final StringBuilder buf = new StringBuilder(5)
                  .append(channelName.getName())
                  .append(DELIMITER_CHANNELNAME_FIELDNAME)
                  .append(EventField.value.name())
                  .append(DELIMITER_CHANNELNAME_FIELDNAME)
                  .append(aggregation.name());

            header.add(buf.toString());
            accessors.add(Pair.of(channelName, new AggregationStringifyer(aggregation, EMPTY_VALUE)));
         }
      }

      if (extrema != null && queryAnalyzer.isAggregationEnabled()) {
         for (final Extrema extremum : extrema) {
            for (final EventField field : eventFields) {
               final Function<Object, Object> accessor = extremum.getAccessor(field);
               if (accessor != null) {
                  final StringBuilder buf = new StringBuilder(7)
                        .append(channelName.getName())
                        .append(DELIMITER_CHANNELNAME_FIELDNAME)
                        .append(FIELDNAME_EXTREMA)
                        .append(DELIMITER_CHANNELNAME_FIELDNAME)
                        .append(extremum.name())
                        .append(DELIMITER_CHANNELNAME_FIELDNAME)
                        .append(field.name());

                  header.add(buf.toString());
                  accessors
                        .add(Pair.of(
                              channelName,
                              new EventFieldStringifyer(
                                    accessor,
                                    EMPTY_VALUE,
                                    DELIMITER_ARRAY,
                                    null)));
               }
            }
         }
      }
   }
}
