package ch.psi.daq.queryrest.response.formatter;

import java.io.OutputStream;
import java.util.LinkedHashSet;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.stream.Stream;

import org.apache.commons.lang3.tuple.Triple;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;

import com.fasterxml.jackson.core.JsonEncoding;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ObjectWriter;

import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQConfigQueryElement;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.operation.ConfigField;
import ch.psi.daq.domain.query.operation.EventField;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.queryrest.response.AbstractHTTPResponse;
import ch.psi.daq.queryrest.response.ResponseFormatter;

public class DAQConfigQueryResponseFormatter
      implements ResponseFormatter<Entry<DAQConfigQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>>,
      ApplicationContextAware {
   private static final Logger LOGGER = LoggerFactory.getLogger(DAQConfigQueryResponseFormatter.class);

   public static final String CONFIGS_RESP_FIELD = "configs";

   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {}

   @Override
   public void format(
         final JsonFactory factory,
         final ObjectMapper mapper,
         final Entry<DAQConfigQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>> result,
         final OutputStream out,
         final AbstractHTTPResponse response) throws Exception {
      final AtomicReference<Exception> exception = new AtomicReference<>();
      final JsonGenerator generator = factory.createGenerator(out, JsonEncoding.UTF8);

      final DAQConfigQueryElement daqQuery = result.getKey();
      final Set<String> includedFields = getConfigFields(daqQuery.getConfigFields(), true);
      final ObjectWriter writer = DAQQueriesResponseFormatter.configureWriter(mapper, null, includedFields);

      try {
         writeArrayFormat(generator, writer, result, exception);
      } finally {
         generator.flush();
         generator.close();
      }

      if (exception.get() != null) {
         throw exception.get();
      }
   }

   private void writeArrayFormat(final JsonGenerator generator, final ObjectWriter writer,
         final Entry<DAQConfigQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>> entryy,
         final AtomicReference<Exception> exception) {
      final DAQConfigQueryElement daqQuery = entryy.getKey();

      try {
         generator.writeStartArray();

         entryy.getValue()
               /* ensure elements are sequentially written */
               .sequential()
               .forEach(
                     triple -> {
                        try {
                           generator.writeStartObject();
                           generator.writeFieldName(EventField.channel.name());
                           writer.writeValue(generator, triple.getMiddle());
                           generator.writeFieldName(CONFIGS_RESP_FIELD);
                           writer.writeValue(generator, triple.getRight());
                           generator.writeEndObject();
                        } catch (Exception e) {
                           LOGGER.error("Could not write channel name of channel configuration '{}'",
                                 triple.getMiddle(),
                                 e);
                           exception.compareAndSet(null, e);
                        } finally {
                           if (triple.getRight() instanceof Stream) {
                              ((Stream<?>) (triple.getRight())).close();
                           }
                        }
                     });

         generator.writeEndArray();
      } catch (Exception e) {
         LOGGER.error("Exception while writing json for '{}'", daqQuery.getChannels(), e);
         exception.compareAndSet(null, e);
      }
   }

   public static Set<String> getConfigFields(final Set<? extends QueryField> queryFields,
         final boolean removeIdentifiers) {
      final Set<String> includedFields =
            new LinkedHashSet<String>(queryFields.size());

      for (final QueryField field : queryFields) {
         includedFields.add(field.getName());
      }

      if (removeIdentifiers) {
         // do not write channel/backend since it is already provided as key in mapping
         includedFields.remove(ConfigField.channel.name());
         includedFields.remove(ConfigField.name.name());
         includedFields.remove(ConfigField.backend.name());
      }

      return includedFields;
   }
}
