package ch.psi.daq.queryrest.response.formatter;

import java.io.OutputStream;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.function.ToLongFunction;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;

import com.fasterxml.jackson.core.JsonEncoding;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ObjectWriter;
import com.fasterxml.jackson.databind.ser.impl.SimpleBeanPropertyFilter;
import com.fasterxml.jackson.databind.ser.impl.SimpleFilterProvider;

import ch.psi.daq.common.stream.match.ListFiller;
import ch.psi.daq.common.stream.match.Padder;
import ch.psi.daq.common.stream.match.ReuseingListCreator;
import ch.psi.daq.common.stream.match.StreamMatcher;
import ch.psi.daq.common.tuple.Quadruple;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.json.mixin.HistoricChannelConfigurationPropertyFilterMixin;
import ch.psi.daq.domain.json.mixin.PropertyFilterMixin;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.bin.BinningStrategy;
import ch.psi.daq.domain.query.mapping.Alignment;
import ch.psi.daq.domain.query.mapping.IncompleteStrategy;
import ch.psi.daq.domain.query.mapping.Mapping;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.EventField;
import ch.psi.daq.domain.query.operation.Extrema;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.domain.request.range.RequestRange;
import ch.psi.daq.query.bin.aggregate.BinnedValueCombiner;
import ch.psi.daq.queryrest.config.QueryRestConfig;
import ch.psi.daq.queryrest.response.AbstractHTTPResponse;
import ch.psi.daq.queryrest.response.ResponseFormatter;
import ch.psi.daq.queryrest.response.json.JSONResponseStreamWriter;

public class DAQQueriesResponseFormatter
      implements ResponseFormatter<List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>>>,
      ApplicationContextAware {
   private static final Logger LOGGER = LoggerFactory.getLogger(JSONResponseStreamWriter.class);

   public static final String DATA_RESP_FIELD = "data";
   public static final String META_RESP_FIELD = "meta";

   public static final Mapping DEFAULT_MAPPING = new Mapping(IncompleteStrategy.PROVIDE_AS_IS);
   public static final Function<DataEvent, ChannelName> KEY_PROVIDER = (event) -> new ChannelName(event.getChannel(),
         event.getBackend());

   // In case ArchiverAppliance had several events within the 10ms mapping interval, return these
   // aggregations (only used for table format)
   private Set<String> defaultEventResponseAggregations;

   @SuppressWarnings("unchecked")
   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      final Set<Aggregation> defaultEventResponseAggregations =
            context.getBean(QueryRestConfig.BEAN_NAME_DEFAULT_EVENT_RESPONSE_AGGREGATIONS, Set.class);
      this.defaultEventResponseAggregations =
            defaultEventResponseAggregations.stream().map(Aggregation::name)
                  .collect(Collectors.toCollection(LinkedHashSet::new));
   }

   @Override
   public void format(
         final JsonFactory factory,
         final ObjectMapper mapper,
         final List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>> results,
         final OutputStream out,
         final AbstractHTTPResponse response) throws Exception {
      final AtomicReference<Exception> exception = new AtomicReference<>();
      final JsonGenerator generator = factory.createGenerator(out, JsonEncoding.UTF8);

      try {
         if (results.size() > 1) {
            generator.writeStartArray();
         }

         results
               .forEach(entryy -> {
                  final DAQQueryElement daqQuery = entryy.getKey();

                  if (response.useTableFormat(daqQuery)) {
                     final Set<String> includedConfigFields =
                           DAQConfigQueryResponseFormatter.getConfigFields(daqQuery.getConfigFields(), true);

                     final Set<String> includedFields = getEventFields(daqQuery, false);
                     /* make sure identifiers are available */
                     includedFields.add(EventField.channel.name());
                     includedFields.add(EventField.backend.name());
                     // issue ATEST-633
                     if (!containsAggregation(includedFields)) {
                        includedFields.addAll(defaultEventResponseAggregations);
                     }

                     final ObjectWriter writer = configureWriter(mapper, includedFields, includedConfigFields);

                     writeTableFormat(generator, writer, entryy, exception);
                  } else {
                     final Set<String> includedConfigFields =
                           DAQConfigQueryResponseFormatter.getConfigFields(daqQuery.getConfigFields(), true);
                     final Set<String> includedFields = getEventFields(daqQuery, true);
                     final ObjectWriter writer = configureWriter(mapper, includedFields, includedConfigFields);

                     writeArrayFormat(generator, writer, entryy, exception);
                  }
               });
      } finally {
         if (results.size() > 1) {
            generator.writeEndArray();
         }

         generator.flush();
         generator.close();
      }

      if (exception.get() != null) {
         throw exception.get();
      }
   }

   private static void writeArrayFormat(final JsonGenerator generator, final ObjectWriter writer,
         final Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>> entryy,
         final AtomicReference<Exception> exception) {
      final DAQQueryElement daqQuery = entryy.getKey();

      try {
         generator.writeStartArray();

         entryy.getValue()
               /* ensure elements are sequentially written */
               .sequential()
               .forEach(
                     quadruple -> {
                        try {
                           generator.writeStartObject();
                           generator.writeFieldName(EventField.channel.name());
                           writer.writeValue(generator, quadruple.getSecond());
                           if (daqQuery.hasConfigFields()) {
                              generator.writeFieldName(DAQConfigQueryResponseFormatter.CONFIGS_RESP_FIELD);
                              writer.writeValue(generator, quadruple.getThird());
                           }
                           generator.writeFieldName(DATA_RESP_FIELD);
                           writer.writeValue(generator, quadruple.getFourth());
                           generator.writeEndObject();
                        } catch (Exception e) {
                           LOGGER.error("Could not write channel name of channel '{}'", quadruple.getSecond(),
                                 e);
                           exception.compareAndSet(null, e);
                        } finally {
                           if (quadruple.getThird() instanceof Stream) {
                              ((Stream<?>) (quadruple.getThird())).close();
                           }
                           if (quadruple.getFourth() instanceof Stream) {
                              ((Stream<?>) (quadruple.getFourth())).close();
                           }
                        }
                     });

         generator.writeEndArray();
      } catch (Exception e) {
         LOGGER.error("Exception while writing json for '{}'", daqQuery.getChannels(), e);
         exception.compareAndSet(null, e);
      }
   }

   @SuppressWarnings("unchecked")
   private static void writeTableFormat(JsonGenerator generator, ObjectWriter writer,
         Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>> entryy,
         AtomicReference<Exception> exception) {
      final Map<ChannelName, Stream<ChannelConfiguration>> configStreams = new LinkedHashMap<>();
      /* get DataEvent stream of sub-queries for later match */
      final Map<ChannelName, Stream<DataEvent>> eventStreams = new LinkedHashMap<>();
      final AtomicReference<BackendQuery> backendQueryRef = new AtomicReference<>();
      final DAQQueryElement daqQuery = entryy.getKey();

      entryy.getValue()
            .sequential()
            .forEach(
                  quadruple -> {
                     backendQueryRef.compareAndSet(null, quadruple.getFirst());

                     if (entryy.getKey().hasConfigFields() && quadruple.getThird() instanceof Stream) {
                        configStreams.put(quadruple.getSecond(), ((Stream<ChannelConfiguration>) quadruple.getThird()));
                     }
                     if (quadruple.getFourth() instanceof Stream) {
                        eventStreams.put(quadruple.getSecond(), ((Stream<DataEvent>) quadruple.getFourth()));
                     } else {
                        final String message =
                              String.format("Expect a DataEvent Stream for '%s' but got '%s'.",
                                    quadruple.getSecond(), quadruple.getFourth().getClass().getSimpleName());
                        LOGGER.warn(message);
                        eventStreams.put(quadruple.getSecond(), Stream.empty());
                     }
                  });

      final BackendQuery backendQuery = backendQueryRef.get();
      final RequestRange requestRange = backendQuery.getRequest().getRequestRange();
      BinningStrategy binningStrategy = backendQuery.getBinningStrategy();

      final Mapping mapping = daqQuery.getMappingOrDefault(DEFAULT_MAPPING);
      final Padder<ChannelName, DataEvent> padder = mapping.getIncomplete().getPadder(backendQuery);

      ToLongFunction<DataEvent> alignmentProvider =
            Alignment.getAlignmentProvider(daqQuery.getMapping().getAlignment(), binningStrategy);
      if (binningStrategy == null) {
         binningStrategy = Alignment.getBinningStrategy(daqQuery.getMapping().getAlignment(), requestRange);
      }
      BiFunction<DataEvent, DataEvent, DataEvent> valueCombiner = null;
      if (binningStrategy != null) {
         binningStrategy.setRequestRange(requestRange);
         valueCombiner = new BinnedValueCombiner(binningStrategy);
      }

      /* online matching of the stream's content */
      final StreamMatcher<ChannelName, DataEvent, List<DataEvent>> streamMatcher =
            new StreamMatcher<>(
                  DAQQueriesResponseFormatter.KEY_PROVIDER,
                  alignmentProvider,
                  new ReuseingListCreator<ChannelName, DataEvent>(),
                  new ListFiller<ChannelName, DataEvent>(),
                  valueCombiner,
                  padder,
                  eventStreams.keySet(),
                  eventStreams.values());
      final Iterator<List<DataEvent>> streamsMatchIter = streamMatcher.iterator();

      try {
         generator.writeStartObject();
         // configs if available (use same format as for array form)
         if (!configStreams.isEmpty()) {
            generator.writeFieldName(META_RESP_FIELD);
            generator.writeStartArray();

            configStreams.entrySet()
                  /* ensure elements are sequentially written */
                  .stream()
                  .forEach(
                        entry -> {
                           try {
                              generator.writeStartObject();
                              generator.writeFieldName(EventField.channel.name());
                              writer.writeValue(generator, entry.getKey());

                              generator.writeFieldName(DAQConfigQueryResponseFormatter.CONFIGS_RESP_FIELD);
                              writer.writeValue(generator, entry.getValue());

                              generator.writeEndObject();
                           } catch (Exception e) {
                              LOGGER.error("Could not write channel name of channel '{}'", entry.getKey(),
                                    e);
                              exception.compareAndSet(null, e);
                           } finally {
                              entry.getValue().close();
                           }
                        });

            generator.writeEndArray();
         }

         // write event table
         generator.writeFieldName(DATA_RESP_FIELD);
         writer.writeValue(generator, streamsMatchIter);
         generator.writeEndObject();
      } catch (Exception e) {
         LOGGER.error("Exception while writing json for '{}'", daqQuery.getChannels(), e);
         exception.compareAndSet(null, e);
      } finally {
         if (streamMatcher != null) {
            try {
               streamMatcher.close();
            } catch (Throwable t) {
               LOGGER.error(
                     "Something went wrong while closing stream matcher for JSON table response writer.",
                     t);
            }
         }
      }
   }

   /**
    * Configures the writer dynamically by including the fields which should be included in the
    * response.
    * 
    * @param mapper The ObjectMapper
    * @param includedEventFields set of strings which correspond to the getter method names of the
    *        classes registered as a mixed-in
    * @param includedConfigFields set of strings which correspond to the getter method names of the
    *        classes registered as a mixed-in
    * @return the configured writer that includes the specified fields
    */
   public static ObjectWriter configureWriter(final ObjectMapper mapper, final Set<String> includedEventFields,
         final Set<String> includedConfigFields) {
      final SimpleFilterProvider propertyFilter = new SimpleFilterProvider();
      if (includedEventFields != null) {
         propertyFilter.addFilter(PropertyFilterMixin.FILTER_NAME,
               SimpleBeanPropertyFilter.filterOutAllExcept(includedEventFields));
      }
      if (includedConfigFields != null) {
         propertyFilter.addFilter(HistoricChannelConfigurationPropertyFilterMixin.FILTER_NAME,
               SimpleBeanPropertyFilter.filterOutAllExcept(includedConfigFields));
      }
      // only write the properties not excluded in the filter
      final ObjectWriter writer = mapper.writer(propertyFilter);
      return writer;
   }

   private static Set<String> getEventFields(final DAQQueryElement query, final boolean removeIdentifiers) {
      final Set<? extends QueryField> eventFields = query.getEventFields();
      List<Aggregation> aggregations =
            query.getAggregation() != null ? query.getAggregation().getAggregations() : null;
      if (aggregations == null && query.getMapping() != null) {
         // global aggregation has precedence
         aggregations = query.getMapping().getAggregations();
      }
      final List<Extrema> extrema = query.getAggregation() != null ? query.getAggregation().getExtrema() : null;

      final Set<String> includedFields =
            new LinkedHashSet<String>(eventFields.size() + (aggregations != null ? aggregations.size() : 0)
                  + (extrema != null ? extrema.size() : 0));

      for (final QueryField field : eventFields) {
         includedFields.add(field.getName());
      }
      if (aggregations != null) {
         for (final Aggregation aggregation : aggregations) {
            includedFields.add(aggregation.name());
         }
      }
      if (extrema != null) {
         // field of ExtremaCalculator (extrema in BinnedValueCombinedDataEvent and
         // BinnedIndexCombinedDataEvent)
         includedFields.add("extrema");
      }

      if (removeIdentifiers) {
         // do not write channel/backend since it is already provided as key in mapping
         includedFields.remove(EventField.channel.name());
         includedFields.remove(EventField.backend.name());
      }

      return includedFields;
   }

   private static boolean containsAggregation(final Set<String> includedFields) {
      for (final Aggregation aggregation : Aggregation.values()) {
         if (includedFields.contains(aggregation.name())) {
            return true;
         }
      }
      return false;
   }
}
