package ch.psi.daq.queryrest.response.raw;

import java.io.BufferedOutputStream;
import java.io.OutputStream;
import java.nio.ByteOrder;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Function;
import java.util.function.ToLongFunction;
import java.util.stream.Stream;

import org.apache.commons.lang3.tuple.MutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.collect.Sets;

import ch.psi.bsread.compression.Compression;
import ch.psi.daq.common.stream.match.KeepAsIsPadder;
import ch.psi.daq.common.stream.match.KeepPreviousPadder;
import ch.psi.daq.common.stream.match.MapFiller;
import ch.psi.daq.common.stream.match.Padder;
import ch.psi.daq.common.stream.match.RecreateMapCreator;
import ch.psi.daq.common.stream.match.StreamMatcher;
import ch.psi.daq.common.time.TimeUtils;
import ch.psi.daq.common.tuple.Quadruple;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.StreamEvent;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.mapping.Mapping;
import ch.psi.daq.domain.query.mapping.Alignment;
import ch.psi.daq.queryrest.response.AbstractHTTPResponse;
import ch.psi.daq.queryrest.response.ResponseFormatter;
import ch.psi.daq.queryrest.response.csv.CSVResponseStreamWriter;
import ch.psi.daq.queryrest.response.formatter.DAQQueriesResponseFormatter;

public abstract class AbstractRawResponseFormatter implements ResponseFormatter<List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>>> {
   private static final Logger LOGGER = LoggerFactory.getLogger(AbstractRawResponseFormatter.class);

   public static final Function<ChannelConfiguration, ChannelName> CONFIG_KEY_PROVIDER =
         (config) -> new ChannelName(config.getChannel(),
               config.getBackend());
   // try to match sync data (bsread) with non sync data (epics) based on the time usin 10 millis
   // buckets.
   public static final ToLongFunction<ChannelConfiguration> CONFIG_MATCHER_PROVIDER =
         (config) -> config.getGlobalMillis()
               / TimeUtils.MILLIS_PER_PULSE;
         
   @SuppressWarnings("unchecked")
   public void format(
         final JsonFactory factory, 
         final ObjectMapper mapper,
         final List<Entry<DAQQueryElement, Stream<Quadruple<BackendQuery, ChannelName, ?, ?>>>> results,
         final OutputStream out,
         final AbstractHTTPResponse response) throws Exception {
      final AtomicReference<Exception> exception = new AtomicReference<>();

      try (final BufferedOutputStream os = new BufferedOutputStream(out, 64 * 1024)) {

         results
               .forEach(entryy -> {
                  final DAQQueryElement daqQuery = entryy.getKey();
                  final AtomicReference<BackendQuery> backendQueryRef = new AtomicReference<>();

                  final Map<ChannelName, Stream<ChannelConfiguration>> configStreams =
                        new LinkedHashMap<>(results.size());
                  final Map<ChannelName, Stream<DataEvent>> eventStreams = new LinkedHashMap<>(results.size());

                  entryy.getValue()
                        .forEach(quadruple -> {
                           // single BackendQuery instance for all
                           backendQueryRef.compareAndSet(null, quadruple.getFirst());

                           configStreams.put(quadruple.getSecond(),
                                 (Stream<ChannelConfiguration>) quadruple.getThird());
                           eventStreams.put(quadruple.getSecond(),
                                 (Stream<DataEvent>) quadruple.getFourth());
                        });


                  if (exception.get() == null) {
                     try {
                        if (response.useTableFormat(daqQuery)) {
                           final Mapping mapping =
                                 daqQuery.getMappingOrDefault(CSVResponseStreamWriter.DEFAULT_MAPPING);
                           final Padder<ChannelName, DataEvent> padder =
                                 mapping.getIncomplete().getPadder(backendQueryRef.get());

                           writeTableFormat(
                                 os,
                                 configStreams,
                                 eventStreams,
                                 padder);
                        } else {
                           writeArrayFormat(
                                 os,
                                 configStreams,
                                 eventStreams);
                        }
                     } catch (Exception e) {
                        final String message = String.format("Could not write data for '{}'.", daqQuery);
                        LOGGER.warn(message, e);
                        exception.compareAndSet(null, new IllegalStateException(message, e));
                     }
                  }

                  for (Stream<ChannelConfiguration> stream : configStreams.values()) {
                     stream.close();
                  }
                  for (Stream<DataEvent> stream : eventStreams.values()) {
                     stream.close();
                  }
               });
      }

      if (exception.get() != null) {
         throw exception.get();
      }
   }

   @SuppressWarnings("unchecked")
   protected void writeArrayFormat(
         final OutputStream os,
         final Map<ChannelName, Stream<ChannelConfiguration>> configStreams,
         final Map<ChannelName, Stream<DataEvent>> eventStreams) throws Exception {

      // padder do not matter as we are streaming channel by channel and do no inter Channel mapping
      final Padder<ChannelName, ChannelConfiguration> configPadder = new KeepAsIsPadder<>();
      final Padder<ChannelName, DataEvent> eventPadder = new KeepAsIsPadder<>();

      for (final Entry<ChannelName, Stream<DataEvent>> entry : eventStreams.entrySet()) {
         final Stream<ChannelConfiguration> configStream = configStreams.get(entry.getKey());
         final Set<ChannelName> keySet = Sets.newHashSet(entry.getKey());
         final Collection<Stream<ChannelConfiguration>> configValues = Sets.newHashSet(configStream);
         final Collection<Stream<DataEvent>> eventValues = Sets.newHashSet(entry.getValue());

         // online matching of the stream's content
         final StreamMatcher<ChannelName, ChannelConfiguration, Map<ChannelName, ChannelConfiguration>> configStreamMatcher =
               new StreamMatcher<>(
                     CONFIG_KEY_PROVIDER,
                     CONFIG_MATCHER_PROVIDER,
                     new RecreateMapCreator<>(),
                     new MapFiller<>(),
                     null,
                     configPadder,
                     keySet,
                     configValues);
         final Iterator<Map<ChannelName, ChannelConfiguration>> configStreamsMatchIter = configStreamMatcher.iterator();

         final StreamMatcher<ChannelName, DataEvent, Map<ChannelName, DataEvent>> eventStreamMatcher =
               new StreamMatcher<>(
                     DAQQueriesResponseFormatter.KEY_PROVIDER,
                     Alignment.MATCHER_PROVIDER_BY_PULSE,
                     new RecreateMapCreator<>(),
                     new MapFiller<>(),
                     null,
                     eventPadder,
                     keySet,
                     eventValues);
         final Iterator<Map<ChannelName, DataEvent>> eventStreamsMatchIter = eventStreamMatcher.iterator();

         try {
            write(os, configStreamsMatchIter, eventStreamsMatchIter);
         } finally {
            eventStreamMatcher.close();
            configStreamMatcher.close();
         }
      }
   }

   protected void writeTableFormat(
         final OutputStream os,
         final Map<ChannelName, Stream<ChannelConfiguration>> configStreams,
         final Map<ChannelName, Stream<DataEvent>> eventStreams,
         final Padder<ChannelName, DataEvent> eventPadder) {
      // INFO: Do not use matchProvider = binningStrategy; since this is a raw data downloader (no aggregations supported so far)
      
      // online matching of the stream's content
      final StreamMatcher<ChannelName, ChannelConfiguration, Map<ChannelName, ChannelConfiguration>> configStreamMatcher =
            new StreamMatcher<>(
                  CONFIG_KEY_PROVIDER,
                  CONFIG_MATCHER_PROVIDER,
                  new RecreateMapCreator<>(),
                  new MapFiller<>(),
                  null,
                  new KeepPreviousPadder<>(),
                  configStreams.keySet(),
                  configStreams.values());
      final Iterator<Map<ChannelName, ChannelConfiguration>> configStreamsMatchIter = configStreamMatcher.iterator();

      final StreamMatcher<ChannelName, DataEvent, Map<ChannelName, DataEvent>> eventStreamMatcher =
            new StreamMatcher<>(
                  DAQQueriesResponseFormatter.KEY_PROVIDER,
                  Alignment.MATCHER_PROVIDER_BY_PULSE,
                  new RecreateMapCreator<>(),
                  new MapFiller<>(),
                  null,
                  eventPadder,
                  eventStreams.keySet(),
                  eventStreams.values());
      final Iterator<Map<ChannelName, DataEvent>> eventStreamsMatchIter = eventStreamMatcher.iterator();

      try {
         write(os, configStreamsMatchIter, eventStreamsMatchIter);
      } finally {
         eventStreamMatcher.close();
         configStreamMatcher.close();
      }
   }

   protected void write(
         final OutputStream os,
         final Iterator<Map<ChannelName, ChannelConfiguration>> configStreamsMatchIter,
         final Iterator<Map<ChannelName, DataEvent>> eventStreamsMatchIter) {

      final Iterator<Pair<Map<ChannelName, ChannelConfiguration>, Map<ChannelName, DataEvent>>> iter =
            new ConfigEventIterator(configStreamsMatchIter, eventStreamsMatchIter);
      final Map<ChannelName, DataEvent> latestEvents = new HashMap<>();
      final ConfigChangeDetector coonfigChangeDetector = new ConfigChangeDetector();

      while (iter.hasNext()) {
         final Pair<Map<ChannelName, ChannelConfiguration>, Map<ChannelName, DataEvent>> currentPair = iter.next();
         latestEvents.putAll(currentPair.getRight());

         // check if changed
         if (coonfigChangeDetector.hasConfigChanged(currentPair)) {
            writeConfigs(os, latestEvents, currentPair);
         }

         writeEvents(os, currentPair);
      }
   }

   protected abstract void writeConfigs(final OutputStream os, final Map<ChannelName, DataEvent> latestEvents,
         final Pair<Map<ChannelName, ChannelConfiguration>, Map<ChannelName, DataEvent>> pair);

   protected abstract void writeEvents(final OutputStream os,
         final Pair<Map<ChannelName, ChannelConfiguration>, Map<ChannelName, DataEvent>> pair);

   protected Compression getCompression(final DataEvent event) {
      if (event instanceof StreamEvent) {
         return ((StreamEvent) event).getCompression().getCompression();
      } else {
         if (event != null && event.isArray()) {
            return Compression.bitshuffle_lz4;
         } else {
            return Compression.none;
         }
      }
   }

   protected ByteOrder getEncoding(final DataEvent event) {
      if (event instanceof StreamEvent) {
         return ((StreamEvent) event).getValueByteOrder();
      } else {
         // use native as it is most likely fastest
         return ByteOrder.nativeOrder();
      }
   }

   public class ConfigEventIterator
         implements Iterator<Pair<Map<ChannelName, ChannelConfiguration>, Map<ChannelName, DataEvent>>> {
      private final Iterator<Map<ChannelName, ChannelConfiguration>> configIter;
      private Map<ChannelName, ChannelConfiguration> nextNextConfigMap;
      private Entry<ChannelName, ChannelConfiguration> nextNextConfig;
      private final Iterator<Map<ChannelName, DataEvent>> eventIter;
      private MutablePair<Map<ChannelName, ChannelConfiguration>, Map<ChannelName, DataEvent>> next;

      public ConfigEventIterator(
            final Iterator<Map<ChannelName, ChannelConfiguration>> configIter,
            final Iterator<Map<ChannelName, DataEvent>> eventIter) {
         this.configIter = configIter;
         this.eventIter = eventIter;

         if (configIter.hasNext() && eventIter.hasNext()) {
            next = MutablePair.of(configIter.next(), eventIter.next());
         }

         extractNextNextConfig();
      }

      @Override
      public boolean hasNext() {
         return next != null;
      }

      @Override
      public Pair<Map<ChannelName, ChannelConfiguration>, Map<ChannelName, DataEvent>> next() {
         final Pair<Map<ChannelName, ChannelConfiguration>, Map<ChannelName, DataEvent>> next = this.next;

         extractNext();
         return next;
      }

      private void extractNext() {
         if (eventIter.hasNext()) {
            next = MutablePair.of(next.getKey(), eventIter.next());

            if (nextNextConfig != null) {
               final DataEvent event = next.getValue().get(nextNextConfig.getKey());
               if (event != null
                     && TimeUtils.isAfterEquals(event.getGlobalTime(), nextNextConfig.getValue().getGlobalTime())) {
                  next.setLeft(nextNextConfigMap);
                  extractNextNextConfig();
               }
            }
         } else {
            next = null;
         }
      }

      private void extractNextNextConfig() {
         if (configIter.hasNext()) {
            nextNextConfigMap = configIter.next();
            ChannelConfiguration tmp = nextNextConfigMap.values()
                  .stream()
                  .sorted(ChannelConfiguration.COMPARATOR.reversed())
                  .findFirst()
                  .orElse(null);
            if (tmp != null) {
               nextNextConfig = Pair.of(new ChannelName(tmp.getChannel(), tmp.getBackend()), tmp);
            } else {
               // should not happen
               nextNextConfigMap = null;
               nextNextConfig = null;
            }
         } else {
            nextNextConfigMap = null;
            nextNextConfig = null;
         }
      }
   }

   public class ConfigChangeDetector {
      private Map<ChannelName, ChannelConfiguration> currentConfigs;
      private final Map<ChannelName, Compression> latestCompressions = new HashMap<>();
      private final Map<ChannelName, ByteOrder> latestEncoding = new HashMap<>();

      public boolean hasConfigChanged(
            final Pair<Map<ChannelName, ChannelConfiguration>, Map<ChannelName, DataEvent>> pair) {
         boolean hasChanged = currentConfigs == null || currentConfigs != pair.getKey();
         Compression compression;
         ByteOrder encoding;

         for (final Entry<ChannelName, DataEvent> entry : pair.getValue().entrySet()) {
            compression = getCompression(entry.getValue());
            if (!compression.equals(latestCompressions.put(entry.getKey(), compression))) {
               hasChanged = true;
            }

            encoding = getEncoding(entry.getValue());
            if (!encoding.equals(latestEncoding.put(entry.getKey(), encoding))) {
               hasChanged = true;
            }
         }

         currentConfigs = pair.getKey();

         return hasChanged;
      }
   }
}
