package ch.psi.daq.queryrest.response.raw.event.serialize;

import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.Function;
import java.util.function.IntFunction;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.psi.bsread.common.allocator.ByteBufferAllocator;
import ch.psi.bsread.common.helper.ByteBufferHelper;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.json.BackendChannel;
import ch.psi.daq.queryrest.response.raw.Ids;
import ch.psi.daq.queryrest.response.raw.serialize.RawBlobDeserializer;
import ch.psi.daq.queryrest.response.raw.serialize.RawBlobSerializer;
import ch.psi.daq.queryrest.response.raw.serialize.RawDeserializer;
import ch.psi.daq.queryrest.response.raw.serialize.RawSerializer;

public class EventDeSerializer implements RawSerializer<EventHeader, Map<? extends BackendChannel, DataEvent>>,
      RawDeserializer<EventHeader, DataEvent> {
   private static final Logger LOGGER = LoggerFactory.getLogger(EventDeSerializer.class);

   private static final short ID = Ids.DATA_EVENTS;
   private static final ByteBuffer ID_BYTES;
   static {
      ID_BYTES = ByteBufferAllocator.DEFAULT_ALLOCATOR.apply(Short.BYTES)
            .order(RawSerializer.DEFAULT_ENCODING)
            .putShort(0, ID);
   }

   private final IntFunction<ByteBuffer> allocator;
   private EventHeader header;
   private Map<EventConfig, BackendChannel> cache = new HashMap<>();
   private Map<EventConfig, RawBlobSerializer<EventConfig, DataEvent>> blobSerializers = new HashMap<>();

   private ConfigValuesHolder<EventHeader, DataEvent> valueHolder;
   private Function<EventConfig, RawBlobDeserializer<EventConfig, DataEvent>> blobDeserializerCreator;
   private Map<EventConfig, RawBlobDeserializer<EventConfig, DataEvent>> blobDeserializers = new HashMap<>();

   public EventDeSerializer(final IntFunction<ByteBuffer> allocator) {
      this.allocator = allocator;
   }

   public EventDeSerializer(
         final IntFunction<ByteBuffer> allocator,
         final Function<EventConfig, RawBlobDeserializer<EventConfig, DataEvent>> blobDeserializerCreator) {
      this.allocator = allocator;
      this.blobDeserializerCreator = blobDeserializerCreator;
   }

   @Override
   public short getId() {
      return ID;
   }

   @Override
   public ByteBuffer getIdBytes() {
      return ID_BYTES.duplicate().order(ID_BYTES.order());
   }

   @Override
   public void configure(final EventHeader header,
         final Map<? extends BackendChannel, ChannelConfiguration> channelConfigs) {
      this.header = header;
      cache.clear();
      header.addEventSerializerId(getId());

      final Map<BackendChannel, EventConfig> eventConfigs = header.getChannelsMapping();
      ChannelConfiguration channelConfig;
      EventConfig eventConfig;
      RawBlobSerializer<EventConfig, DataEvent> serializer;

      for (final Entry<BackendChannel, EventConfig> entry : eventConfigs.entrySet()) {
         eventConfig = entry.getValue();

         channelConfig = channelConfigs.get(entry.getKey());
         serializer = new StandardEventDeSerializer(allocator);
         serializer.configure(eventConfig, channelConfig);
         
         // add at end since configure can modify eventConfig
         cache.put(eventConfig, entry.getKey());
         blobSerializers.put(eventConfig, serializer);
      }
   }

   @Override
   public void serialize(OutputStream os, Map<? extends BackendChannel, DataEvent> events) {
      try {
         final Collection<EventConfig> channels = header.getChannels();

         final List<ByteBuffer> buffers = new ArrayList<>();
         final ByteBuffer idBuf = getIdBytes();
         buffers.add(idBuf);

         BackendChannel channelName;
         DataEvent event;

         for (final EventConfig channel : channels) {
            channelName = cache.get(channel);
            event = events.get(channelName);

            final RawBlobSerializer<EventConfig, DataEvent> serializer = blobSerializers.get(channel);
            serializer.serialize(event, buffers);
         }

         long size = 0;
         for (final ByteBuffer buffer : buffers) {
            size += buffer.remaining();
         }
         final ByteBuffer sizeBuf = allocator
               .apply(Long.BYTES)
               .order(RawSerializer.DEFAULT_ENCODING)
               .putLong(0, size);
         // size (long)
         ByteBufferHelper.writeByteBuffer(sizeBuf, os);

         for (final ByteBuffer buffer : buffers) {
            ByteBufferHelper.writeByteBuffer(buffer, os);
         }
      } catch (Exception e) {
         final String message = String.format("Could not write Events '%s'.", events.keySet());
         LOGGER.warn(message, e);
         throw new IllegalStateException(message, e);
      }
   }

   @Override
   public void init(final ConfigValuesHolder<EventHeader, DataEvent> valueHolder) {
      this.header = valueHolder.getConfig();
      this.valueHolder = valueHolder;
      cache.clear();

      final Map<BackendChannel, EventConfig> eventConfigs = header.getChannelsMapping();
      EventConfig eventConfig;

      for (final Entry<BackendChannel, EventConfig> entry : eventConfigs.entrySet()) {
         eventConfig = entry.getValue();
         cache.put(eventConfig, entry.getKey());

         final RawBlobDeserializer<EventConfig, DataEvent> blobDeserializer =
               blobDeserializerCreator.apply(eventConfig);
         blobDeserializer.init(eventConfig);
         blobDeserializers.put(eventConfig, blobDeserializer);
      }
   }

   @Override
   public void deserialize(ByteBuffer buffer) {
      try {
         // skip id
         buffer.position(buffer.position() + Short.BYTES);

         final Collection<EventConfig> channels = header.getChannels();
         BackendChannel channelName;
         RawBlobDeserializer<EventConfig, DataEvent> blobDeserializer;

         for (final EventConfig channel : channels) {
            blobDeserializer = blobDeserializers.get(channel);

            final DataEvent event = blobDeserializer.deserialize(buffer);
            if (event != null) {
               channelName = cache.get(channel);
               valueHolder.add(channelName, event);
            }
         }
      } catch (Exception e) {
         final String message = "Could not read DataHeader.";
         LOGGER.warn(message, e);
         throw new IllegalStateException(message, e);
      }
   }
}
