package ch.psi.daq.queryrest.response.raw.event.serialize;

import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.util.Map;
import java.util.function.IntFunction;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.hash.Hasher;
import com.google.common.hash.Hashing;

import ch.psi.bsread.common.allocator.ByteBufferAllocator;
import ch.psi.bsread.common.helper.ByteBufferHelper;
import ch.psi.bsread.compression.Compression;
import ch.psi.daq.common.io.ByteBufferInputStream;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.json.BackendChannel;
import ch.psi.daq.queryrest.response.raw.receive.ReceiverState;
import ch.psi.daq.queryrest.response.raw.serialize.RawDeserializer;
import ch.psi.daq.queryrest.response.raw.serialize.RawSerializer;

public class EventHeaderDeSerializer implements RawSerializer<Void, EventHeader>, RawDeserializer<Void, EventHeader> {
   private static final Logger LOGGER = LoggerFactory.getLogger(EventHeaderDeSerializer.class);

   private static final short ID = EventHeader.ID;
   private static final ByteBuffer ID_BYTES;
   static {
      ID_BYTES = ByteBufferAllocator.DEFAULT_ALLOCATOR.apply(Short.BYTES)
            .order(RawSerializer.DEFAULT_ENCODING)
            .putShort(0, ID);
   }

   private final IntFunction<ByteBuffer> allocator;
   private final ObjectMapper mapper;
   private final ReceiverState<EventHeader, ?> state;

   public EventHeaderDeSerializer(final IntFunction<ByteBuffer> allocator, final ObjectMapper mapper) {
      this(allocator, mapper, null);
   }

   public EventHeaderDeSerializer(final IntFunction<ByteBuffer> allocator, final ObjectMapper mapper, final ReceiverState<EventHeader, ?> state) {
      this.allocator = allocator;
      this.mapper = mapper;
      this.state = state;
   }

   @Override
   public short getId() {
      return ID;
   }

   @Override
   public ByteBuffer getIdBytes() {
      return ID_BYTES.duplicate().order(ID_BYTES.order());
   }

   @Override
   public void configure(Void config, Map<? extends BackendChannel, ChannelConfiguration> channelConfigs) {
      // noop
   }

   @Override
   public void serialize(OutputStream os, EventHeader value) {
      try {
         final Compression compression = Compression.bitshuffle_lz4;
         final byte[] eventHeaderByteArray = mapper.writeValueAsBytes(value);
         final ByteBuffer eventHeaderBytes = compression.getCompressor()
               .compressDataHeader(
                     ByteBuffer.wrap(eventHeaderByteArray),
                     allocator);
         final Hasher eventHeaderHash = Hashing.murmur3_128().newHasher(eventHeaderBytes.remaining());
         // dataHeaderHash.putBytes(dataHeaderBytes.duplicate());
         eventHeaderHash.putBytes(eventHeaderByteArray);

         final ByteBuffer idBuf = getIdBytes();
         final ByteBuffer hashCompressBuf = allocator
               .apply(Long.BYTES + Byte.BYTES)
               .order(RawSerializer.DEFAULT_ENCODING)
               .putLong(0, eventHeaderHash.hash().asLong())
               .put(Long.BYTES, compression.getId());
         final ByteBuffer sizeBuf = allocator
               .apply(Long.BYTES)
               .order(RawSerializer.DEFAULT_ENCODING)
               .putLong(0, idBuf.remaining() + hashCompressBuf.remaining() + eventHeaderBytes.remaining());

         // size (long)
         ByteBufferHelper.writeByteBuffer(sizeBuf, os);
         // id (short)
         ByteBufferHelper.writeByteBuffer(idBuf, os);
         // hash (long)
         ByteBufferHelper.writeByteBuffer(hashCompressBuf, os);
         // DataHeader (bitshuffle_lz4 compressed json string)
         ByteBufferHelper.writeByteBuffer(eventHeaderBytes, os);
      } catch (Exception e) {
         final String message = String.format("Could not write DataHeader '%s'.", value);
         LOGGER.warn(message, e);
         throw new IllegalStateException(message, e);
      }
   }

   @Override
   public void init(final ConfigValuesHolder<Void, EventHeader> config) {}

   @Override
   public void deserialize(ByteBuffer buffer) {
      try {
         // skip id
         buffer.position(buffer.position() + Short.BYTES);

         final long hash = buffer.getLong();
         if (state.getHash() != hash) {
            final Compression compression = Compression.byId(buffer.get());
            final ByteBuffer eventHeaderBytes = compression.getCompressor()
                  .decompressDataHeader(buffer, allocator);
            final EventHeader eventHeader =
                  mapper.readValue(new ByteBufferInputStream(eventHeaderBytes), EventHeader.class);

            state.setHash(hash);
            state.setConfig(eventHeader);
         }
      } catch (Exception e) {
         final String message = "Could not read EventHeader.";
         LOGGER.warn(message, e);
         throw new IllegalStateException(message, e);
      }
   }
}
