package ch.psi.daq.queryrest.response.raw.event.serialize;

import java.nio.ByteBuffer;
import java.util.Collection;
import java.util.function.IntFunction;

import ch.psi.daq.common.time.Time;
import ch.psi.daq.domain.ByteValue;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.DataHelper;
import ch.psi.daq.domain.StreamEvent;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.compression.CompressionAlgo;
import ch.psi.daq.domain.compression.CompressionRequest;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.events.ChannelEvent;
import ch.psi.daq.domain.events.Severity;
import ch.psi.daq.domain.events.Status;
import ch.psi.daq.domain.events.impl.ProcessingEvent;
import ch.psi.daq.queryrest.response.raw.Ids;
import ch.psi.daq.queryrest.response.raw.serialize.RawBlobDeserializer;
import ch.psi.daq.queryrest.response.raw.serialize.RawBlobSerializer;

public class StandardEventDeSerializer
      implements RawBlobSerializer<EventConfig, DataEvent>, RawBlobDeserializer<EventConfig, DataEvent> {
   public static final int ID = Ids.BLOB_DATA_EVENT_STANDARD;
   
   private static final int META_VALUE_BLOB_SIZE_LENGTH = Integer.BYTES;

   private static final int POS_META_BLOB_SIZE = 0;
   private static final int POS_META_IOC_TIME = POS_META_BLOB_SIZE + Integer.BYTES;
   private static final int POS_META_PULSE_ID = POS_META_IOC_TIME + Long.BYTES;
   private static final int POS_META_GLOBAL_TIME = POS_META_PULSE_ID + Long.BYTES;
   private static final int POS_META_STATUS = POS_META_GLOBAL_TIME + Long.BYTES;
   private static final int POS_META_SEVERITY = POS_META_STATUS + Byte.BYTES;
   private static final int POS_VALUE_BLOB = POS_META_SEVERITY + Byte.BYTES;
   private static final int META_BLOB_LENGTH = POS_VALUE_BLOB - POS_META_IOC_TIME;
   private static final int META_SIZE = META_VALUE_BLOB_SIZE_LENGTH + META_BLOB_LENGTH;

   private final IntFunction<ByteBuffer> allocator;
   private EventConfig config;
   private ByteBuffer noEventBytes;

   public StandardEventDeSerializer(final IntFunction<ByteBuffer> allocator) {
      this.allocator = allocator;
   }

   @Override
   public short getId() {
      return ID;
   }

   @Override
   public void configure(EventConfig config, ChannelConfiguration channelConfig) {
      this.config = config;

      config.setBlobId(getId());

      noEventBytes = allocator.apply(Integer.BYTES)
            .order(config.getByteOrder())
            .putInt(0, 0);
   }

   @Override
   public void serialize(final DataEvent event, final Collection<ByteBuffer> buffers) {
      if (event == null) {
         buffers.add(noEventBytes.duplicate().order(noEventBytes.order()));
      } else {
         Status status = Status.NO_ALARM;
         Severity severity = Severity.NO_ALARM;
         
         final ByteBuffer valueBuffer;
         if (event instanceof StreamEvent) {
            valueBuffer = ((StreamEvent) event).getValueBytes(CompressionRequest.any);
         } else if (event instanceof ChannelEvent) {
            valueBuffer = ((ChannelEvent) event).getValueBytes(CompressionRequest.any);

            status = ((ChannelEvent) event).getStatus();
            severity = ((ChannelEvent) event).getSeverity();
         } else {
            final ByteBuffer dbBytes = DataHelper.convertToDBBytes(
                  // allocator?
                  event.getValue(),
                  config.getType().getKey(),
                  config.getByteOrder(),
                  config.getShape(),
                  CompressionAlgo.getCompressionAlgo(config.getCompression().name()));
            valueBuffer = DataHelper.extractValueBytesAsIs(dbBytes);
         }

         final ByteBuffer metaBuffer = allocator
               .apply(META_SIZE)
               .order(config.getByteOrder());
         // meta value blob size
         metaBuffer.putInt(POS_META_BLOB_SIZE,
               metaBuffer.remaining() - META_VALUE_BLOB_SIZE_LENGTH
                     + valueBuffer.remaining());
         metaBuffer.putLong(POS_META_IOC_TIME, event.getIocTime().getUnscaled());
         metaBuffer.putLong(POS_META_PULSE_ID, event.getPulseId());
         metaBuffer.putLong(POS_META_GLOBAL_TIME, event.getGlobalTime().getUnscaled());
         metaBuffer.put(POS_META_STATUS, status.getDBKey());
         metaBuffer.put(POS_META_SEVERITY, severity.getDBKey());

         buffers.add(metaBuffer);
         buffers.add(valueBuffer);
      }
   }

   @Override
   public void init(final EventConfig config) {
      this.config = config;
   }

   @Override
   public DataEvent deserialize(ByteBuffer buffer) {
      int metaValueBlobSize;
      long iocTime;
      long pulseId;
      long globalTime;
      Status status;
      Severity severity;
      ByteBuffer byteValue;

      buffer.order(config.getByteOrder());
      // size
      metaValueBlobSize = buffer.getInt();
      if (metaValueBlobSize > 0) {
         iocTime = buffer.getLong();
         pulseId = buffer.getLong();
         globalTime = buffer.getLong();
         status = Status.newInstance(buffer.get());
         severity = Severity.newInstance(buffer.get());

         byteValue = buffer.duplicate().order(buffer.order());
         final int endPos = byteValue.position() + metaValueBlobSize - META_BLOB_LENGTH;
         byteValue.limit(endPos);
         ByteBuffer valueBytes =
               config
                     .getCompression()
                     .getCompressor()
                     .decompressData(byteValue, byteValue.position(), DataHelper.ALLOCATOR_DEFAULT,
                           config.getType().getBytes());
         buffer.position(endPos);

         ProcessingEvent event =
               new ProcessingEvent(
                     config.getName(),
                     Backend.byName(config.getBackend()),
                     new Time(iocTime),
                     pulseId,
                     new Time(globalTime),
                     config.getType().getKey(),
                     config.getShape(),
                     status,
                     severity,
                     new ByteValue(
                           config.getType().getKey(),
                           config.getShape(),
                           valueBytes));
         return event;
      } else {
         return null;
      }
   }
}
