package ch.psi.daq.queryrest.response.raw.receive;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.util.function.Function;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.psi.daq.queryrest.response.raw.RawDeserializerProvider;
import ch.psi.daq.queryrest.response.raw.event.serialize.EventHeader;
import ch.psi.daq.queryrest.response.raw.serialize.RawDeserializer;

public class InputStreamMessageExtractor<V> implements MessageReader<EventHeader, V> {
   private static final Logger LOGGER = LoggerFactory.getLogger(InputStreamMessageExtractor.class);

   private final InputStream is;
   private final Function<ReceiverState<EventHeader, V>, RawDeserializer<Void, EventHeader>> configDeserializerCreator;
   private final RawDeserializerProvider<EventHeader, V> deserializerProvider;
   private ReceiverState<EventHeader, V> state;
   private RawDeserializer<Void, EventHeader> configDeserializer;

   public InputStreamMessageExtractor(final InputStream is,
         final Function<ReceiverState<EventHeader, V>, RawDeserializer<Void, EventHeader>> configDeserializerCreator,
         final RawDeserializerProvider<EventHeader, V> deserializerProvider) {
      this.is = is;
      this.configDeserializerCreator = configDeserializerCreator;
      this.deserializerProvider = deserializerProvider;
   }

   @Override
   public void connect(final ReceiverState<EventHeader, V> state) {
      this.state = state;

      configDeserializer = configDeserializerCreator.apply(state);
      configDeserializer.init(null);
   }

   @Override
   public void readNext() {
      state.init();

      try {
         while (!state.hasValues() && state.isRunning()) {
            final ByteBuffer buffer = RawDeserializer.readBytes(is);
            if(buffer.remaining() > 0){
            final short id = RawDeserializer.getDataId(buffer);
            if (id == configDeserializer.getId()) {
               configDeserializer.deserialize(buffer);
               final EventHeader config = state.getConfig();

               // some checkes
               if (deserializerProvider.hasAllDeserializer(config.getEventSerializerIds())) {
                  deserializerProvider.init(state);
               } else {
                  final String message = String.format("Supported Deserializers '%s' but extected are '%s'.",
                        deserializerProvider.getDeserializerIds(), config.getEventSerializerIds());
                  LOGGER.error(message);
                  throw new IllegalStateException(message);
               }
            } else {
               final RawDeserializer<EventHeader, V> deserializer = deserializerProvider.getDeserializer(id);

               if (deserializer != null) {
                  deserializer.deserialize(buffer);
               } else {
                   final String message =
                   String.format("There is no Deserializer for '%d'. Available ids are '%s'.", id, deserializerProvider.getDeserializerIds());
                   LOGGER.warn(message);
                   throw new IllegalStateException(message);
               }
            }
            } else{
               // stop
               state.stop();
            }
         }
      } catch (IOException e) {
         // assume end of stream reached
         LOGGER.debug("Reached end of stream.", e);
      }
   }

   @Override
   public void close() {
      try {
         is.close();
      } catch (Exception e) {
         LOGGER.warn("Could not close InputStream.", e);
      }
   }

}
