package ch.psi.daq.queryrest.response.raw.receive;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.psi.daq.domain.json.BackendChannel;
import ch.psi.daq.queryrest.response.raw.event.serialize.EventHeader;

public class Receiver<C, V> implements IReceiver<C, V> {
   private static final Logger LOGGER = LoggerFactory.getLogger(Receiver.class);

   private final AtomicBoolean isRunning = new AtomicBoolean();

   private final List<Consumer<C>> configHandlers = new ArrayList<>();
   private final List<Consumer<Map<BackendChannel, V>>> valueHandlers = new ArrayList<>();

   private final ReceiverConfig<C, V> receiverConfig;
   private ReceiverState<C, V> receiverState;

   public Receiver(ReceiverConfig<C, V> receiverConfig) {
      this.receiverConfig = receiverConfig;

      // this.dataHeaderHandlers.add(this.receiverConfig.getMessageExtractor());
   }

   public void connect() {
      if (isRunning.compareAndSet(false, true)) {
         // ensures new state after reconnect
         receiverState = new ReceiverStateImpl<>(isRunning);

         receiverConfig.getMessageReader().connect(receiverState);
      }
   }

   @Override
   public void close() {
      if (isRunning.compareAndSet(true, false)) {
         cleanup();
      }
   }

   protected void cleanup() {
      receiverConfig.getMessageReader().close();
   }

   public Message<C, V> receive() throws RuntimeException {
      receiverState.init();
      
      try {
         while (receiverState.isRunning() && !receiverState.hasValues()) {
            receiverConfig.getMessageReader().readNext();
         }

         if (receiverState.hasValues()) {
            final Message<C, V> message = receiverState.getMessage();

            if (message.isConfigChanged()) {
               final C config = message.getConfig();
               if (receiverConfig.isParallelHandlerProcessing()) {
                  configHandlers.parallelStream().forEach(handler -> handler.accept(config));
               } else {
                  for (Consumer<C> eventHeaderHandler : configHandlers) {
                     eventHeaderHandler.accept(config);
                  }
               }
            }

            final Map<BackendChannel, V> values = message.getValues();
            if (receiverConfig.isParallelHandlerProcessing()) {
               valueHandlers.parallelStream().forEach(handler -> handler.accept(values));
            } else {
               for (Consumer<Map<BackendChannel, V>> valueHandler : valueHandlers) {
                  valueHandler.accept(values);
               }
            }

            return message;
         } else {
            return null;
         }
      } catch (Exception e) {
         LOGGER.error(
               "Receiver of '{}' stopped unexpectedly.", receiverConfig, e);
         isRunning.set(false);
         throw e;
      } finally {
         if (!isRunning.get()) {
            cleanup();
         }
      }
   }

   public ReceiverConfig<C, V> getReceiverConfig() {
      return receiverConfig;
   }

   public ReceiverState<C, V> getReceiverState() {
      return receiverState;
   }

   @Override
   public Collection<Consumer<Map<BackendChannel, V>>> getValueHandlers() {
      return valueHandlers;
   }

   public void addValueHandler(Consumer<Map<BackendChannel, V>> handler) {
      valueHandlers.add(handler);
   }

   public void removeValueHandler(Consumer<Map<BackendChannel, V>> handler) {
      valueHandlers.remove(handler);
   }

   @Override
   public Collection<Consumer<C>> getConfigHandlers() {
      return configHandlers;
   }

   public void addConfigHandler(Consumer<C> handler) {
      configHandlers.add(handler);
   }

   public void removeEventHeaderHandler(Consumer<EventHeader> handler) {
      configHandlers.remove(handler);
   }
}
