package ch.psi.daq.test.queryrest.backend;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNotSame;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;

import java.util.HashSet;
import java.util.Set;

import javax.annotation.Resource;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.springframework.beans.factory.NoSuchBeanDefinitionException;
import org.springframework.context.ApplicationContext;

import ch.psi.daq.common.serialization.SerializationHelper;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.filestorage.config.FileStorageConfig;
import ch.psi.daq.test.queryrest.AbstractDaqRestTest;

public class BackendTest extends AbstractDaqRestTest {

   @Resource
   private ApplicationContext context;

   @Before
   public void setUp() throws Exception {
      context.getBean(DomainConfig.BEAN_NAME_BACKENDS);
   }

   @After
   public void tearDown() {}

   @Test
   public void testSerialization_01() throws Exception {
      boolean backendsInit = false;

      for (final Backend backend : Backend.getBackends()) {
         backendsInit = true;
         final Backend copy = SerializationHelper.copy(backend);

         assertSame(backend, copy);
         assertNotNull(Backend.byName(backend.getName().toLowerCase()));
         assertNotNull(Backend.byName(backend.getName().toUpperCase()));
         assertNotNull(Backend.byId(backend.getId()));
      }
      assertTrue(backendsInit);
   }

   @Test
   public void testOverload_01() throws Exception {
      int timeout;
      boolean backendsInit = false;
      Backend backendContext;
      String userHome = System.getProperty("user.home");
      String storageRootDir;
      boolean compaction;

      for (final Backend backend : Backend.getBackends()) {
         backendsInit = true;
         final ApplicationContext context = backend.getApplicationContext();

         // from domain.properties
         timeout = context.getBean(DomainConfig.BEAN_NAME_UPDATE_TIMEOUT, Integer.class);
         assertEquals(30, timeout);

         // from queryrest-test.properties
         timeout = context.getBean(DomainConfig.BEAN_NAME_QUERY_TIMEOUT, Integer.class);
         assertEquals(11, timeout);

         if (backend.getId() == 255) {
            // from test-overload.properties
            timeout = context.getBean(DomainConfig.BEAN_NAME_SHUTDOWN_TIMEOUT, Integer.class);
            assertEquals(32, timeout);

            // from test-overload.properties
            backendContext = context.getBean(DomainConfig.BEAN_NAME_BACKEND_OF_CONTEXT, Backend.class);
            assertEquals("queryrest-1", backendContext.getName());
            assertSame(backend, backendContext);

            timeout = context.getBean(DomainConfig.BEAN_NAME_DELETE_TIMEOUT, Integer.class);
            assertEquals(42, timeout);
            storageRootDir =
                  context.getBean(FileStorageConfig.BEAN_NAME_ROOT_DIR, String.class);
            assertEquals(userHome + "/daq/queryrest1", storageRootDir);

            // from filestorage.properties
            compaction =
                  context.getBean(FileStorageConfig.BEAN_NAME_FILESTORAGE_COMPACTION_TTL_UNCOMPACTED_DELETES,
                        Boolean.class);
            assertEquals(false, compaction);
         } else if (backend.getId() == 254) {
            // from queryrest-test.properties
            timeout = context.getBean(DomainConfig.BEAN_NAME_SHUTDOWN_TIMEOUT, Integer.class);
            assertEquals(31, timeout);

            // from test-overload2.properties
            backendContext = context.getBean(DomainConfig.BEAN_NAME_BACKEND_OF_CONTEXT, Backend.class);
            assertEquals("queryrest-2", backendContext.getName());
            assertSame(backend, backendContext);

            timeout = context.getBean(DomainConfig.BEAN_NAME_DELETE_TIMEOUT, Integer.class);
            assertEquals(43, timeout);
         } else if (backend.getId() == 253) {
            // from queryrest-test.properties
            timeout = context.getBean(DomainConfig.BEAN_NAME_SHUTDOWN_TIMEOUT, Integer.class);
            assertEquals(31, timeout);

            // from test-overload3.properties
            backendContext = context.getBean(DomainConfig.BEAN_NAME_BACKEND_OF_CONTEXT, Backend.class);
            assertEquals("queryrest-3", backendContext.getName());
            assertSame(backend, backendContext);

            timeout = context.getBean(DomainConfig.BEAN_NAME_DELETE_TIMEOUT, Integer.class);
            assertEquals(50, timeout);
         }
      }
      assertTrue(backendsInit);

      // from domain.properties
      timeout = context.getBean(DomainConfig.BEAN_NAME_UPDATE_TIMEOUT, Integer.class);
      assertEquals(30, timeout);

      // from queryrest-test.properties
      timeout = context.getBean(DomainConfig.BEAN_NAME_QUERY_TIMEOUT, Integer.class);
      assertEquals(11, timeout);

      // from queryrest-test.properties
      timeout = context.getBean(DomainConfig.BEAN_NAME_SHUTDOWN_TIMEOUT, Integer.class);
      assertEquals(31, timeout);

      // does not see these beans
      try {
         storageRootDir = context.getBean(FileStorageConfig.BEAN_NAME_ROOT_DIR, String.class);
         assertTrue(false);
      } catch (NoSuchBeanDefinitionException e) {
         assertTrue(true);
      }
      try {
         compaction =
               context.getBean(FileStorageConfig.BEAN_NAME_FILESTORAGE_COMPACTION_TTL_UNCOMPACTED_DELETES,
                     Boolean.class);
         assertTrue(false);
      } catch (NoSuchBeanDefinitionException e) {
         assertTrue(true);
      }
   }

   @Test
   public void testInstance_01() throws Exception {
      final Set<ApplicationContext> parentContexts = new HashSet<>();
      final Set<ApplicationContext> backendContexts = new HashSet<>();
      for (final Backend backend : Backend.getBackends()) {
         parentContexts.add(backend.getParentApplicationContext());
         backendContexts.add(backend.getApplicationContext());
      }

      assertEquals(1, parentContexts.size());
      assertEquals(Backend.getBackends().size(), backendContexts.size());

      final ApplicationContext parentContext = parentContexts.iterator().next();
      assertSame(context, parentContext);
      for (final ApplicationContext context : backendContexts) {
         assertNotSame(context, parentContext);
         assertSame(parentContext, context.getParent());
      }
   }
}
