package ch.psi.daq.test.queryrest.controller;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVParser;
import org.apache.commons.csv.CSVRecord;
import org.apache.http.client.utils.URIBuilder;
import org.junit.After;
import org.junit.Test;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.result.MockMvcResultHandlers;
import org.springframework.test.web.servlet.result.MockMvcResultMatchers;

import ch.psi.bsread.message.Type;
import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.common.time.TimeUtils;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.mapping.Mapping;
import ch.psi.daq.domain.query.mapping.Alignment;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.AggregationDescriptor;
import ch.psi.daq.domain.query.operation.AggregationType;
import ch.psi.daq.domain.query.operation.Compression;
import ch.psi.daq.domain.query.operation.ConfigField;
import ch.psi.daq.domain.query.operation.Extrema;
import ch.psi.daq.domain.query.operation.EventField;
import ch.psi.daq.domain.request.range.RequestRangeDate;
import ch.psi.daq.domain.request.range.RequestRangePulseId;
import ch.psi.daq.domain.request.range.RequestRangeTime;
import ch.psi.daq.domain.test.TestTimeUtils;
import ch.psi.daq.queryrest.response.csv.CSVHTTPResponse;
import ch.psi.daq.queryrest.response.csv.CSVResponseStreamWriter;
import ch.psi.daq.test.queryrest.AbstractDaqRestTest;

/**
 * Tests the {@link DaqController} implementation.
 */
public class CSVQueryRestControllerTest extends AbstractDaqRestTest implements ApplicationContextAware {

   public static final String TEST_CHANNEL = "testChannel";
   public static final String TEST_CHANNEL_01 = TEST_CHANNEL + "1";
   public static final String TEST_CHANNEL_02 = TEST_CHANNEL + "2";
   public static final String[] TEST_CHANNEL_NAMES = new String[] {TEST_CHANNEL_01, TEST_CHANNEL_02};

   private Backend backend;

   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      backend = context.getBean(DomainConfig.BEAN_NAME_BACKEND_DEFAULT, Backend.class);
      context = backend.getApplicationContext();
   }

   @After
   public void tearDown() throws Exception {}

   @Test
   public void testPulseRangeQuery() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.iocSeconds);
      eventFields.add(EventField.iocMillis);
      eventFields.add(EventField.globalSeconds);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.shape);
      eventFields.add(EventField.eventCount);
      eventFields.add(EventField.value);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals(eventFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals(eventFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("" + pulse, record.get(column++));
            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testPulseRangeQuery_Match() throws Exception {
      Map<String, Set<Long>> channelSeq = new HashMap<>();
      Set<Long> channel_01_Seq = new LinkedHashSet<>(Arrays.asList(0L, 2L, 4L, 5L));
      String channel_01 = "TestChannel_Sequence_" + Arrays.toString(channel_01_Seq.toArray());
      channelSeq.put(channel_01, channel_01_Seq);
      Set<Long> channel_02_Seq = new LinkedHashSet<>(Arrays.asList(1L, 3L, 5L));
      String channel_02 = "TestChannel_Sequence_" + Arrays.toString(channel_02_Seq.toArray());
      channelSeq.put(channel_02, channel_02_Seq);
      Set<Long> channel_03_Seq = new LinkedHashSet<>(Arrays.asList(0L, 1L, 2L, 3L, 4L, 5L));
      String channel_03 = "TestChannel_Sequence_" + Arrays.toString(channel_03_Seq.toArray());
      channelSeq.put(channel_03, channel_03_Seq);
      List<String> channels = Arrays.asList(channel_01, channel_02, channel_03);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  // dummy range as range is defined by channel_Seq (see
                  // DummyCassandraReader.getDummyEventStream())
                  0,
                  0),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.value);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 6;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         CSVRecord record = records.remove(0);
         assertEquals(eventFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals(eventFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               if (channelSeq.get(channel).contains(pulse)) {
                  assertEquals(channel, record.get(column++));
                  assertEquals("" + pulse, record.get(column++));
                  assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
                  assertEquals("" + pulse, record.get(column++));
               } else {
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
               }

            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testPulseRangeQuery_DuplicatedPulse() throws Exception {
      Map<String, Collection<Long>> channelSeq = new HashMap<>();
      List<Long> channel_01_Seq = Arrays.asList(0L, 1L, 1L, 2L);
      String channel_01 = "TestChannel_Scalar_01_" + Arrays.toString(channel_01_Seq.toArray());
      channelSeq.put(channel_01, channel_01_Seq);
      List<Long> channel_02_Seq = Arrays.asList(0L, 1L, 1L, 2L);
      String channel_02 = "TestChannel_Waveform_02_" + Arrays.toString(channel_01_Seq.toArray());
      channelSeq.put(channel_02, channel_02_Seq);
      List<String> channels = Arrays.asList(channel_01, channel_02);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  2),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.value);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 3;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         CSVRecord record = records.remove(0);
         assertEquals(eventFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals(eventFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               if (channelSeq.get(channel).contains(pulse)) {
                  assertEquals(channel, record.get(column++));
                  assertEquals("" + pulse, record.get(column++));
                  assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
                  if (channel.toLowerCase().contains("waveform")) {
                     if (pulse != 1) {
                        assertEquals("[" + pulse + "," + pulse + ",2,3,4,5,6,7]", record.get(column++));
                     } else {
                        assertEquals(
                              "[{min:1.0 mean:1.0 max:1.0},{min:1.0 mean:1.0 max:1.0},{min:2.0 mean:2.0 max:2.0},{min:3.0 mean:3.0 max:3.0},{min:4.0 mean:4.0 max:4.0},{min:5.0 mean:5.0 max:5.0},{min:6.0 mean:6.0 max:6.0},{min:7.0 mean:7.0 max:7.0}]",
                              record.get(column++));
                     }
                  } else {
                     if (pulse != 1) {
                        assertEquals("" + pulse, record.get(column++));
                     } else {
                        assertEquals("{min:1.0 mean:1.0 max:1.0}", record.get(column++));
                     }
                  }
               } else {
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
               }

            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }

      request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  2),
            channels);
      request.setResponse(new CSVHTTPResponse());
      request.setEventFields(eventFields);
      request.setMapping(new Mapping());
      request.getMapping().setAggregations(Arrays.asList(Aggregation.count, Aggregation.mean));

      content = mapper.writeValueAsString(request);
      System.out.println(content);

      result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      reader = new StringReader(response);
      csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 3;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         CSVRecord record = records.remove(0);
         assertEquals(eventFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals(eventFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               if (channelSeq.get(channel).contains(pulse)) {
                  assertEquals(channel, record.get(column++));
                  assertEquals("" + pulse, record.get(column++));
                  assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
                  if (channel.toLowerCase().contains("waveform")) {
                     if (pulse != 1) {
                        assertEquals("[" + pulse + "," + pulse + ",2,3,4,5,6,7]", record.get(column++));
                     } else {
                        assertEquals(
                              "[{count:2.0 mean:1.0},{count:2.0 mean:1.0},{count:2.0 mean:2.0},{count:2.0 mean:3.0},{count:2.0 mean:4.0},{count:2.0 mean:5.0},{count:2.0 mean:6.0},{count:2.0 mean:7.0}]",
                              record.get(column++));
                     }
                  } else {
                     if (pulse != 1) {
                        assertEquals("" + pulse, record.get(column++));
                     } else {
                        assertEquals("{count:2.0 mean:1.0}", record.get(column++));
                     }
                  }
               } else {
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
               }

            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }

      request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  2),
            channels);
      request.setResponse(new CSVHTTPResponse());
      request.setEventFields(eventFields);
      request.setMapping(new Mapping());
      request.getMapping().setAggregations(Arrays.asList(Aggregation.mean));

      content = mapper.writeValueAsString(request);
      System.out.println(content);

      result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      reader = new StringReader(response);
      csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 3;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         CSVRecord record = records.remove(0);
         assertEquals(eventFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals(eventFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               if (channelSeq.get(channel).contains(pulse)) {
                  assertEquals(channel, record.get(column++));
                  assertEquals("" + pulse, record.get(column++));
                  assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
                  if (channel.toLowerCase().contains("waveform")) {
                     if (pulse != 1) {
                        assertEquals("[" + pulse + "," + pulse + ",2,3,4,5,6,7]", record.get(column++));
                     } else {
                        assertEquals(
                              "[1.0,1.0,2.0,3.0,4.0,5.0,6.0,7.0]",
                              record.get(column++));
                     }
                  } else {
                     if (pulse != 1) {
                        assertEquals("" + pulse, record.get(column++));
                     } else {
                        assertEquals("1.0", record.get(column++));
                     }
                  }
               } else {
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
               }

            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testPulseRangeQuery_Match_NoTimeField() throws Exception {
      Map<String, Set<Long>> channelSeq = new HashMap<>();
      Set<Long> channel_01_Seq = new LinkedHashSet<>(Arrays.asList(0L, 2L, 4L, 5L));
      String channel_01 = "TestChannel_Sequence_" + Arrays.toString(channel_01_Seq.toArray());
      channelSeq.put(channel_01, channel_01_Seq);
      Set<Long> channel_02_Seq = new LinkedHashSet<>(Arrays.asList(1L, 3L, 5L));
      String channel_02 = "TestChannel_Sequence_" + Arrays.toString(channel_02_Seq.toArray());
      channelSeq.put(channel_02, channel_02_Seq);
      Set<Long> channel_03_Seq = new LinkedHashSet<>(Arrays.asList(0L, 1L, 2L, 3L, 4L, 5L));
      String channel_03 = "TestChannel_Sequence_" + Arrays.toString(channel_03_Seq.toArray());
      channelSeq.put(channel_03, channel_03_Seq);
      List<String> channels = Arrays.asList(channel_01, channel_02, channel_03);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  // dummy range as range is defined by channel_Seq (see
                  // DummyCassandraReader.getDummyEventStream())
                  0,
                  0),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.value);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 6;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals((eventFields.size() + 1) * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
            assertEquals(
                  channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + EventField.globalMillis.name(),
                  record.get(column++));
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals((eventFields.size() + 1) * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               if (channelSeq.get(channel).contains(pulse)) {
                  assertEquals(channel, record.get(column++));
                  assertEquals("" + pulse, record.get(column++));
                  assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               } else {
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
                  assertEquals(CSVResponseStreamWriter.EMPTY_VALUE, record.get(column++));
               }

            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   // @Test
   // public void testPulseRangeQueries() throws Exception {
   // List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
   // String testChannel3 = "testChannel3";
   // DAQQueries request = new DAQQueries(
   // new DAQQueryElement(
   // new RequestRangePulseId(
   // 0,
   // 1),
   // channels),
   // new DAQQueryElement(
   // new RequestRangePulseId(
   // 0,
   // 1),
   // testChannel3));
   // request.setResponse(new CSVHTTPResponse());
   // channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02, testChannel3);
   //
   // LinkedHashSet<eventField> eventFields = new LinkedHashSet<>();
   // eventFields.add(eventField.channel);
   // eventFields.add(eventField.pulseId);
   // eventFields.add(eventField.iocSeconds);
   // eventFields.add(eventField.iocMillis);
   // eventFields.add(eventField.globalSeconds);
   // eventFields.add(eventField.globalMillis);
   // eventFields.add(eventField.shape);
   // eventFields.add(eventField.eventCount);
   // eventFields.add(eventField.value);
   // for (DAQQueryElement element : request) {
   // element.setEventFields(eventFields);
   // }
   //
   // String content = mapper.writeValueAsString(request);
   // System.out.println(content);
   //
   // MvcResult result = this.mockMvc
   // .perform(MockMvcRequestBuilders
   // .post(DomainConfig.PATH_QUERIES)
   // .contentType(MediaType.APPLICATION_JSON)
   // .content(content))
   // .andDo(MockMvcResultHandlers.print())
   // .andExpect(MockMvcResultMatchers.status().isOk())
   // .andReturn();
   //
   // String response = result.getResponse().getContentAsString();
   // System.out.println("Response: " + response);
   //
   // CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
   // StringReader reader = new StringReader(response);
   // CSVParser csvParser = new CSVParser(reader, csvFormat);
   //
   // try {
   // long pulse = 0;
   // int totalEventRows = 2;
   //
   // List<CSVRecord> records = csvParser.getRecords();
   // assertEquals(totalEventRows + 1, records.size());
   // // remove header
   // CSVRecord record = records.remove(0);
   // assertEquals(eventFields.size() * channels.size(), record.size());
   // int column = 0;
   // for (String channel : channels) {
   // for (eventField eventField : eventFields) {
   // assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME +
   // eventField.name(),
   // record.get(column++));
   // }
   // }
   //
   // for (int row = 0; row < totalEventRows; ++row) {
   // record = records.get(row);
   //
   // assertEquals(eventFields.size() * channels.size(), record.size());
   //
   // column = 0;
   // for (String channel : channels) {
   // assertEquals(channel, record.get(column++));
   // assertEquals("" + pulse, record.get(column++));
   // assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
   // record.get(column++));
   // assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse)),
   // record.get(column++));
   // assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
   // record.get(column++));
   // assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse)),
   // record.get(column++));
   // assertEquals("[1]", record.get(column++));
   // assertEquals("1", record.get(column++));
   // assertEquals("" + pulse, record.get(column++));
   // }
   // ++pulse;
   // }
   // } finally {
   // reader.close();
   // csvParser.close();
   // }
   // }

   @Test
   public void testPulseRangeQueryWaveform() throws Exception {
      List<String> channels = Arrays.asList("XYWaveform");
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.iocSeconds);
      eventFields.add(EventField.iocMillis);
      eventFields.add(EventField.globalSeconds);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.shape);
      eventFields.add(EventField.eventCount);
      eventFields.add(EventField.value);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals(eventFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals(eventFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals("[8]", record.get(column++));
               assertEquals("1", record.get(column++));
               assertTrue(record.get(column).startsWith("[" + pulse + ","));
               assertTrue(record.get(column++).endsWith("]"));
            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testTimeRangeQuery() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(0, 0),
                  TimeUtils.getTimeFromMillis(10, 0)),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.iocSeconds);
      eventFields.add(EventField.iocMillis);
      eventFields.add(EventField.globalSeconds);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.shape);
      eventFields.add(EventField.eventCount);
      eventFields.add(EventField.value);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals(eventFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals(eventFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("" + pulse, record.get(column++));
            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testTimeRangeQueryConfigFields() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(0, 0),
                  TimeUtils.getTimeFromMillis(10, 0)),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<ConfigField> configFields = new LinkedHashSet<>();
      configFields.add(ConfigField.name);
      configFields.add(ConfigField.pulseId);
      configFields.add(ConfigField.globalSeconds);
      configFields.add(ConfigField.globalMillis);
      configFields.add(ConfigField.shape);
      configFields.add(ConfigField.description);
      configFields.add(ConfigField.backend);
      configFields.add(ConfigField.modulo);
      configFields.add(ConfigField.offset);
      configFields.add(ConfigField.keyspace);
      configFields.add(ConfigField.precision);
      configFields.add(ConfigField.source);
      configFields.add(ConfigField.type);
      configFields.add(ConfigField.unit);
      request.setConfigFields(configFields);

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.iocSeconds);
      eventFields.add(EventField.iocMillis);
      eventFields.add(EventField.globalSeconds);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.shape);
      eventFields.add(EventField.eventCount);
      eventFields.add(EventField.value);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         // pulse of last element
         long configPulse = 1;
         int totalConfigRows = 1;
         int totalEventRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1 + totalConfigRows + 1 + 1, records.size());
         // remove config header
         CSVRecord record = records.remove(0);
         assertEquals(configFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (ConfigField configField : configFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + configField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalConfigRows; ++row) {
            record = records.remove(0);

            assertEquals(configFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + configPulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(configPulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(configPulse).getMillis(),
                     record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("", record.get(column++));
               assertEquals(backend.getName(), record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("0", record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("0", record.get(column++));
               assertEquals("unknown", record.get(column++));
               assertEquals(Type.Int32.getKey(), record.get(column++));
               assertEquals("", record.get(column++));
            }
         }

         // empty line
         record = records.remove(0);
         assertEquals(1, record.size());
         assertEquals("", record.get(0));

         // remove event header
         record = records.remove(0);
         assertEquals(eventFields.size() * channels.size(), record.size());
         column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals(eventFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("" + pulse, record.get(column++));
            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testDateRangeQuery() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      String startDate = TimeUtils.format(0);
      String endDate = TimeUtils.format(10);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.iocSeconds);
      eventFields.add(EventField.iocDate);
      eventFields.add(EventField.iocMillis);
      eventFields.add(EventField.globalSeconds);
      eventFields.add(EventField.globalDate);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.shape);
      eventFields.add(EventField.eventCount);
      eventFields.add(EventField.value);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals(eventFields.size() * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals(eventFields.size() * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.format(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TimeUtils.format(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("" + pulse, record.get(column++));
            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testExtremaAggregation() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            Ordering.asc,
            new AggregationDescriptor(AggregationType.extrema),
            TEST_CHANNEL_NAMES[0]);
      request.setResponse(new CSVHTTPResponse());

      String content = mapper.writeValueAsString(request);

      try {
         this.mockMvc
               .perform(MockMvcRequestBuilders
                     .post(DomainConfig.PATH_QUERY)
                     .contentType(MediaType.APPLICATION_JSON)
                     .content(content))
               .andDo(MockMvcResultHandlers.print());
         // .andExpect(MockMvcResultMatchers.model().attribute(
         // "exception",
         // Matchers.isA(IllegalArgumentException.class)));
         // should throw an exception
         assertTrue(false);
      } catch (Throwable e) {
         assertTrue(true);
      }
   }

   @Test
   public void testIndexAggregation() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            Ordering.asc,
            new AggregationDescriptor(AggregationType.index),
            TEST_CHANNEL_NAMES[0]);
      request.setResponse(new CSVHTTPResponse());

      String content = mapper.writeValueAsString(request);

      try {
         this.mockMvc
               .perform(MockMvcRequestBuilders
                     .post(DomainConfig.PATH_QUERY)
                     .contentType(MediaType.APPLICATION_JSON)
                     .content(content))
               .andDo(MockMvcResultHandlers.print());
         // .andExpect(MockMvcResultMatchers.model().attribute(
         // "exception",
         // Matchers.isA(IllegalArgumentException.class)));
         // should throw an exception
         assertTrue(false);
      } catch (Throwable e) {
         assertTrue(true);
      }
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregate() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01);
      long startTime = 0;
      long endTime = 99;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      List<Aggregation> aggregations = new ArrayList<>();
      aggregations.add(Aggregation.min);
      aggregations.add(Aggregation.mean);
      aggregations.add(Aggregation.max);

      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            channels);
      request.setAggregation(new AggregationDescriptor().setNrOfBins(2).setAggregations(aggregations));
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.iocSeconds);
      eventFields.add(EventField.iocMillis);
      eventFields.add(EventField.globalSeconds);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.shape);
      eventFields.add(EventField.eventCount);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals((eventFields.size() + aggregations.size()) * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
            for (Aggregation aggregation : aggregations) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + EventField.value
                     + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + aggregation.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals((eventFields.size() + aggregations.size()) * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("5", record.get(column++));
               assertEquals("" + pulse + ".0", record.get(column++));
               assertEquals("" + (pulse + 2) + ".0", record.get(column++));
               assertEquals("" + (pulse + 4) + ".0", record.get(column++));
            }
            pulse += 5;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregateExtrema() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01);
      long startTime = 0;
      long endTime = 99;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      List<Aggregation> aggregations = new ArrayList<>();
      aggregations.add(Aggregation.min);
      aggregations.add(Aggregation.mean);
      aggregations.add(Aggregation.max);
      List<Extrema> extrema = new ArrayList<>();
      extrema.add(Extrema.minValue);
      extrema.add(Extrema.maxValue);

      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            channels);
      request.setAggregation(new AggregationDescriptor().setNrOfBins(2).setAggregations(aggregations)
            .setExtrema(extrema));
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.iocSeconds);
      eventFields.add(EventField.iocMillis);
      eventFields.add(EventField.globalSeconds);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.shape);
      eventFields.add(EventField.eventCount);
      eventFields.add(EventField.value);
      request.setEventFields(eventFields);

      Set<EventField> extremaFields = new LinkedHashSet<>();
      for (Extrema extremum : extrema) {
         for (EventField eventField : eventFields) {
            if (extremum.getAccessor(eventField) != null) {
               extremaFields.add(eventField);
            }
         }
      }

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      // will not be included as it is an aggregation
      eventFields.remove(EventField.value);
      try {
         long pulse = 0;
         int totalEventRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals(
               (eventFields.size() + aggregations.size() + (extremaFields.size() * extrema.size())) * channels.size(),
               record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
            for (Aggregation aggregation : aggregations) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + EventField.value
                     + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + aggregation.name(),
                     record.get(column++));
            }
            for (Extrema extremum : extrema) {
               for (EventField eventField : extremaFields) {
                  assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME
                        + CSVResponseStreamWriter.FIELDNAME_EXTREMA
                        + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + extremum.name()
                        + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                        record.get(column++));
               }
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals((eventFields.size() + aggregations.size() + (extremaFields.size() * extrema.size()))
                  * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("5", record.get(column++));
               assertEquals("" + pulse + ".0", record.get(column++));
               assertEquals("" + (pulse + 2) + ".0", record.get(column++));
               assertEquals("" + (pulse + 4) + ".0", record.get(column++));

               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("" + pulse + ".0", record.get(column++));

               assertEquals("" + (pulse + 4), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId((pulse + 4))), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId((pulse + 4)).getMillis(),
                     record.get(column++));
               assertEquals("1", record.get(column++));
               assertEquals("" + (pulse + 4) + ".0", record.get(column++));
            }
            pulse += 5;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testDateRangeQueryBinSizeAggregate_01() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01);
      long startTime = 0;
      long endTime = 999;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      List<Aggregation> aggregations = new ArrayList<>();
      aggregations.add(Aggregation.min);
      aggregations.add(Aggregation.mean);
      aggregations.add(Aggregation.max);

      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            channels);
      request.setAggregation(new AggregationDescriptor().setDurationPerBin(100).setAggregations(aggregations));
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.iocSeconds);
      eventFields.add(EventField.iocMillis);
      eventFields.add(EventField.globalSeconds);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.shape);
      eventFields.add(EventField.eventCount);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      checkDateRangeQueryBinSizeAggregate(channels, aggregations, eventFields, response);
   }

   @Test
   public void testDateRangeQueryBinSizeAggregate_02() throws Exception {
      List<String> channels = Arrays.asList(
            TEST_CHANNEL_01 + "_{0,100,200,300,400,500,600,700,800,900}",
            TEST_CHANNEL_02 + "_{50,150,250,350,450,550,650,750,850,950}");
      long startTime = 0;
      long endTime = 999;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      List<Aggregation> aggregations = new ArrayList<>();
      aggregations.add(Aggregation.min);
      aggregations.add(Aggregation.mean);
      aggregations.add(Aggregation.max);

      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            channels);
      request.setAggregation(new AggregationDescriptor().setDurationPerBin(100).setAggregations(aggregations));
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.iocSeconds);
      eventFields.add(EventField.iocMillis);
      eventFields.add(EventField.globalSeconds);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.shape);
      eventFields.add(EventField.eventCount);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 10;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals((eventFields.size() + aggregations.size()) * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
            for (Aggregation aggregation : aggregations) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + EventField.value
                     + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + aggregation.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals((eventFields.size() + aggregations.size()) * channels.size(), record.size());

            column = 0;
            String channel = channels.get(0);
            assertEquals(channel, record.get(column++));
            assertEquals("" + pulse, record.get(column++));
            assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
            assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
            assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
            assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
            assertEquals("[1]", record.get(column++));
            assertEquals("1", record.get(column++));
            assertEquals("" + pulse + ".0", record.get(column++));
            assertEquals("" + pulse + ".0", record.get(column++));
            assertEquals("" + pulse + ".0", record.get(column++));

            pulse += 5;
            channel = channels.get(1);
            assertEquals(channel, record.get(column++));
            assertEquals("" + pulse, record.get(column++));
            assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
            assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
            assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
            assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
            assertEquals("[1]", record.get(column++));
            assertEquals("1", record.get(column++));
            assertEquals("" + pulse + ".0", record.get(column++));
            assertEquals("" + pulse + ".0", record.get(column++));
            assertEquals("" + pulse + ".0", record.get(column++));

            pulse += 5;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testDateRangeQueryBinSizeAggregate_MissingPulses() throws Exception {
      List<String> channels = Arrays.asList(
            TEST_CHANNEL_01 + "_{100,200}",
            TEST_CHANNEL_01 + "_{10,110,210}",
            TEST_CHANNEL_01 + "_{120}");
      long startTime = 0;
      long endTime = 400;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      List<Aggregation> aggregations = new ArrayList<>();
      aggregations.add(Aggregation.min);
      aggregations.add(Aggregation.mean);
      aggregations.add(Aggregation.max);

      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            channels);
      request.setAggregation(new AggregationDescriptor().setDurationPerBin(100).setAggregations(aggregations));
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.iocSeconds);
      eventFields.add(EventField.iocMillis);
      eventFields.add(EventField.globalSeconds);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.shape);
      eventFields.add(EventField.eventCount);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 3;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals((eventFields.size() + aggregations.size()) * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
            for (Aggregation aggregation : aggregations) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + EventField.value
                     + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + aggregation.name(),
                     record.get(column++));
            }
         }

         int row = 0;
         record = records.get(row);
         assertEquals((eventFields.size() + aggregations.size()) * channels.size(), record.size());
         column = 0;
         String channel = channels.get(0);
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));

         pulse = 1;
         channel = channels.get(1);
         assertEquals(channel, record.get(column++));
         assertEquals("" + pulse, record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals("[1]", record.get(column++));
         assertEquals("1", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));

         pulse = 2;
         channel = channels.get(2);
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));

         row = 1;
         record = records.get(row);
         assertEquals((eventFields.size() + aggregations.size()) * channels.size(), record.size());
         column = 0;
         pulse = 10;
         channel = channels.get(0);
         assertEquals(channel, record.get(column++));
         assertEquals("" + pulse, record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals("[1]", record.get(column++));
         assertEquals("1", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));

         pulse = 11;
         channel = channels.get(1);
         assertEquals(channel, record.get(column++));
         assertEquals("" + pulse, record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals("[1]", record.get(column++));
         assertEquals("1", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));

         pulse = 12;
         channel = channels.get(2);
         assertEquals(channel, record.get(column++));
         assertEquals("" + pulse, record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals("[1]", record.get(column++));
         assertEquals("1", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));

         row = 2;
         record = records.get(row);
         assertEquals((eventFields.size() + aggregations.size()) * channels.size(), record.size());
         column = 0;
         pulse = 20;
         channel = channels.get(0);
         assertEquals(channel, record.get(column++));
         assertEquals("" + pulse, record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals("[1]", record.get(column++));
         assertEquals("1", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));

         pulse = 21;
         channel = channels.get(1);
         assertEquals(channel, record.get(column++));
         assertEquals("" + pulse, record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals("[1]", record.get(column++));
         assertEquals("1", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));
         assertEquals("" + pulse + ".0", record.get(column++));

         pulse = 22;
         channel = channels.get(2);
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testDateRangeQuery_NoneAlignment() throws Exception {
      List<String> channels = Arrays.asList(
            TEST_CHANNEL_01 + "_{100,200}",
            TEST_CHANNEL_01 + "_{10,110,210}",
            TEST_CHANNEL_01 + "_{120}");
      long startTime = 0;
      long endTime = 400;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);

      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            channels);
      request.setResponse(new CSVHTTPResponse());
      request.setMapping(new Mapping(Alignment.NONE));

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.iocSeconds);
      eventFields.add(EventField.iocMillis);
      eventFields.add(EventField.globalSeconds);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.shape);
      eventFields.add(EventField.eventCount);
      eventFields.add(EventField.value);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 3;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals((eventFields.size()) * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
         }

         int row = 0;
         record = records.get(row);
         assertEquals((eventFields.size()) * channels.size(), record.size());
         column = 0;
         pulse = 10;
         String channel = channels.get(0);
         assertEquals(channel, record.get(column++));
         assertEquals("" + pulse, record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals("[1]", record.get(column++));
         assertEquals("1", record.get(column++));
         assertEquals("" + pulse, record.get(column++));

         pulse = 1;
         channel = channels.get(1);
         assertEquals(channel, record.get(column++));
         assertEquals("" + pulse, record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals("[1]", record.get(column++));
         assertEquals("1", record.get(column++));
         assertEquals("" + pulse, record.get(column++));

         pulse = 12;
         channel = channels.get(2);
         assertEquals(channel, record.get(column++));
         assertEquals("" + pulse, record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals("[1]", record.get(column++));
         assertEquals("1", record.get(column++));
         assertEquals("" + pulse, record.get(column++));

         row = 1;
         record = records.get(row);
         assertEquals((eventFields.size()) * channels.size(), record.size());
         column = 0;
         pulse = 20;
         channel = channels.get(0);
         assertEquals(channel, record.get(column++));
         assertEquals("" + pulse, record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals("[1]", record.get(column++));
         assertEquals("1", record.get(column++));
         assertEquals("" + pulse, record.get(column++));

         pulse = 11;
         channel = channels.get(1);
         assertEquals(channel, record.get(column++));
         assertEquals("" + pulse, record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals("[1]", record.get(column++));
         assertEquals("1", record.get(column++));
         assertEquals("" + pulse, record.get(column++));

         pulse = 12;
         channel = channels.get(2);
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));

         row = 2;
         record = records.get(row);
         assertEquals((eventFields.size()) * channels.size(), record.size());
         column = 0;
         pulse = 20;
         channel = channels.get(0);
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));

         pulse = 21;
         channel = channels.get(1);
         assertEquals(channel, record.get(column++));
         assertEquals("" + pulse, record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
         assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
         assertEquals("[1]", record.get(column++));
         assertEquals("1", record.get(column++));
         assertEquals("" + pulse, record.get(column++));

         pulse = 22;
         channel = channels.get(2);
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
         assertEquals("", record.get(column++));
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   private void checkDateRangeQueryBinSizeAggregate(final List<String> channels, final List<Aggregation> aggregations,
         final Set<EventField> eventFields, final String response) throws Exception {
      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 10;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         // remove header
         CSVRecord record = records.remove(0);
         assertEquals((eventFields.size() + aggregations.size()) * channels.size(), record.size());
         int column = 0;
         for (String channel : channels) {
            for (EventField eventField : eventFields) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + eventField.name(),
                     record.get(column++));
            }
            for (Aggregation aggregation : aggregations) {
               assertEquals(channel + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + EventField.value
                     + CSVResponseStreamWriter.DELIMITER_CHANNELNAME_FIELDNAME + aggregation.name(),
                     record.get(column++));
            }
         }

         for (int row = 0; row < totalEventRows; ++row) {
            record = records.get(row);

            assertEquals((eventFields.size() + aggregations.size()) * channels.size(), record.size());

            column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals(TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)), record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
               assertEquals("[1]", record.get(column++));
               assertEquals("10", record.get(column++));
               assertEquals("" + pulse + ".0", record.get(column++));
               assertEquals("" + (pulse + 4.5), record.get(column++));
               assertEquals("" + (pulse + 9) + ".0", record.get(column++));
            }
            pulse += 10;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testDateRangeQueryBinSizeAggregate_GET() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01);
      long startTime = 0;
      long endTime = 999;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      List<Aggregation> aggregations = new ArrayList<>();
      aggregations.add(Aggregation.min);
      aggregations.add(Aggregation.mean);
      aggregations.add(Aggregation.max);

      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            channels);
      // request.setAggregation(new
      // AggregationDescriptor().setDurationPerBin(100).setAggregations(aggregations));
      // make sure defaults are set
      request.setAggregation(new AggregationDescriptor().setDurationPerBin(100));
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.pulseId);
      eventFields.add(EventField.iocSeconds);
      eventFields.add(EventField.iocMillis);
      eventFields.add(EventField.globalSeconds);
      eventFields.add(EventField.globalMillis);
      eventFields.add(EventField.shape);
      eventFields.add(EventField.eventCount);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      URIBuilder ub = new URIBuilder(DomainConfig.PATH_QUERY);
      ub.addParameter(DomainConfig.PATH_QUERY_PARAM_JSON_BODY, content);
      String url = ub.toString();
      System.out.println(url);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .get(url))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      checkDateRangeQueryBinSizeAggregate(channels, aggregations, eventFields, response);
   }

   @Test
   public void testQuery_NoTimeFields_01() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            channels);
      request.setResponse(new CSVHTTPResponse());

      LinkedHashSet<EventField> eventFields = new LinkedHashSet<>();
      eventFields.add(EventField.channel);
      eventFields.add(EventField.value);
      request.setEventFields(eventFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CSVFormat csvFormat = CSVFormat.EXCEL.withDelimiter(CSVResponseStreamWriter.DELIMITER_CVS);
      StringReader reader = new StringReader(response);
      CSVParser csvParser = new CSVParser(reader, csvFormat);

      try {
         long pulse = 0;
         int totalEventRows = 2;

         List<CSVRecord> records = csvParser.getRecords();
         assertEquals(totalEventRows + 1, records.size());
         // remove header
         records.remove(0);

         for (int row = 0; row < totalEventRows; ++row) {
            CSVRecord record = records.get(row);

            assertEquals((eventFields.size() + 1) * channels.size(), record.size());

            int column = 0;
            for (String channel : channels) {
               assertEquals(channel, record.get(column++));
               assertEquals("" + pulse, record.get(column++));
               assertEquals("" + TestTimeUtils.getTimeFromPulseId(pulse).getMillis(), record.get(column++));
            }
            ++pulse;
         }
      } finally {
         reader.close();
         csvParser.close();
      }
   }

   @Test
   public void testGzipFileSuffixHeader() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);
      request.setResponse(new CSVHTTPResponse(Compression.GZIP));

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Content-Disposition", "attachment; filename=data.csv"));
   }

   @Test
   public void testCvsFileSuffixHeader() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);
      request.setResponse(new CSVHTTPResponse());

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Content-Disposition", "attachment; filename=data.csv"));
   }
}
