package ch.psi.daq.test.queryrest.controller;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.awt.Color;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.junit.After;
import org.junit.Test;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.result.MockMvcResultHandlers;
import org.springframework.test.web.servlet.result.MockMvcResultMatchers;

import com.jayway.jsonpath.Configuration;
import com.jayway.jsonpath.JsonPath;

import ch.psi.bsread.message.Type;
import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.common.time.TimeUtils;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.mapping.IncompleteStrategy;
import ch.psi.daq.domain.query.mapping.Mapping;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.AggregationDescriptor;
import ch.psi.daq.domain.query.operation.AggregationType;
import ch.psi.daq.domain.query.operation.ConfigField;
import ch.psi.daq.domain.query.operation.Extrema;
import ch.psi.daq.domain.query.operation.EventField;
import ch.psi.daq.domain.query.transform.ValueTransformationSequence;
import ch.psi.daq.domain.query.transform.image.ImageDownScaleValueTransformation;
import ch.psi.daq.domain.query.transform.image.ImageEncodingValueTransformation;
import ch.psi.daq.domain.query.transform.image.ImageFormat;
import ch.psi.daq.domain.query.transform.image.color.ColorModelType;
import ch.psi.daq.domain.query.transform.image.color.TypedColorModel;
import ch.psi.daq.domain.query.transform.image.encoding.Base64ImageEncoder;
import ch.psi.daq.domain.query.transform.sampling.ValueSamplerValueTransformation;
import ch.psi.daq.domain.request.range.RequestRangeDate;
import ch.psi.daq.domain.request.range.RequestRangePulseId;
import ch.psi.daq.domain.request.range.RequestRangeTime;
import ch.psi.daq.domain.test.TestTimeUtils;
import ch.psi.daq.domain.test.reader.TestReader;
import ch.psi.daq.test.queryrest.AbstractDaqRestTest;

/**
 * Tests the {@link DaqController} implementation.
 */
public class JsonQueryRestControllerTableTest extends AbstractDaqRestTest implements ApplicationContextAware {

   public static final String TEST_CHANNEL_01 = "testChannel1";
   public static final String TEST_CHANNEL_02 = "testChannel2";
   public static final String TEST_CHANNEL_WAVEFORM_01 = "testChannelWaveform1";
   public static final String TEST_CHANNEL_WAVEFORM_02 = "testChannelWaveform2";
   public static final String[] TEST_CHANNEL_NAMES = new String[] {TEST_CHANNEL_01, TEST_CHANNEL_02};

   private Backend backend;
   private Backend backend2;

   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      backend = context.getBean(DomainConfig.BEAN_NAME_BACKEND_DEFAULT, Backend.class);
      context = backend.getApplicationContext();

      for (final Backend baknd : Backend.getBackends()) {
         if (baknd != backend) {
            backend2 = baknd;
            break;
         }
      }
   }

   @After
   public void tearDown() throws Exception {}

   @Test
   public void testPulseRangeQuery() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            TEST_CHANNEL_NAMES);
      request.setMapping(new Mapping());
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].iocMillis").value(1000))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].iocMillis").value(1000))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].iocMillis").value(1010))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].iocMillis").value(1010));
   }

   @Test
   public void testPulseRangeQuery_FillNull_NotExists() throws Exception {
      final String[] TEST_CHANNEL_NAMES =
            new String[] {TEST_CHANNEL_01, TestReader.DOES_NOT_EXIST_CHANNEL, TEST_CHANNEL_02};
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            TEST_CHANNEL_NAMES);
      request.setMapping(new Mapping(IncompleteStrategy.FILL_NULL));
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].iocMillis").value(1000))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].iocMillis").value(1000))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].iocMillis").value(1010))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2].iocMillis").value(1010));
   }

   @Test
   public void testPulseRangeQuery_NrOfBins() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            TEST_CHANNEL_NAMES);
      request.setMapping(new Mapping());
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);
      request.addEventField(EventField.value);

      AggregationDescriptor aggregation = new AggregationDescriptor(AggregationType.value);
      aggregation.setNrOfBins(1);
      aggregation.setAggregations(Arrays.asList(Aggregation.mean));
      request.setAggregation(aggregation);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].iocMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.mean").value(100.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.min").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.max").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].iocMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value.mean").value(100.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value.min").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value.max").doesNotExist());
   }

   @Test
   public void testPulseRangeQuery_ChannelOrder() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            TEST_CHANNEL_02, TEST_CHANNEL_01);
      request.setMapping(new Mapping());
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].iocMillis").value(1000))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].iocMillis").value(1000))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].iocMillis").value(1010))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].iocMillis").value(1010));
   }

   @Test
   public void testPulseRangeQuery_SingleChannel() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            TEST_CHANNEL_01);
      request.setMapping(new Mapping());
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].iocMillis").value(1000))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].iocMillis").value(1010))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").doesNotExist());
   }

   @Test
   public void testPulseRangeQuery_Fields() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  199),
            new AggregationDescriptor().setNrOfBins(2),
            TEST_CHANNEL_NAMES);
      request.setMapping(new Mapping());
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.eventCount);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].eventCount").value(50))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].eventCount").value(50))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(150))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].eventCount").value(50))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(150))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].eventCount").value(50));
   }

   @Test
   public void testPulseRangeQueryBackends() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            new ChannelName(TEST_CHANNEL_01, backend),
            new ChannelName(TEST_CHANNEL_02, backend2));
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_02))
            .andExpect(
                  MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend2.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(TEST_CHANNEL_02))
            .andExpect(
                  MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend2.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)));
   }

   @Test
   public void testPulseRangeQueries() throws Exception {
      String testChannel3 = "testChannel3";
      DAQQueryElement daqQuery1 = new DAQQueryElement(
            new RequestRangePulseId(
                  100,
                  101),
            TEST_CHANNEL_NAMES);
      daqQuery1.setMapping(new Mapping());
      DAQQueryElement daqQuery2 = new DAQQueryElement(
            new RequestRangePulseId(
                  100,
                  101),
            testChannel3);
      daqQuery2.setMapping(new Mapping());
      DAQQueries request = new DAQQueries(
            daqQuery1, daqQuery2);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERIES)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0][1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1][0].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1][1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))

            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0][0].channel").value(testChannel3))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))

            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1][0].channel").value(testChannel3))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1][0].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)));
   }

   @Test
   public void testTimeRangeQuery_01() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(2000, 0),
                  TimeUtils.getTimeFromMillis(2010, 0)),
            TEST_CHANNEL_NAMES);
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)));
   }

   @Test
   public void testTimeRangeQuery_01_ConfigFields() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(2000, 0),
                  TimeUtils.getTimeFromMillis(2010, 0)),
            TEST_CHANNEL_NAMES);
      request.addConfigField(ConfigField.pulseId);
      request.addConfigField(ConfigField.globalSeconds);
      request.addConfigField(ConfigField.globalMillis);
      request.addConfigField(ConfigField.shape);
      request.addConfigField(ConfigField.description);
      request.addConfigField(ConfigField.backend);
      request.addConfigField(ConfigField.modulo);
      request.addConfigField(ConfigField.offset);
      request.addConfigField(ConfigField.keyspace);
      request.addConfigField(ConfigField.precision);
      request.addConfigField(ConfigField.source);
      request.addConfigField(ConfigField.type);
      request.addConfigField(ConfigField.unit);
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].channel.backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[0].globalMillis").value(2000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[0].shape").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[0].description").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[0].modulo").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[0].offset").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[0].keyspace").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[0].precision").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[0].source").value("unknown"))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[0].type").value(Type.Int32.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[0].unit").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[1].globalMillis").value(2010))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[1].shape").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[1].description").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[1].modulo").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[1].offset").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[1].keyspace").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[1].precision").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[1].source").value("unknown"))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[1].type").value(Type.Int32.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[0].configs[1].unit").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].channel.backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[0].globalMillis").value(2000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[0].shape").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[0].description").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[0].modulo").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[0].offset").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[0].keyspace").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[0].precision").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[0].source").value("unknown"))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[0].type").value(Type.Int32.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[0].unit").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[1].globalMillis").value(2010))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[1].shape").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[1].description").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[1].modulo").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[1].offset").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[1].keyspace").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[1].precision").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[1].source").value("unknown"))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[1].type").value(Type.Int32.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.meta[1].configs[1].unit").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)));
   }

   @Test
   public void testDateRangeQuery() throws Exception {
      String startDate = TimeUtils.format(1000);
      String endDate = TimeUtils.format(1010);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            TEST_CHANNEL_NAMES);
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)));
   }

   @Test
   public void testExtremaAggregation() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            Ordering.asc,
            new AggregationDescriptor(AggregationType.extrema),
            TEST_CHANNEL_NAMES[0]);
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders.post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").doesNotExist());
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].channel").isMap())
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].channel.name").value(TEST_CHANNEL_NAMES[0]))
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.minima").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.minima.min").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.minima.min.value").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.minima.min.value").value(Double.valueOf(100)))
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.minima.min.event").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.minima.min.event.pulseId").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.minima.min.event.pulseId").value(100))
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.maxima").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.maxima.max").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.maxima.max.value").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.maxima.max.value").value(Double.valueOf(101)))
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.maxima.max.event").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.maxima.max.event.pulseId").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$.data[0].data.maxima.max.event.pulseId").value(101));
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregate() throws Exception {
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor().setNrOfBins(2).setAggregations(
                        Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max)),
                  TEST_CHANNEL_NAMES);
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.max").value(104.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value.min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value.mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value.max").value(104.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value.min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value.mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value.max").value(109.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value.min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value.mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value.max").value(109.0));
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregateExtrema() throws Exception {
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor()
                        .setNrOfBins(2)
                        .setAggregations(Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max))
                        .setExtrema(Arrays.asList(Extrema.minValue, Extrema.maxValue)),
                  TEST_CHANNEL_NAMES);
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema.minValue.value").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema.minValue.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema.minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema.minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema.maxValue.value").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema.maxValue.pulseId").value(104))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema.maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 40000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema.maxValue.eventCount").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value.min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value.mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value.max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema.minValue.value").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema.minValue.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema.minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema.minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema.maxValue.value").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema.maxValue.pulseId").value(104))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema.maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 40000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema.maxValue.eventCount").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value.min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value.mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value.max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema.minValue.value").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema.minValue.pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema.minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema.minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema.maxValue.value").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema.maxValue.pulseId").value(109))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema.maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 90000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema.maxValue.eventCount").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value.min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value.mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value.max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema.minValue.value").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema.minValue.pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema.minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema.minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema.maxValue.value").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema.maxValue.pulseId").value(109))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema.maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 90000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema.maxValue.eventCount").value(1));
   }

   @Test
   public void testDateRangeQueryBinSizeAggregate() throws Exception {
      long startTime = 10000;
      long endTime = 10999;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            new AggregationDescriptor().setDurationPerBin(100),
            TEST_CHANNEL_NAMES);
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.min").value(1000.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.mean").value(1004.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.max").value(1009.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value.min").value(1000.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value.mean").value(1004.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value.max").value(1009.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 100000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value.min").value(1010.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value.mean").value(1014.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value.max").value(1019.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 100000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value.min").value(1010.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value.mean").value(1014.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value.max").value(1019.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].pulseId").value(1020))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 200000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].value.min").value(1020.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].value.mean").value(1024.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].value.max").value(1029.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].pulseId").value(1020))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 200000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].value.min").value(1020.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].value.mean").value(1024.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].value.max").value(1029.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].pulseId").value(1030))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 300000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].value.min").value(1030.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].value.mean").value(1034.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].value.max").value(1039.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].pulseId").value(1030))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 300000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].value.min").value(1030.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].value.mean").value(1034.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].value.max").value(1039.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].pulseId").value(1040))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 400000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].value.min").value(1040.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].value.mean").value(1044.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].value.max").value(1049.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].pulseId").value(1040))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 400000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].value.min").value(1040.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].value.mean").value(1044.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].value.max").value(1049.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].pulseId").value(1050))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 500000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].value.min").value(1050.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].value.mean").value(1054.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].value.max").value(1059.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].pulseId").value(1050))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 500000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value.min").value(1050.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value.mean").value(1054.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value.max").value(1059.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].pulseId").value(1060))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 600000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value.min").value(1060.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value.mean").value(1064.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value.max").value(1069.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1].pulseId").value(1060))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 600000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1].value.min").value(1060.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1].value.mean").value(1064.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1].value.max").value(1069.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][0].pulseId").value(1070))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 700000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][0].value.min").value(1070.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][0].value.mean").value(1074.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][0].value.max").value(1079.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][1].pulseId").value(1070))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 700000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][1].value.min").value(1070.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][1].value.mean").value(1074.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7][1].value.max").value(1079.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][0].pulseId").value(1080))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 800000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][0].value.min").value(1080.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][0].value.mean").value(1084.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][0].value.max").value(1089.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][1].pulseId").value(1080))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 800000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][1].value.min").value(1080.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][1].value.mean").value(1084.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[8][1].value.max").value(1089.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][0].channel").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][0].pulseId").value(1090))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 900000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][0].value.min").value(1090.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][0].value.mean").value(1094.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][0].value.max").value(1099.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][1].channel").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][1].pulseId").value(1090))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 900000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][1].value.min").value(1090.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][1].value.mean").value(1094.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[9][1].value.max").value(1099.0));
   }

   @Test
   public void testDateRangeQueryIndexAggregate() throws Exception {
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor()
                        .setAggregationType(AggregationType.index)
                        .setNrOfBins(2)
                        .setAggregations(Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max)),
                  TEST_CHANNEL_WAVEFORM_01, TEST_CHANNEL_WAVEFORM_02);
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_WAVEFORM_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[0].min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[0].mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[0].max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[1].min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[1].mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[1].max").value(104.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_WAVEFORM_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[0].min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[0].mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[0].max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[1].min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[1].mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[1].max").value(104.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_WAVEFORM_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[0].min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[0].mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[0].max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[1].min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[1].mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[1].max").value(109.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(TEST_CHANNEL_WAVEFORM_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[0].min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[0].mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[0].max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[1].min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[1].mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[1].max").value(109.0));
   }

   @Test
   public void testDateRangeQueryIndexAggregateExtrema() throws Exception {
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor()
                        .setAggregationType(AggregationType.index)
                        .setNrOfBins(2)
                        .setAggregations(Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max))
                        .setExtrema(Arrays.asList(Extrema.minValue, Extrema.maxValue)),
                  TEST_CHANNEL_WAVEFORM_01, TEST_CHANNEL_WAVEFORM_02);
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(TEST_CHANNEL_WAVEFORM_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[0].min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[0].mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[0].max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[1].min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[1].mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[1].max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[0].minValue.value").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[0].minValue.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[0].minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[0].minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[0].maxValue.value").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[0].maxValue.pulseId").value(104))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[0].maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 40000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[0].maxValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[1].minValue.value").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[1].minValue.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[1].minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[1].minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[1].maxValue.value").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[1].maxValue.pulseId").value(104))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[1].maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 40000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].extrema[1].maxValue.eventCount").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(TEST_CHANNEL_WAVEFORM_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[0].min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[0].mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[0].max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[1].min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[1].mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[1].max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[0].minValue.value").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[0].minValue.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[0].minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[0].minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[0].maxValue.value").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[0].maxValue.pulseId").value(104))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[0].maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 40000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[0].maxValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[1].minValue.value").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[1].minValue.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[1].minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[1].minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[1].maxValue.value").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[1].maxValue.pulseId").value(104))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[1].maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 40000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].extrema[1].maxValue.eventCount").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(TEST_CHANNEL_WAVEFORM_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[0].min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[0].mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[0].max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[1].min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[1].mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[1].max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[0].minValue.value").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[0].minValue.pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[0].minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[0].minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[0].maxValue.value").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[0].maxValue.pulseId").value(109))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[0].maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 90000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[0].maxValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[1].minValue.value").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[1].minValue.pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[1].minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[1].minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[1].maxValue.value").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[1].maxValue.pulseId").value(109))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[1].maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 90000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].extrema[1].maxValue.eventCount").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(TEST_CHANNEL_WAVEFORM_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value[0].min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value[0].mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value[0].max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value[1].min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value[1].mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value[1].max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[0].minValue.value").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[0].minValue.pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[0].minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[0].minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[0].maxValue.value").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[0].maxValue.pulseId").value(109))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[0].maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 90000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[0].maxValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[1].minValue.value").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[1].minValue.pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[1].minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[1].minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[1].maxValue.value").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[1].maxValue.pulseId").value(109))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[1].maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 90000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].extrema[1].maxValue.eventCount").value(1));
   }

   @Test
   public void testPulseRangeQuery_Match_Drop() throws Exception {
      Map<String, Set<Long>> channelSeq = new HashMap<>();
      Set<Long> channel_01_Seq = new LinkedHashSet<>(Arrays.asList(0L, 2L, 4L, 5L));
      String channel_01 = "TestChannel_Sequence_" + Arrays.toString(channel_01_Seq.toArray());
      channelSeq.put(channel_01, channel_01_Seq);
      Set<Long> channel_02_Seq = new LinkedHashSet<>(Arrays.asList(1L, 3L, 5L, 6L));
      String channel_02 = "TestChannel_Sequence_" + Arrays.toString(channel_02_Seq.toArray());
      channelSeq.put(channel_02, channel_02_Seq);
      Set<Long> channel_03_Seq = new LinkedHashSet<>(Arrays.asList(0L, 1L, 2L, 3L, 4L, 5L));
      String channel_03 = "TestChannel_Sequence_" + Arrays.toString(channel_03_Seq.toArray());
      channelSeq.put(channel_03, channel_03_Seq);
      List<String> channels = Arrays.asList(channel_01, channel_02, channel_03);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  // dummy range as range is defined by channel_Seq (see
                  // DummyCassandraReader.getDummyEventStream())
                  0,
                  0),
            channels);
      request.setMapping(new Mapping(IncompleteStrategy.DROP));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(5))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(5))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].pulseId").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][3]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1]").doesNotExist());

   }

   @Test
   public void testPulseRangeQuery_Match_ProvideAsIs() throws Exception {
      Map<String, Set<Long>> channelSeq = new HashMap<>();
      Set<Long> channel_01_Seq = new LinkedHashSet<>(Arrays.asList(0L, 2L, 4L, 5L));
      String channel_01 = "TestChannel_Sequence_" + Arrays.toString(channel_01_Seq.toArray());
      channelSeq.put(channel_01, channel_01_Seq);
      Set<Long> channel_02_Seq = new LinkedHashSet<>(Arrays.asList(1L, 3L, 5L, 6L));
      String channel_02 = "TestChannel_Sequence_" + Arrays.toString(channel_02_Seq.toArray());
      channelSeq.put(channel_02, channel_02_Seq);
      Set<Long> channel_03_Seq = new LinkedHashSet<>(Arrays.asList(0L, 1L, 2L, 3L, 4L, 5L));
      String channel_03 = "TestChannel_Sequence_" + Arrays.toString(channel_03_Seq.toArray());
      channelSeq.put(channel_03, channel_03_Seq);
      List<String> channels = Arrays.asList(channel_01, channel_02, channel_03);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  // dummy range as range is defined by channel_Seq (see
                  // DummyCassandraReader.getDummyEventStream())
                  0,
                  0),
            channels);
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].pulseId").value(2))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].pulseId").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].pulseId").value(3))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].pulseId").value(3))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].pulseId").value(4))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].pulseId").value(4))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].pulseId").value(5))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].pulseId").value(5))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].pulseId").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][3]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].pulseId").value(6))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7]").doesNotExist());

   }

   @Test
   public void testPulseRangeQuery_Match_FillNull() throws Exception {
      Map<String, Set<Long>> channelSeq = new HashMap<>();
      Set<Long> channel_01_Seq = new LinkedHashSet<>(Arrays.asList(0L, 2L, 4L, 5L));
      String channel_01 = "TestChannel_Sequence_" + Arrays.toString(channel_01_Seq.toArray());
      channelSeq.put(channel_01, channel_01_Seq);
      Set<Long> channel_02_Seq = new LinkedHashSet<>(Arrays.asList(1L, 3L, 5L, 6L));
      String channel_02 = "TestChannel_Sequence_" + Arrays.toString(channel_02_Seq.toArray());
      channelSeq.put(channel_02, channel_02_Seq);
      Set<Long> channel_03_Seq = new LinkedHashSet<>(Arrays.asList(0L, 1L, 2L, 3L, 4L, 5L));
      String channel_03 = "TestChannel_Sequence_" + Arrays.toString(channel_03_Seq.toArray());
      channelSeq.put(channel_03, channel_03_Seq);
      List<String> channels = Arrays.asList(channel_01, channel_02, channel_03);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  // dummy range as range is defined by channel_Seq (see
                  // DummyCassandraReader.getDummyEventStream())
                  0,
                  0),
            channels);
      request.setMapping(new Mapping(IncompleteStrategy.FILL_NULL));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2].pulseId").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][3]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2].pulseId").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][3]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].pulseId").value(2))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][2]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][2].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][2].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][2].pulseId").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][3]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].pulseId").value(3))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][2]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][2].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][2].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][2].pulseId").value(3))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][3]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].pulseId").value(4))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][2]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][2].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][2].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][2].pulseId").value(4))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][3]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].pulseId").value(5))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].pulseId").value(5))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].pulseId").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][3]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1].pulseId").value(6))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][2]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][3]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7]").doesNotExist());

   }

   @Test
   public void testPulseRangeQuery_Match_KeepAsIs_Combine_Scalar() throws Exception {
      Map<String, Set<Long>> channelSeq = new HashMap<>();
      Set<Long> channel_01_Seq = new LinkedHashSet<>(Arrays.asList(0L, 5L, 20L, 40L, 50L));
      String channel_01 =
            "TestChannel_Sequence_"
                  + channel_01_Seq.stream().map(val -> val.toString()).collect(Collectors.joining(",", "{", "}"));
      channelSeq.put(channel_01, channel_01_Seq);
      Set<Long> channel_02_Seq = new LinkedHashSet<>(Arrays.asList(10L, 30L, 50L, 55L, 60L, 65L));
      String channel_02 =
            "TestChannel_Sequence_"
                  + channel_02_Seq.stream().map(val -> val.toString()).collect(Collectors.joining(",", "{", "}"));
      channelSeq.put(channel_02, channel_02_Seq);
      Set<Long> channel_03_Seq = new LinkedHashSet<>(Arrays.asList(0L, 10L, 20L, 30L, 40L, 50L));
      String channel_03 =
            "TestChannel_Sequence_"
                  + channel_03_Seq.stream().map(val -> val.toString()).collect(Collectors.joining(",", "{", "}"));
      channelSeq.put(channel_03, channel_03_Seq);
      List<String> channels = Arrays.asList(channel_01, channel_02, channel_03);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  // dummy range as range is defined by channel_Seq (see
                  // TestReader.getDummyEventStream())
                  0,
                  0),
            channels);
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].eventCount").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.min").value(0.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.mean").value(0.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value.max").value(0.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].pulseId").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].value").value(2))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].pulseId").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].value").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].pulseId").value(3))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].value").value(3))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].pulseId").value(3))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].value").value(3))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].pulseId").value(4))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].value").value(4))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].pulseId").value(4))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].value").value(4))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].pulseId").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].value").value(5))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].pulseId").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].eventCount").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value.min").value(5.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value.mean").value(5.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value.max").value(5.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].pulseId").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].value").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][3]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].pulseId").value(6))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].eventCount").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value.min").value(6.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value.mean").value(6.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value.max").value(6.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7]").doesNotExist());

   }

   @Test
   public void testPulseRangeQuery_Match_KeepAsIs_Combine_Waveform() throws Exception {
      Map<String, Set<Long>> channelSeq = new HashMap<>();
      Set<Long> channel_01_Seq = new LinkedHashSet<>(Arrays.asList(0L, 5L, 20L, 40L, 50L));
      String channel_01 =
            "TestChannel_Waveform_Sequence_"
                  + channel_01_Seq.stream().map(val -> val.toString()).collect(Collectors.joining(",", "{", "}"));
      channelSeq.put(channel_01, channel_01_Seq);
      Set<Long> channel_02_Seq = new LinkedHashSet<>(Arrays.asList(10L, 30L, 50L, 55L, 60L, 65L));
      String channel_02 =
            "TestChannel_Waveform_Sequence_"
                  + channel_02_Seq.stream().map(val -> val.toString()).collect(Collectors.joining(",", "{", "}"));
      channelSeq.put(channel_02, channel_02_Seq);
      Set<Long> channel_03_Seq = new LinkedHashSet<>(Arrays.asList(0L, 10L, 20L, 30L, 40L, 50L));
      String channel_03 =
            "TestChannel_Waveform_Sequence_"
                  + channel_03_Seq.stream().map(val -> val.toString()).collect(Collectors.joining(",", "{", "}"));
      channelSeq.put(channel_03, channel_03_Seq);
      List<String> channels = Arrays.asList(channel_01, channel_02, channel_03);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  // dummy range as range is defined by channel_Seq (see
                  // TesTetReader.getDummyEventStream())
                  0,
                  0),
            channels);
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].pulseId").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].eventCount").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[0].min").value(0.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[0].mean").value(0.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[0].max").value(0.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[1].min").value(0.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[1].mean").value(0.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][0].value[1].max").value(0.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].pulseId").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[0]").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][1].value[1]").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[0][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].pulseId").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[0]").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][0].value[1]").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].pulseId").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value[0]").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][1].value[1]").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[1][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].pulseId").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].value[0]").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][0].value[1]").value(2))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].pulseId").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].value[0]").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][1].value[1]").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[2][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].pulseId").value(3))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].value[0]").value(3))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][0].value[1]").value(3))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].pulseId").value(3))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].value[0]").value(3))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][1].value[1]").value(3))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[3][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].pulseId").value(4))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].value[0]").value(4))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][0].value[1]").value(4))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].pulseId").value(4))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].value[0]").value(4))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][1].value[1]").value(4))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[4][2]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].channel").value(channel_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].pulseId").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].value[0]").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][0].value[1]").value(5))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].pulseId").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].eventCount").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value[0].min").value(5.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value[0].mean").value(5.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value[0].max").value(5.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value[1].min").value(5.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value[1].mean").value(5.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][1].value[1].max").value(5.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].channel").value(channel_03))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].pulseId").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].value[0]").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][2].value[1]").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[5][3]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0]").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].channel").value(channel_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].pulseId").value(6))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].eventCount").value(2))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value[0].min").value(6.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value[0].mean").value(6.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value[0].max").value(6.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value[1].min").value(6.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value[1].mean").value(6.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][0].value[1].max").value(6.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$.data[6][1]").doesNotExist())

            .andExpect(MockMvcResultMatchers.jsonPath("$.data[7]").doesNotExist());

   }

   @Test
   public void testValueTransformation_01() throws Exception {
      String channelName = "TestImage";
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            channelName);
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);
      request.setMapping(new Mapping());
      request.addValueTransformation(
            new ValueTransformationSequence(
                  ValueTransformationSequence.ALL_CHANNELS,
                  new ImageDownScaleValueTransformation(8,
                        new TypedColorModel(
                              ColorModelType.LINEAR,
                              64,
                              Arrays.asList(Color.BLUE, Color.RED))),
                  new ImageEncodingValueTransformation(
                        ImageFormat.PNG,
                        new Base64ImageEncoder(StandardCharsets.UTF_8, true, false))));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andReturn();
      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      String imageScheme = "data:image/png;charset=UTF-8;base64,";
      assertTrue(response.contains(imageScheme));
      Object document = Configuration.defaultConfiguration().jsonProvider().parse(response);
      String image = JsonPath.read(document, "$.data[0][0].transformedValue");
      assertTrue(image.startsWith(imageScheme));
      image = JsonPath.read(document, "$.data[1][0].transformedValue");
      assertTrue(image.startsWith(imageScheme));
   }

   @Test
   public void testValueTransformation_02() throws Exception {
      String channelName = "TestImage";
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            channelName);
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);
      request.setMapping(new Mapping());
      request.addValueTransformation(
            new ValueTransformationSequence(
                  channelName,
                  new ValueSamplerValueTransformation(2)));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andReturn();
      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      Object document = Configuration.defaultConfiguration().jsonProvider().parse(response);
      String shape = JsonPath.read(document, "$.data[0][0].transformedValue.shape").toString();
      assertEquals("[40,20]", shape);
      String value = JsonPath.read(document, "$.data[0][0].transformedValue.value").toString();
      assertTrue(value.startsWith("[100.0,"));

      shape = JsonPath.read(document, "$.data[1][0].transformedValue.shape").toString();
      assertEquals("[40,20]", shape);
      value = JsonPath.read(document, "$.data[1][0].transformedValue.value").toString();
      assertTrue(value.startsWith("[101.0,"));
   }

   @Test
   public void testValueTransformation_03() throws Exception {
      String channelName = "TestImage";
      String channelName2 = "Test2Image";
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            channelName, channelName2);
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);
      request.setMapping(new Mapping());
      request.addValueTransformation(
            new ValueTransformationSequence(
                  null,
                  new ImageDownScaleValueTransformation(8,
                        new TypedColorModel(
                              ColorModelType.LINEAR,
                              64,
                              Arrays.asList(Color.BLUE, Color.RED))),
                  new ImageEncodingValueTransformation(
                        ImageFormat.PNG,
                        new Base64ImageEncoder(StandardCharsets.UTF_8, true, false))));
      request.addValueTransformation(
            new ValueTransformationSequence(
                  "Test2",
                  new ValueSamplerValueTransformation(2)));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andReturn();
      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      String imageScheme = "data:image/png;charset=UTF-8;base64,";
      assertTrue(response.contains(imageScheme));
      Object document = Configuration.defaultConfiguration().jsonProvider().parse(response);
      String image = JsonPath.read(document, "$.data[0][0].transformedValue");
      assertTrue(image.startsWith(imageScheme));
      image = JsonPath.read(document, "$.data[1][0].transformedValue");
      assertTrue(image.startsWith(imageScheme));

      String shape = JsonPath.read(document, "$.data[0][1].transformedValue.shape").toString();
      assertEquals("[40,20]", shape);
      String value = JsonPath.read(document, "$.data[0][1].transformedValue.value").toString();
      assertTrue(value.startsWith("[100.0,"));

      shape = JsonPath.read(document, "$.data[1][1].transformedValue.shape").toString();
      assertEquals("[40,20]", shape);
      value = JsonPath.read(document, "$.data[1][1].transformedValue.value").toString();
      assertTrue(value.startsWith("[101.0,"));
   }
}
