package ch.psi.daq.test.queryrest.controller;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.awt.Color;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;

import org.junit.After;
import org.junit.Test;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.result.MockMvcResultHandlers;
import org.springframework.test.web.servlet.result.MockMvcResultMatchers;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.jayway.jsonpath.Configuration;
import com.jayway.jsonpath.JsonPath;

import ch.psi.bsread.message.Type;
import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.common.time.TimeUtils;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.channels.ChannelsRequest;
import ch.psi.daq.domain.query.channels.LongHash;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.AggregationDescriptor;
import ch.psi.daq.domain.query.operation.AggregationType;
import ch.psi.daq.domain.query.operation.Compression;
import ch.psi.daq.domain.query.operation.ConfigField;
import ch.psi.daq.domain.query.operation.Extrema;
import ch.psi.daq.domain.query.operation.EventField;
import ch.psi.daq.domain.query.transform.ValueTransformationSequence;
import ch.psi.daq.domain.query.transform.image.ImageDownScaleValueTransformation;
import ch.psi.daq.domain.query.transform.image.ImageEncodingValueTransformation;
import ch.psi.daq.domain.query.transform.image.ImageFormat;
import ch.psi.daq.domain.query.transform.image.color.ColorModelType;
import ch.psi.daq.domain.query.transform.image.color.TypedColorModel;
import ch.psi.daq.domain.query.transform.image.encoding.Base64ImageEncoder;
import ch.psi.daq.domain.query.transform.sampling.ValueSamplerValueTransformation;
import ch.psi.daq.domain.request.range.RequestRangeDate;
import ch.psi.daq.domain.request.range.RequestRangeEndDate;
import ch.psi.daq.domain.request.range.RequestRangeEndPulseId;
import ch.psi.daq.domain.request.range.RequestRangeEndTime;
import ch.psi.daq.domain.request.range.RequestRangePulseId;
import ch.psi.daq.domain.request.range.RequestRangeStartDate;
import ch.psi.daq.domain.request.range.RequestRangeStartPulseId;
import ch.psi.daq.domain.request.range.RequestRangeStartTime;
import ch.psi.daq.domain.request.range.RequestRangeTime;
import ch.psi.daq.domain.test.TestTimeUtils;
import ch.psi.daq.queryrest.response.json.JSONHTTPResponse;
import ch.psi.daq.test.queryrest.AbstractDaqRestTest;

/**
 * Tests the {@link DaqController} implementation.
 */
public class JsonQueryRestControllerTest extends AbstractDaqRestTest implements ApplicationContextAware {

   public static final String TEST_CHANNEL_01 = "testChannel1";
   public static final String TEST_CHANNEL_02 = "testChannel2";
   public static final String TEST_CHANNEL_WAVEFORM_01 = "testChannelWaveform1";
   public static final String[] TEST_CHANNEL_NAMES = new String[] {TEST_CHANNEL_01, TEST_CHANNEL_02};

   private Backend backend;
   private Backend backend2;
   private Backend backend3;

   private ObjectMapper objectMapper = new ObjectMapper();

   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      backend = context.getBean(DomainConfig.BEAN_NAME_BACKEND_DEFAULT, Backend.class);
      context = backend.getApplicationContext();

      for (final Backend baknd : Backend.getBackends()) {
         if (baknd != backend) {
            if (backend2 == null) {
               backend2 = baknd;
            } else if (backend3 == null) {
               backend3 = baknd;
               break;
            }
         }
      }
   }

   @After
   public void tearDown() throws Exception {}

   @Test
   public void testChannelNameQuery() throws Exception {
      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .get(DomainConfig.PATH_CHANNELS)
                        .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[0]").value("BoolScalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[1]").value("BoolWaveform"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].backend").value(backend2.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].backend").value(backend3.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[0]").value("BoolScalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[1]").value("BoolWaveform"));

   }

   @Test
   public void testChannelsHash() throws Exception {
      MvcResult result = this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .get(DomainConfig.PATH_CHANNELS_HASH)
                        .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      LongHash hash1 =
            objectMapper.readValue(response, LongHash.class);

      result = this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .get(DomainConfig.PATH_CHANNELS_HASH)
                        .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      LongHash hash2 =
            objectMapper.readValue(response, LongHash.class);

      assertEquals(hash1.getHash(), hash2.getHash());
   }

   @Test
   public void testSpecificChannelSearch() throws Exception {
      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .get(DomainConfig.PATH_CHANNELS + "/int32")
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[0]").value("Int32Scalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[1]").value("Int32Waveform"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[2]").value("UInt32Scalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[3]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[3]").value("UInt32Waveform"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].backend").value(backend2.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].backend").value(backend3.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[0]").value("Int32Scalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[1]").value("Int32Waveform"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[2]").value("UInt32Scalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[3]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[3]").value("UInt32Waveform"));
   }

   @Test
   public void testChannelNameQueryBackendOrder() throws Exception {
      ChannelsRequest request = new ChannelsRequest("int64", Ordering.desc, backend);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_CHANNELS)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[0]").value("UInt64Waveform"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[1]").value("UInt64Scalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[2]").value("Int64Waveform"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[3]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[3]").value("Int64Scalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").doesNotExist());
   }

   @Test
   public void testChannelNameQueryReload() throws Exception {
      ChannelsRequest request = new ChannelsRequest();

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_CHANNELS)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[23]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[24]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].backend").value(backend2.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels[1]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].backend").value(backend3.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[23]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[24]").doesNotExist());

      // each reload add another channel
      request.setReload(true);

      content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_CHANNELS)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[24]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[25]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].backend").value(backend2.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].backend").value(backend3.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[24]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[2].channels[25]").doesNotExist());
   }

   @Test
   public void testCorsFilterNoHeaders() throws Exception {
      this.mockMvc = MockMvcBuilders.webAppContextSetup(webApplicationContext).build();

      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .options(DomainConfig.PATH_CHANNELS)
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            // we didn't set the 'Origin' header so no access-control
            .andExpect(MockMvcResultMatchers.header().doesNotExist("Access-Control-Allow-Origin"));
   }

   @Test
   public void testCorsFilterIncludesHeaders() throws Exception {
      // all headers are set
      this.mockMvc = MockMvcBuilders.webAppContextSetup(webApplicationContext).build();

      // curl -H "Origin: *" -H "Access-Control-Request-Method: POST" -X OPTIONS -v
      // http://localhost:8080/channels
      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .options(DomainConfig.PATH_CHANNELS)
                  .header("Origin", "*")
                  .header("Access-Control-Request-Method", "POST")
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Access-Control-Allow-Origin", "*"));

      // curl -H "Origin: http://localhost:8080" -H "Access-Control-Request-Method: POST" -X OPTIONS
      // -v http://localhost:8080/channels
      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .options(DomainConfig.PATH_CHANNELS)
                  .header("Origin", "http://localhost:8080")
                  .header("Access-Control-Request-Method", "POST")
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Access-Control-Allow-Origin", "http://localhost:8080"));
   }

   @Test
   public void testPulseRangeQuery() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            TEST_CHANNEL_NAMES);
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].iocMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].iocMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].iocMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].iocMillis").value(1010));
   }

   @Test
   public void testPulseRangeQuery_StartExpansion() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  true,
                  101, 
                  false),
            TEST_CHANNEL_NAMES);
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(99))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(0, 990000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalMillis").value(990))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(0, 990000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].iocMillis").value(990))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].iocMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].iocMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(99))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(0, 990000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalMillis").value(990))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(0, 990000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].iocMillis").value(990))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].iocMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].iocMillis").value(1010));
   }

   @Test
   public void testPulseRangeQuery_EndExpansion() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  99,
                  false,
                  100, 
                  true),
            TEST_CHANNEL_NAMES);
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(99))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(0, 990000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalMillis").value(990))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(0, 990000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].iocMillis").value(990))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].iocMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].iocMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(99))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(0, 990000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalMillis").value(990))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(0, 990000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].iocMillis").value(990))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].iocMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].iocMillis").value(1010));
   }
   
   @Test
   public void testPulseRangeQuery_Fields() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  199),
            new AggregationDescriptor().setNrOfBins(2),
            TEST_CHANNEL_NAMES);
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.eventCount);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(50))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(150))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(50))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].eventCount").value(50))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(150))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].eventCount").value(50));
   }

   @Test
   public void testPulseRangeQueryBackends() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            new ChannelName(TEST_CHANNEL_01, backend),
            new ChannelName(TEST_CHANNEL_02, backend2));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)));
   }

   @Test
   public void testPulseRangeQueries() throws Exception {
      String testChannel3 = "testChannel3";
      DAQQueries request = new DAQQueries(
            new DAQQueryElement(
                  new RequestRangePulseId(
                        100,
                        101),
                  TEST_CHANNEL_NAMES),
            new DAQQueryElement(
                  new RequestRangePulseId(
                        100,
                        101),
                  testChannel3));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERIES)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][0].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1][0].channel.name").value(testChannel3))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1][0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1][0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1][0].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1][0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)));
   }

   @Test
   public void testTimeRangeQuery_01() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(2000, 0),
                  TimeUtils.getTimeFromMillis(2010, 0)),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)));
   }
   
   @Test
   public void testTimeRangeQuery_01_StartExpansion() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(2000, 0),
                  true,
                  TimeUtils.getTimeFromMillis(2010, 0),
                  false),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(199))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 990000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(199))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 990000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)));
   }
   
   @Test
   public void testTimeRangeQuery_01_EndExpansion() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(1990, 0),
                  false,
                  TimeUtils.getTimeFromMillis(2000, 0),
                  true),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(199))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 990000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(199))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 990000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)));
   }

   @Test
   public void testTimeRangeQuery_01_ConfigFields() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(2000, 0),
                  TimeUtils.getTimeFromMillis(2010, 0)),
            TEST_CHANNEL_NAMES);
      request.addConfigField(ConfigField.pulseId);
      request.addConfigField(ConfigField.globalSeconds);
      request.addConfigField(ConfigField.globalMillis);
      request.addConfigField(ConfigField.shape);
      request.addConfigField(ConfigField.description);
      request.addConfigField(ConfigField.backend);
      request.addConfigField(ConfigField.modulo);
      request.addConfigField(ConfigField.offset);
      request.addConfigField(ConfigField.keyspace);
      request.addConfigField(ConfigField.precision);
      request.addConfigField(ConfigField.source);
      request.addConfigField(ConfigField.type);
      request.addConfigField(ConfigField.unit);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[0].globalMillis").value(2000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[0].shape").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[0].description").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[0].modulo").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[0].offset").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[0].keyspace").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[0].precision").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[0].source").value("unknown"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[0].type").value(Type.Int32.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[0].unit").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[1].globalMillis").value(2010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[1].shape").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[1].description").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[1].modulo").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[1].offset").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[1].keyspace").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[1].precision").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[1].source").value("unknown"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[1].type").value(Type.Int32.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].configs[1].unit").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[0].globalMillis").value(2000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[0].shape").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[0].description").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[0].modulo").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[0].offset").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[0].keyspace").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[0].precision").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[0].source").value("unknown"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[0].type").value(Type.Int32.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[0].unit").doesNotExist()) 
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))  
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[1].globalMillis").value(2010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[1].shape").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[1].description").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[1].modulo").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[1].offset").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[1].keyspace").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[1].precision").value(0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[1].source").value("unknown"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[1].type").value(Type.Int32.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].configs[1].unit").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)));
   }

   @Test
   public void testTimeRangeQuery_02() throws Exception {
      String content =
            "{\"channels\":[{\"name\":\"testChannel1\"},{\"name\":\"testChannel2\"}],\"fields\":[],\"ordering\":\"asc\",\"range\":{\"startSeconds\":\"2.0\",\"endSeconds\":\"2.01\"}}";

      MvcResult result = this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)));
   }

   @Test
   public void testDateRangeQuery() throws Exception {
      String startDate = TimeUtils.format(1000);
      String endDate = TimeUtils.format(1010);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)));
   }

   @Test
   public void testOpenTimeRangeQueryStart_01() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeStartTime(
                  TimeUtils.getTimeFromMillis(2000, 0)),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2]").doesNotExist());
   }

   @Test
   public void testOpenTimeRangeQueryStart_01_Exclusive() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeStartTime(
                  TimeUtils.getTimeFromMillis(2000, 0),
                  false),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(202))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 20000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(202))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 20000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2]").doesNotExist());
   }

   @Test
   public void testOpenTimeRangeQueryStartDate_01_Exclusive() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeStartDate(
                  TimeUtils.format(TimeUtils.getTimeFromMillis(2000, 0)),
                  false),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(202))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 20000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(202))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 20000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2]").doesNotExist());
   }

   @Test
   public void testOpenTimeRangeQueryStart_02() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeStartTime(
                  TimeUtils.getTimeFromMillis(2000, 0)),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);

      try {
         // limit needs to be defined
         this.mockMvc.perform(MockMvcRequestBuilders
               .post(DomainConfig.PATH_QUERY)
               .contentType(MediaType.APPLICATION_JSON)
               .content(content))

               .andDo(MockMvcResultHandlers.print())
               .andExpect(MockMvcResultMatchers.status().isOk());
         assertTrue(false);
      } catch (Exception e) {
         assertTrue(true);
      }
   }

   @Test
   public void testOpenTimeRangeQueryStart_03() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeStartTime(
                  TimeUtils.getTimeFromMillis(2000, 0)),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);
      request.setOrdering(Ordering.desc);

      String content = mapper.writeValueAsString(request);

      try {
         // asc needs to be defined
         this.mockMvc.perform(MockMvcRequestBuilders
               .post(DomainConfig.PATH_QUERY)
               .contentType(MediaType.APPLICATION_JSON)
               .content(content))

               .andDo(MockMvcResultHandlers.print())
               .andExpect(MockMvcResultMatchers.status().isOk());
         assertTrue(false);
      } catch (Exception e) {
         assertTrue(true);
      }
   }

   @Test
   public void testOpenTimeRangeQueryEnd_01() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeEndTime(
                  TimeUtils.getTimeFromMillis(2010, 0)),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);
      request.setOrdering(Ordering.desc);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2]").doesNotExist());
   }

   @Test
   public void testOpenTimeRangeQueryEnd_01_Exclusive() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeEndTime(
                  TimeUtils.getTimeFromMillis(2020, 0),
                  false),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);
      request.setOrdering(Ordering.desc);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2]").doesNotExist());
   }

   @Test
   public void testOpenTimeRangeQueryEndDate_01_Exclusive() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeEndDate(
                  TimeUtils.format(TimeUtils.getTimeFromMillis(2020, 0)),
                  false),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);
      request.setOrdering(Ordering.desc);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2]").doesNotExist());
   }

   @Test
   public void testOpenTimeRangeQueryEnd_02() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeEndTime(
                  TimeUtils.getTimeFromMillis(2010, 0)),
            TEST_CHANNEL_NAMES);
      request.setOrdering(Ordering.desc);

      String content = mapper.writeValueAsString(request);

      try {
         // limit needs to be defined
         this.mockMvc.perform(MockMvcRequestBuilders
               .post(DomainConfig.PATH_QUERY)
               .contentType(MediaType.APPLICATION_JSON)
               .content(content))

               .andDo(MockMvcResultHandlers.print())
               .andExpect(MockMvcResultMatchers.status().isOk());
         assertTrue(false);
      } catch (Exception e) {
         assertTrue(true);
      }
   }

   @Test
   public void testOpenTimeRangeQueryEnd_03() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeEndTime(
                  TimeUtils.getTimeFromMillis(2010, 0)),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);
      request.setOrdering(Ordering.asc);

      String content = mapper.writeValueAsString(request);

      try {
         // desc needs to be defined
         this.mockMvc.perform(MockMvcRequestBuilders
               .post(DomainConfig.PATH_QUERY)
               .contentType(MediaType.APPLICATION_JSON)
               .content(content))

               .andDo(MockMvcResultHandlers.print())
               .andExpect(MockMvcResultMatchers.status().isOk());
         assertTrue(false);
      } catch (Exception e) {
         assertTrue(true);
      }
   }

   @Test
   public void testOpenPulseRangeQueryStart_01() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeStartPulseId(
                  200),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2]").doesNotExist());
   }

   @Test
   public void testOpenPulseRangeQueryStart_01_Exclusive() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeStartPulseId(
                  200,
                  false),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(202))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 20000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(202))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 20000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2]").doesNotExist());
   }

   @Test
   public void testOpenPulseRangeQueryStart_02() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeStartPulseId(
                  200),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);

      try {
         // limit needs to be defined
         this.mockMvc.perform(MockMvcRequestBuilders
               .post(DomainConfig.PATH_QUERY)
               .contentType(MediaType.APPLICATION_JSON)
               .content(content))

               .andDo(MockMvcResultHandlers.print())
               .andExpect(MockMvcResultMatchers.status().isOk());
         assertTrue(false);
      } catch (Exception e) {
         assertTrue(true);
      }
   }

   @Test
   public void testOpenPulseRangeQueryStart_03() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeStartPulseId(
                  200),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);
      request.setOrdering(Ordering.desc);

      String content = mapper.writeValueAsString(request);

      try {
         // asc needs to be defined
         this.mockMvc.perform(MockMvcRequestBuilders
               .post(DomainConfig.PATH_QUERY)
               .contentType(MediaType.APPLICATION_JSON)
               .content(content))

               .andDo(MockMvcResultHandlers.print())
               .andExpect(MockMvcResultMatchers.status().isOk());
         assertTrue(false);
      } catch (Exception e) {
         assertTrue(true);
      }
   }

   @Test
   public void testOpenPulseRangeQueryEnd_01() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeEndPulseId(
                  201),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);
      request.setOrdering(Ordering.desc);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2]").doesNotExist());
   }

   @Test
   public void testOpenPulseRangeQueryEnd_01_Exclusive() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeEndPulseId(
                  202,
                  false),
            TEST_CHANNEL_NAMES);
      request.setLimit(2);
      request.setOrdering(Ordering.desc);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(DomainConfig.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2]").doesNotExist())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[2]").doesNotExist());
   }

   @Test
   public void testOpenPulseRangeQueryEnd_02() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeEndPulseId(
                  201),
            TEST_CHANNEL_NAMES);
      request.setOrdering(Ordering.desc);

      String content = mapper.writeValueAsString(request);

      try {
         // limit needs to be defined
         this.mockMvc.perform(MockMvcRequestBuilders
               .post(DomainConfig.PATH_QUERY)
               .contentType(MediaType.APPLICATION_JSON)
               .content(content))

               .andDo(MockMvcResultHandlers.print())
               .andExpect(MockMvcResultMatchers.status().isOk());
         assertTrue(false);
      } catch (Exception e) {
         assertTrue(true);
      }
   }

   @Test
   public void testOpenPulseRangeQueryEnd_03() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeEndPulseId(
                  201),
            TEST_CHANNEL_NAMES);
      request.setOrdering(Ordering.asc);
      request.setLimit(2);

      String content = mapper.writeValueAsString(request);

      try {
         // desc needs to be defined
         this.mockMvc.perform(MockMvcRequestBuilders
               .post(DomainConfig.PATH_QUERY)
               .contentType(MediaType.APPLICATION_JSON)
               .content(content))

               .andDo(MockMvcResultHandlers.print())
               .andExpect(MockMvcResultMatchers.status().isOk());
         assertTrue(false);
      } catch (Exception e) {
         assertTrue(true);
      }
   }

   @Test
   public void testExtremaAggregation() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            Ordering.asc,
            new AggregationDescriptor(AggregationType.extrema),
            TEST_CHANNEL_NAMES[0]);

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders.post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())

            .andExpect(MockMvcResultMatchers.jsonPath("$").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_NAMES[0]))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.value").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.value").value(Double.valueOf(100)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.event").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.event.pulseId").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.event.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.value").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.value").value(Double.valueOf(101)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.event").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.event.pulseId").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.event.pulseId").value(101));
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregate() throws Exception {
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor().setNrOfBins(2).setAggregations(
                        Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max)),
                  TEST_CHANNEL_01);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.max").value(104.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.max").value(109.0));
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregate_Bool() throws Exception {
      String channelName = Type.Bool.getKey() + TEST_CHANNEL_01;
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor().setNrOfBins(2).setAggregations(
                        Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max)),
                  channelName);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(channelName))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.min").isNumber())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.mean").isNumber())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.max").isNumber())

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.min").isNumber())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.mean").isNumber())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.max").isNumber());
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregateExtrema() throws Exception {
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor()
                        .setNrOfBins(2)
                        .setAggregations(Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max))
                        .setExtrema(Arrays.asList(Extrema.minValue, Extrema.maxValue)),
                  TEST_CHANNEL_01);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.minValue.value").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.minValue.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.maxValue.value").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.maxValue.pulseId").value(104))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 40000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.maxValue.eventCount").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.minValue.value").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.minValue.pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.maxValue.value").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.maxValue.pulseId").value(109))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 90000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.maxValue.eventCount").value(1));
   }

   @Test
   public void testDateRangeQueryBinSizeAggregate() throws Exception {
      long startTime = 10000;
      long endTime = 10999;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            new AggregationDescriptor().setDurationPerBin(100),
            TEST_CHANNEL_01);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.min").value(1000.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.mean").value(1004.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.max").value(1009.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 100000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.min").value(1010.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.mean").value(1014.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.max").value(1019.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].pulseId").value(1020))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 200000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].value.min").value(1020.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].value.mean").value(1024.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].value.max").value(1029.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].pulseId").value(1030))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 300000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].value.min").value(1030.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].value.mean").value(1034.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].value.max").value(1039.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].pulseId").value(1040))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 400000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].value.min").value(1040.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].value.mean").value(1044.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].value.max").value(1049.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].pulseId").value(1050))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 500000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].value.min").value(1050.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].value.mean").value(1054.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].value.max").value(1059.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].pulseId").value(1060))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 600000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].value.min").value(1060.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].value.mean").value(1064.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].value.max").value(1069.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].pulseId").value(1070))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 700000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].value.min").value(1070.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].value.mean").value(1074.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].value.max").value(1079.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].pulseId").value(1080))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 800000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].value.min").value(1080.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].value.mean").value(1084.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].value.max").value(1089.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].pulseId").value(1090))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 900000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].value.min").value(1090.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].value.mean").value(1094.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].value.max").value(1099.0));
   }

   @Test
   public void testDateRangeQueryIndexAggregate() throws Exception {
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor()
                        .setAggregationType(AggregationType.index)
                        .setNrOfBins(2)
                        .setAggregations(Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max)),
                  TEST_CHANNEL_WAVEFORM_01);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_WAVEFORM_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[0].min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[0].mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[0].max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[1].min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[1].mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[1].max").value(104.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[0].min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[0].mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[0].max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[1].min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[1].mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[1].max").value(109.0));
   }

   @Test
   public void testDateRangeQueryIndexAggregateExtrema() throws Exception {
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor()
                        .setAggregationType(AggregationType.index)
                        .setNrOfBins(2)
                        .setAggregations(Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max))
                        .setExtrema(Arrays.asList(Extrema.minValue, Extrema.maxValue)),
                  TEST_CHANNEL_WAVEFORM_01);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(DomainConfig.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_WAVEFORM_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(backend.getName()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[0].min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[0].mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[0].max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[1].min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[1].mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[1].max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].minValue.value").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].minValue.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].maxValue.value").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].maxValue.pulseId").value(104))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 40000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].maxValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].minValue.value").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].minValue.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].maxValue.value").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].maxValue.pulseId").value(104))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 40000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].maxValue.eventCount").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[0].min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[0].mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[0].max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[1].min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[1].mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[1].max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].minValue.value").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].minValue.pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].maxValue.value").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].maxValue.pulseId").value(109))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 90000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].maxValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].minValue.value").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].minValue.pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].maxValue.value").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].maxValue.pulseId").value(109))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 90000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].maxValue.eventCount").value(1));
   }

   @Test
   public void testGzipFileSuffixHeader() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);
      request.setResponse(new JSONHTTPResponse(Compression.GZIP));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Content-Disposition", "attachment; filename=data.json"));
   }

   @Test
   public void testJsonFileSuffixHeader() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Content-Disposition", "attachment; filename=data.json"));
   }

   @Test
   public void testBadAggregation_01() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            new AggregationDescriptor().setDurationPerBin(1000),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isBadRequest());
   }

   @Test
   public void testBadAggregation_02() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(0, 0),
                  TimeUtils.getTimeFromMillis(10, 0)),
            new AggregationDescriptor().setPulsesPerBin(100),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isBadRequest());
   }

   @Test
   public void testBadAggregation_03() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            new AggregationDescriptor().setDurationPerBin(1000).setNrOfBins(100),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isBadRequest());
   }

   @Test
   public void testValueTransformation_01() throws Exception {
      String channelName = "TestImage";
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            channelName);
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);
      request.addValueTransformation(
            new ValueTransformationSequence(
                  ValueTransformationSequence.ALL_CHANNELS,
                  new ImageDownScaleValueTransformation(8,
                        new TypedColorModel(
                              ColorModelType.LINEAR,
                              64,
                              Arrays.asList(Color.BLUE, Color.RED))),
                  new ImageEncodingValueTransformation(
                        ImageFormat.PNG,
                        new Base64ImageEncoder(StandardCharsets.UTF_8, true, false))));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andReturn();
      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      String imageScheme = "data:image/png;charset=UTF-8;base64,";
      assertTrue(response.contains(imageScheme));
      Object document = Configuration.defaultConfiguration().jsonProvider().parse(response);
      String image = JsonPath.read(document, "$[0].data[0].transformedValue");
      assertTrue(image.startsWith(imageScheme));
      image = JsonPath.read(document, "$[0].data[1].transformedValue");
      assertTrue(image.startsWith(imageScheme));
   }

   @Test
   public void testValueTransformation_02() throws Exception {
      String channelName = "TestImage";
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            channelName);
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);
      request.addValueTransformation(
            new ValueTransformationSequence(
                  channelName,
                  new ValueSamplerValueTransformation(2)));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andReturn();
      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      Object document = Configuration.defaultConfiguration().jsonProvider().parse(response);
      String shape = JsonPath.read(document, "$[0].data[0].transformedValue.shape").toString();
      assertEquals("[40,20]", shape);
      String value = JsonPath.read(document, "$[0].data[0].transformedValue.value").toString();
      assertTrue(value.startsWith("[100.0,"));

      shape = JsonPath.read(document, "$[0].data[1].transformedValue.shape").toString();
      assertEquals("[40,20]", shape);
      value = JsonPath.read(document, "$[0].data[1].transformedValue.value").toString();
      assertTrue(value.startsWith("[101.0,"));
   }

   @Test
   public void testValueTransformation_03() throws Exception {
      String channelName = "TestImage";
      String channelName2 = "Test2Image";
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            channelName, channelName2);
      request.addEventField(EventField.pulseId);
      request.addEventField(EventField.globalSeconds);
      request.addEventField(EventField.globalMillis);
      request.addEventField(EventField.iocSeconds);
      request.addEventField(EventField.iocMillis);
      request.addValueTransformation(
            new ValueTransformationSequence(
                  null,
                  new ImageDownScaleValueTransformation(8, new TypedColorModel(
                        ColorModelType.LINEAR,
                        64,
                        Arrays.asList(Color.BLUE, Color.RED))),
                  new ImageEncodingValueTransformation(
                        ImageFormat.PNG,
                        new Base64ImageEncoder(StandardCharsets.UTF_8, true, false))));
      request.addValueTransformation(
            new ValueTransformationSequence(
                  "Test2",
                  new ValueSamplerValueTransformation(2)));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andReturn();
      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      String imageScheme = "data:image/png;charset=UTF-8;base64,";
      assertTrue(response.contains(imageScheme));
      Object document = Configuration.defaultConfiguration().jsonProvider().parse(response);
      String image = JsonPath.read(document, "$[0].data[0].transformedValue");
      assertTrue(image.startsWith(imageScheme));
      image = JsonPath.read(document, "$[0].data[1].transformedValue");
      assertTrue(image.startsWith(imageScheme));

      String shape = JsonPath.read(document, "$[1].data[0].transformedValue.shape").toString();
      assertEquals("[40,20]", shape);
      String value = JsonPath.read(document, "$[1].data[0].transformedValue.value").toString();
      assertTrue(value.startsWith("[100.0,"));

      shape = JsonPath.read(document, "$[1].data[1].transformedValue.shape").toString();
      assertEquals("[40,20]", shape);
      value = JsonPath.read(document, "$[1].data[1].transformedValue.value").toString();
      assertTrue(value.startsWith("[101.0,"));
   }
}
