package ch.psi.daq.test.queryrest.controller;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.function.Function;

import org.junit.After;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.result.MockMvcResultHandlers;
import org.springframework.test.web.servlet.result.MockMvcResultMatchers;

import com.fasterxml.jackson.databind.ObjectMapper;

import ch.psi.bsread.common.allocator.ByteBufferAllocator;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.mapping.Mapping;
import ch.psi.daq.domain.request.range.RequestRangePulseId;
import ch.psi.daq.queryrest.response.raw.RawDeserializerProvider;
import ch.psi.daq.queryrest.response.raw.event.RawEventHTTPResponse;
import ch.psi.daq.queryrest.response.raw.event.serialize.EventDeSerializer;
import ch.psi.daq.queryrest.response.raw.event.serialize.EventHeader;
import ch.psi.daq.queryrest.response.raw.event.serialize.EventHeaderDeSerializer;
import ch.psi.daq.queryrest.response.raw.event.serialize.StandardEventDeSerializer;
import ch.psi.daq.queryrest.response.raw.receive.IReceiver;
import ch.psi.daq.queryrest.response.raw.receive.InputStreamMessageExtractor;
import ch.psi.daq.queryrest.response.raw.receive.Message;
import ch.psi.daq.queryrest.response.raw.receive.Receiver;
import ch.psi.daq.queryrest.response.raw.receive.ReceiverConfig;
import ch.psi.daq.queryrest.response.raw.receive.ReceiverState;
import ch.psi.daq.queryrest.response.raw.serialize.RawDeserializer;
import ch.psi.daq.test.queryrest.AbstractDaqRestTest;

public class RawEventQueryRestControllerTest extends AbstractDaqRestTest implements ApplicationContextAware {
   private static final Logger LOGGER = LoggerFactory.getLogger(RawEventQueryRestControllerTest.class);

   public static final String TEST_CHANNEL = "testChannelBig";
   public static final String TEST_CHANNEL_01 = TEST_CHANNEL + "1";
   public static final String TEST_CHANNEL_02 = TEST_CHANNEL + "2";
   public static final String[] TEST_CHANNEL_NAMES = new String[] {TEST_CHANNEL_01, TEST_CHANNEL_02};

   private Backend backend;
   private ObjectMapper mapper;

   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      backend = context.getBean(DomainConfig.BEAN_NAME_BACKEND_DEFAULT, Backend.class);
      context = backend.getApplicationContext();
      mapper = context.getBean(DomainConfig.BEAN_NAME_OBJECT_MAPPER, ObjectMapper.class);
   }

   @After
   public void tearDown() throws Exception {}

   @Test
   public void testPulseRangeQuery_01() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            channels);
      request.setResponse(new RawEventHTTPResponse());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(TEST_CHANNEL_01, event.getChannel());
      assertEquals(Integer.valueOf(0), event.getValue(Integer.class));

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(TEST_CHANNEL_01, event.getChannel());
      assertEquals(Integer.valueOf(1), event.getValue(Integer.class));

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      event = message.getValues().values().iterator().next();
      assertEquals(TEST_CHANNEL_02, event.getChannel());
      assertEquals(Integer.valueOf(0), event.getValue(Integer.class));

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(TEST_CHANNEL_02, event.getChannel());
      assertEquals(Integer.valueOf(1), event.getValue(Integer.class));

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_02() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  3),
            channels);
      request.setResponse(new RawEventHTTPResponse());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(TEST_CHANNEL_01, event.getChannel());
      assertEquals(Integer.valueOf(0), event.getValue(Integer.class));

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(TEST_CHANNEL_01, event.getChannel());
      assertEquals(Integer.valueOf(1), event.getValue(Integer.class));

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(TEST_CHANNEL_01, event.getChannel());
      assertEquals(Integer.valueOf(2), event.getValue(Integer.class));

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(TEST_CHANNEL_01, event.getChannel());
      assertEquals(Integer.valueOf(3), event.getValue(Integer.class));

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(TEST_CHANNEL_02, event.getChannel());
      assertEquals(Integer.valueOf(0), event.getValue(Integer.class));

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(TEST_CHANNEL_02, event.getChannel());
      assertEquals(Integer.valueOf(1), event.getValue(Integer.class));

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(TEST_CHANNEL_02, event.getChannel());
      assertEquals(Integer.valueOf(2), event.getValue(Integer.class));

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(TEST_CHANNEL_02, event.getChannel());
      assertEquals(Integer.valueOf(3), event.getValue(Integer.class));

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_02_Encoding() throws Exception {
      String channel1 = "TestChannelLittle1";
      String channel2 = "TestChannelLittle2";
      List<String> channels = Arrays.asList(channel1, channel2);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  3),
            channels);
      request.setResponse(new RawEventHTTPResponse());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(Integer.valueOf(0), event.getValue(Integer.class));

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(Integer.valueOf(1), event.getValue(Integer.class));

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(Integer.valueOf(2), event.getValue(Integer.class));

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(Integer.valueOf(3), event.getValue(Integer.class));

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(Integer.valueOf(0), event.getValue(Integer.class));

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(Integer.valueOf(1), event.getValue(Integer.class));

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(Integer.valueOf(2), event.getValue(Integer.class));

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(Integer.valueOf(3), event.getValue(Integer.class));

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_03_Waveform() throws Exception {
      String channel1 = "TestChannelWaveform1Big";
      String channel2 = "TestChannelWaveform2Big";
      List<String> channels = Arrays.asList(channel1, channel2);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  3),
            channels);
      request.setResponse(new RawEventHTTPResponse());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_03_Waveform_Encoding() throws Exception {
      String channel1 = "TestChannelWaveform1Little";
      String channel2 = "TestChannelWaveform2Little";
      List<String> channels = Arrays.asList(channel1, channel2);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  3),
            channels);
      request.setResponse(new RawEventHTTPResponse());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_04_Image() throws Exception {
      String channel1 = "TestChannelImage1Big";
      String channel2 = "TestChannelImage2Big";
      List<String> channels = Arrays.asList(channel1, channel2);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  3),
            channels);
      request.setResponse(new RawEventHTTPResponse());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_04_Image_Encoding() throws Exception {
      String channel1 = "TestChannelImage1Little";
      String channel2 = "TestChannelImage2Big";
      List<String> channels = Arrays.asList(channel1, channel2);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  3),
            channels);
      request.setResponse(new RawEventHTTPResponse());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_05_Sequence() throws Exception {
      String channel1 = "TestChannel1WaveformBig_[0,1,3,5]";
      String channel2 = "TestChannel2ImageBig_[1,2,4,5,6]";
      List<String> channels = Arrays.asList(channel1, channel2);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  6),
            channels);
      request.setResponse(new RawEventHTTPResponse());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(5, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(5, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(4, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(4, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(5, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(5, event.getValue(int[].class)[0]);

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(6, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(6, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_06_Sequence() throws Exception {
      String channel1 = "TestChannel1WaveformBig_[1,3,6]";
      String channel2 = "TestChannel2ImageBig_[0,2,3,5]";
      List<String> channels = Arrays.asList(channel1, channel2);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  6),
            channels);
      request.setResponse(new RawEventHTTPResponse());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(6, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(6, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(5, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(5, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_01_Table() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            channels);
      request.setResponse(new RawEventHTTPResponse());
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(TEST_CHANNEL_01, event.getChannel());
      assertEquals(Integer.valueOf(0), event.getValue(Integer.class));

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(TEST_CHANNEL_02, event.getChannel());
      assertEquals(Integer.valueOf(0), event.getValue(Integer.class));

      assertFalse(eventIter.hasNext());

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(TEST_CHANNEL_01, event.getChannel());
      assertEquals(Integer.valueOf(1), event.getValue(Integer.class));

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(TEST_CHANNEL_02, event.getChannel());
      assertEquals(Integer.valueOf(1), event.getValue(Integer.class));

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_02_Table() throws Exception {
      List<String> channels = Arrays.asList(TEST_CHANNEL_01, TEST_CHANNEL_02);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  3),
            channels);
      request.setResponse(new RawEventHTTPResponse());
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(TEST_CHANNEL_01, event.getChannel());
      assertEquals(Integer.valueOf(0), event.getValue(Integer.class));

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(TEST_CHANNEL_02, event.getChannel());
      assertEquals(Integer.valueOf(0), event.getValue(Integer.class));

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(TEST_CHANNEL_01, event.getChannel());
      assertEquals(Integer.valueOf(1), event.getValue(Integer.class));

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(TEST_CHANNEL_02, event.getChannel());
      assertEquals(Integer.valueOf(1), event.getValue(Integer.class));

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(TEST_CHANNEL_01, event.getChannel());
      assertEquals(Integer.valueOf(2), event.getValue(Integer.class));

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(TEST_CHANNEL_02, event.getChannel());
      assertEquals(Integer.valueOf(2), event.getValue(Integer.class));

      assertFalse(eventIter.hasNext());

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(TEST_CHANNEL_01, event.getChannel());
      assertEquals(Integer.valueOf(3), event.getValue(Integer.class));

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(TEST_CHANNEL_02, event.getChannel());
      assertEquals(Integer.valueOf(3), event.getValue(Integer.class));

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_02_Table_Encoding() throws Exception {
      String channel1 = "TestChannel1Big";
      String channel2 = "TestChannel2Little";
      List<String> channels = Arrays.asList(channel1, channel2);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  3),
            channels);
      request.setResponse(new RawEventHTTPResponse());
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(Integer.valueOf(0), event.getValue(Integer.class));

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(Integer.valueOf(0), event.getValue(Integer.class));

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(Integer.valueOf(1), event.getValue(Integer.class));

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(Integer.valueOf(1), event.getValue(Integer.class));

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(Integer.valueOf(2), event.getValue(Integer.class));

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(Integer.valueOf(2), event.getValue(Integer.class));

      assertFalse(eventIter.hasNext());

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(Integer.valueOf(3), event.getValue(Integer.class));

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(Integer.valueOf(3), event.getValue(Integer.class));

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_03_Table_Waveform() throws Exception {
      String channel1 = "TestChannelWaveform1Big";
      String channel2 = "TestChannelWaveform2Big";
      String channel3 = "TestChannelWaveform3Big";
      List<String> channels = Arrays.asList(channel1, channel2, channel3);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  3),
            channels);
      request.setResponse(new RawEventHTTPResponse());
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_03_Table_Waveform_Encoding() throws Exception {
      String channel1 = "TestChannelWaveform1Big";
      String channel2 = "TestChannelWaveform2Little";
      String channel3 = "TestChannelWaveform3Little";
      List<String> channels = Arrays.asList(channel1, channel2, channel3);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  3),
            channels);
      request.setResponse(new RawEventHTTPResponse());
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_04_Table_Image() throws Exception {
      String channel1 = "TestChannelImage1Big";
      String channel2 = "TestChannelImage2Big";
      String channel3 = "TestChannelImage3Big";
      List<String> channels = Arrays.asList(channel1, channel2, channel3);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  3),
            channels);
      request.setResponse(new RawEventHTTPResponse());
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_04_Table_Image_Encoding() throws Exception {
      String channel1 = "TestChannelImage1Little";
      String channel2 = "TestChannelImage2Big";
      String channel3 = "TestChannelImage3Little";
      List<String> channels = Arrays.asList(channel1, channel2, channel3);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  3),
            channels);
      request.setResponse(new RawEventHTTPResponse());
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      // latest has a different config
      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(3, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel3, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_05_Sequence_Table() throws Exception {
      String channel1 = "TestChannel1WaveformBig_[0,1,3,5]";
      String channel2 = "TestChannel2ImageBig_[1,2,4,5,6]";
      List<String> channels = Arrays.asList(channel1, channel2);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  6),
            channels);
      request.setResponse(new RawEventHTTPResponse());
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(4, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(4, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(5, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(5, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(5, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(5, event.getValue(int[].class)[0]);

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(6, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(6, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

   @Test
   public void testPulseRangeQuery_06_Sequence_Table() throws Exception {
      String channel1 = "TestChannel1WaveformBig_[1,3,6]";
      String channel2 = "TestChannel2ImageBig_[0,2,3,5]";
      List<String> channels = Arrays.asList(channel1, channel2);
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  6),
            channels);
      request.setResponse(new RawEventHTTPResponse());
      request.setMapping(new Mapping());

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(DomainConfig.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      final byte[] response = result.getResponse().getContentAsByteArray();
      final InputStream responseStream = new ByteArrayInputStream(response);

      final Function<ReceiverState<EventHeader, DataEvent>, RawDeserializer<Void, EventHeader>> configDeserializerCreator =
            state -> new EventHeaderDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR, mapper, state);
      final RawDeserializerProvider<EventHeader, DataEvent> deserializerProvider =
            new RawDeserializerProvider<>();
      deserializerProvider.add(
            new EventDeSerializer(
                  ByteBufferAllocator.DEFAULT_ALLOCATOR,
                  eventConfig -> {
                     if (eventConfig.getBlobId() == StandardEventDeSerializer.ID) {
                        return new StandardEventDeSerializer(ByteBufferAllocator.DEFAULT_ALLOCATOR);
                     } else {
                        final String message =
                              String.format("There is no blob deserializer for if '%d'.", eventConfig.getBlobId());
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }));
      final InputStreamMessageExtractor<DataEvent> messageExtractor = new InputStreamMessageExtractor<>(
            responseStream,
            configDeserializerCreator,
            deserializerProvider);
      final ReceiverConfig<EventHeader, DataEvent> receiverConfig = new ReceiverConfig<>(messageExtractor);
      final IReceiver<EventHeader, DataEvent> receiver = new Receiver<>(receiverConfig);
      receiver.connect();

      Message<EventHeader, DataEvent> message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      Iterator<DataEvent> eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      DataEvent event = eventIter.next();
      assertEquals(0, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(0, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(1, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(1, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(2, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(2, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertFalse(message.isConfigChanged());
      assertEquals(2, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(3, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(3, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(5, event.getPulseId());
      assertEquals(channel2, event.getChannel());
      assertEquals(5, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNotNull(message);
      assertTrue(message.isConfigChanged());
      assertEquals(1, message.getValues().size());
      eventIter = message.getValues().values().iterator();
      assertTrue(eventIter.hasNext());
      event = eventIter.next();
      assertEquals(6, event.getPulseId());
      assertEquals(channel1, event.getChannel());
      assertEquals(6, event.getValue(int[].class)[0]);

      assertFalse(eventIter.hasNext());

      message = receiver.receive();
      assertNull(message);

      receiver.close();
   }

//   @Test
//   public void testEventHeader() throws Exception {
//      EventHeader header = new EventHeader();
//      header.addEventSerializerId((short)0);
//      EventConfig config = new EventConfig((short) 0, "Test_1", "queryrest-1", Type.Int32);
//      header.addChannel(config);     
//      config = new EventConfig(
//            (short) 0, 
//            "Test_2", 
//            "queryrest-1", 
//            Type.Float32, 
//            new int[]{10}, 
//            10, 
//            1,
//            EventConfig.ENCODING_BIG_ENDIAN,
//            Compression.bitshuffle_lz4,
//            10000,
//            1517929910123456789L,
//            "tcp://SIN-CVME-DBPM0421:9000",
//            (short)-3, 
//            "mV",
//            "A channel in mV");
//      header.addChannel(config);
//      config = new EventConfig();
//      config.setName("Test_3");
//      config.setBackend("queryrest-1");
//      config.setBlobId((short)0);
//      header.addChannel(config);
//      
//      String headerStr = mapper.writeValueAsString(header);
//      System.out.println(headerStr);
//   }
}
