package ch.psi.daq.test.queryrest.query;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.annotation.Resource;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.springframework.context.ApplicationContext;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;

import ch.psi.bsread.message.Type;
import ch.psi.daq.common.concurrent.LoggingMapMergeFunction;
import ch.psi.daq.common.statistic.CombinableStatistics;
import ch.psi.daq.common.statistic.Statistics;
import ch.psi.daq.common.statistic.StorelessStatistics;
import ch.psi.daq.common.time.TimeUtils;
import ch.psi.daq.domain.StreamEvent;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.events.ChannelConfiguration;
import ch.psi.daq.domain.events.impl.ChannelConfigurationImpl;
import ch.psi.daq.domain.json.ChannelEventTable;
import ch.psi.daq.domain.json.ChannelEvents;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.json.Event;
import ch.psi.daq.domain.query.DAQConfigQuery;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.channels.ChannelsRequest;
import ch.psi.daq.domain.query.mapping.Mapping;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.AggregationDescriptor;
import ch.psi.daq.domain.query.operation.AggregationType;
import ch.psi.daq.domain.request.range.RequestRangeTime;
import ch.psi.daq.domain.rest.RestHelper;
import ch.psi.daq.queryrest.config.QueryRestConfig;
import ch.psi.daq.queryrest.query.QueryManager;
import ch.psi.daq.test.queryrest.AbstractDaqRestTest;
import ch.psi.data.converters.ByteConverter;

public class QueryManagerRemoteTest extends AbstractDaqRestTest {
   private final static String DATA_BUFFER = "daqlocal";// "sf-databuffer";

   @Resource
   private ApplicationContext context;
   private QueryManager queryManager;
   private Map<Backend, String> backendToServerAddresses;
   private Backend queryBackend;
   private String queryServer;

   @SuppressWarnings("unchecked")
   @Before
   public void setUp() throws Exception {
      queryManager = context.getBean(QueryRestConfig.BEAN_NAME_QUERY_MANAGER_REMOTE, QueryManager.class);
      backendToServerAddresses =
            context.getBean(QueryRestConfig.BEAN_NAME_QUERY_SERVER_BACKENDS, Map.class);
      queryBackend = backendToServerAddresses.keySet().stream()
            .filter(backend -> DATA_BUFFER.equals(backend.getName()))
            .findFirst()
            .orElse(null);
      queryServer = backendToServerAddresses.get(queryBackend);
   }

   @After
   public void tearDown() {}

   @Test
   public void testStatisticsJSON_01() throws Exception {
      final ObjectMapper objectMapper = context.getBean(DomainConfig.BEAN_NAME_OBJECT_MAPPER, ObjectMapper.class);

      final StorelessStatistics stats1 = new StorelessStatistics(3);
      stats1.add(5);
      final StorelessStatistics stats2 = new StorelessStatistics(3);
      stats2.add(5);
      stats2.add(7);

      List<Statistics> statisticsList = Arrays.asList(
            stats1,
            stats2);

      String statsStr = objectMapper.writeValueAsString(
            statisticsList
                  .stream()
                  .map(stat -> stat.getTyped()));

      List<Statistics> statisticsListDes = objectMapper.readValue(statsStr, new TypeReference<List<Statistics>>() {});
      assertEquals(statisticsList, statisticsListDes);

      List<CombinableStatistics> combStatisticsListDes =
            objectMapper.readValue(statsStr, new TypeReference<List<CombinableStatistics>>() {});
      assertEquals(statisticsList, combStatisticsListDes);
   }

   @Test
   public void testConfigQuery_01() throws Exception {
      assertNotNull(queryBackend);
      assertNotNull(queryServer);

      String channel = "SINEG01-RCIR-PUP10:SIG-AMPLT-AVG";
      DAQConfigQuery query = new DAQConfigQuery(
            new RequestRangeTime(
                  TimeUtils.parse("2016-10-12T14:00"),
                  TimeUtils.parse("2018-12-03T08:00")),
            new ChannelName(channel, queryBackend));


      List<ChannelConfiguration> configs = RestHelper.queryConfigs(context, queryServer, query)
            .stream()
            .flatMap(channelConfigs -> channelConfigs.getConfigs())
            .collect(Collectors.toList());
      assertTrue("Size was " + configs.size(), configs.size() > 0);
      assertEquals(ChannelConfigurationImpl.class, configs.get(0).getClass());
      assertEquals(channel, configs.get(0).getChannel());
      assertEquals(queryBackend, configs.get(0).getBackend());
   }

   @Test
   public void testEventQueryRaw_01() throws Exception {
      assertNotNull(queryBackend);
      assertNotNull(queryServer);

      String scalarChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT-AVG";
      String waveformChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT";
      DAQQuery query = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.parse("2016-10-12T14:00"),
                  TimeUtils.parse("2016-10-12T16:00")),
            new ChannelName(scalarChannel, queryBackend),
            new ChannelName(waveformChannel, queryBackend));

      final List<? extends ChannelEvents> eventsList = RestHelper.queryEvents(context, queryServer, query);
      assertEquals("Size was " + eventsList.size(), 2, eventsList.size());

      List<StreamEvent> events = eventsList.get(0)
            .getEvents()
            .collect(Collectors.toList());
      assertEquals("Size was " + events.size(), 11, events.size());
      StreamEvent event = events.get(0);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(scalarChannel, event.getChannel());
      assertEquals(1, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(1, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {1}, event.getShape());
      assertEquals(1, event.getStatistics().getCount());
      assertEquals(Type.Float64.getKey(), event.getType());
      assertEquals("9.430576E-5", event.getValueAsString());

      events = eventsList.get(1)
            .getEvents()
            .collect(Collectors.toList());
      assertEquals("Size was " + events.size(), 11, events.size());
      event = events.get(0);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(waveformChannel, event.getChannel());
      assertEquals(2048, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(1, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {2048}, event.getShape());
      assertEquals(2048, event.getStatistics().getCount());
      assertEquals(Type.Int32.getKey(), event.getType());
      assertEquals(
            "[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 3, 4, 6, 6, 8, 6, 5, 3, 3, 3, 8, 12, 14... shape [2048]]",
            event.getValueAsString(30));
   }

   @Test
   public void testEventQueryAggregated_01() throws Exception {
      assertNotNull(queryBackend);
      assertNotNull(queryServer);

      String scalarChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT-AVG";
      String waveformChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT";
      DAQQuery query = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.parse("2016-10-12T14:00"),
                  TimeUtils.parse("2016-10-12T16:00")),
            new ChannelName(scalarChannel, queryBackend),
            new ChannelName(waveformChannel, queryBackend));
      query.setAggregation(new AggregationDescriptor(AggregationType.value)
            .setAggregations(Arrays.asList(Aggregation.typed)));

      final List<? extends ChannelEvents> eventsList = RestHelper.queryEvents(context, queryServer, query);
      assertEquals("Size was " + eventsList.size(), 2, eventsList.size());

      List<StreamEvent> events = eventsList.get(0)
            .getEvents()
            .collect(Collectors.toList());
      assertEquals("Size was " + events.size(), 11, events.size());
      StreamEvent event = events.get(0);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(scalarChannel, event.getChannel());
      // assertEquals(1, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(1, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {1}, event.getShape());
      assertEquals(1, event.getStatistics().getCount());
      assertNull(event.getType());
      assertTrue(event.getValueAsString(30).startsWith("StorelessStatistics"));

      events = eventsList.get(1)
            .getEvents()
            .collect(Collectors.toList());
      assertEquals("Size was " + events.size(), 11, events.size());
      event = events.get(0);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(waveformChannel, event.getChannel());
      // assertEquals(1, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(1, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {1}, event.getShape());
      assertEquals(2048, event.getStatistics().getCount());
      assertNull(event.getType());
      assertTrue(event.getValueAsString(30).startsWith("StorelessStatistics"));
   }

   @Test
   public void testEventQueryAggregatedBin_01() throws Exception {
      assertNotNull(queryBackend);
      assertNotNull(queryServer);

      String scalarChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT-AVG";
      String waveformChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT";
      DAQQuery query = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.parse("2016-10-12T14:00"),
                  TimeUtils.parse("2016-10-12T16:00")),
            new ChannelName(scalarChannel, queryBackend),
            new ChannelName(waveformChannel, queryBackend));
      query.setAggregation(new AggregationDescriptor(AggregationType.value)
            .setNrOfBins(2)
            .setAggregations(Arrays.asList(Aggregation.typed)));

      int firstBinEventCount = 6; // 5;
      final List<? extends ChannelEvents> eventsList = RestHelper.queryEvents(context, queryServer, query);
      assertEquals("Size was " + eventsList.size(), 2, eventsList.size());

      List<StreamEvent> events = eventsList.get(0)
            .getEvents()
            .collect(Collectors.toList());
      assertEquals("Size was " + events.size(), 2, events.size());
      StreamEvent event = events.get(0);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(scalarChannel, event.getChannel());
      // assertEquals(1, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(firstBinEventCount, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {1}, event.getShape());
      assertEquals(firstBinEventCount, event.getStatistics().getCount());
      assertNull(event.getType());
      assertTrue(event.getValueAsString(30).startsWith("StorelessStatistics"));

      events = eventsList.get(1)
            .getEvents()
            .collect(Collectors.toList());
      assertEquals("Size was " + events.size(), 11, events.size());
      event = events.get(0);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(waveformChannel, event.getChannel());
      // assertEquals(1, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(firstBinEventCount, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {1}, event.getShape());
      assertEquals(firstBinEventCount * 2048, event.getStatistics().getCount());
      assertNull(event.getType());
      assertTrue(event.getValueAsString(30).startsWith("StorelessStatistics"));
   }

   @Test
   public void testEventQueryIndexAggregateBin_01() throws Exception {
      assertNotNull(queryBackend);
      assertNotNull(queryServer);

      String scalarChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT-AVG";
      String waveformChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT";
      DAQQuery query = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.parse("2016-10-12T14:00"),
                  TimeUtils.parse("2016-10-12T16:00")),
            new ChannelName(scalarChannel, queryBackend),
            new ChannelName(waveformChannel, queryBackend));
      query.setAggregation(new AggregationDescriptor(AggregationType.index)
            .setNrOfBins(2)
            .setAggregations(Arrays.asList(Aggregation.typed)));

      final List<? extends ChannelEvents> eventsList = RestHelper.queryEvents(context, queryServer, query);
      assertEquals("Size was " + eventsList.size(), 2, eventsList.size());

      int firstBinEventCount = 6; // 5;
      List<StreamEvent> events = eventsList.get(0)
            .getEvents()
            .collect(Collectors.toList());
      assertEquals("Size was " + events.size(), 11, events.size());
      StreamEvent event = events.get(0);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(scalarChannel, event.getChannel());
      // assertEquals(1, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(firstBinEventCount, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {1}, event.getShape());
      assertEquals(firstBinEventCount, event.getStatistics().getCount());
      assertEquals(firstBinEventCount, event.getValue(Statistics.class).getCount());
      assertNull(event.getType());
      assertTrue(event.getValueAsString(30).startsWith("StorelessStatistics"));

      events = eventsList.get(1)
            .getEvents()
            .collect(Collectors.toList());
      assertEquals("Size was " + events.size(), 11, events.size());
      event = events.get(0);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(waveformChannel, event.getChannel());
      // assertEquals(1, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(firstBinEventCount, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {2048}, event.getShape());
      assertEquals(2048, event.getValue(List.class).size());
      assertEquals(firstBinEventCount, ((Statistics) event.getValue(List.class).get(0)).getCount());
      assertNull(event.getType());
      //
      // System.out.println(event.getEventCount());
      // System.out.println(event.getGlobalDate());
      // System.out.println(event.getIocDate());
      // System.out.println(event.getPulseId());
      // System.out.println(Arrays.toString(event.getShape()));
      // System.out.println(event.getValue(Object.class));
      // System.out.println(event.getType());
      // System.out.println(event.getValueAsString(30));
   }

   @Test
   public void testEventQueryTableRaw_01() throws Exception {
      assertNotNull(queryBackend);
      assertNotNull(queryServer);

      String scalarChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT-AVG";
      String waveformChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT";
      DAQQuery query = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.parse("2016-10-12T14:00"),
                  TimeUtils.parse("2016-10-12T16:00")),
            new ChannelName(scalarChannel, queryBackend),
            new ChannelName(waveformChannel, queryBackend));
      query.setMapping(new Mapping());

      final ChannelEventTable table = RestHelper.queryEventsTable(context, queryServer, query);
      assertEquals("Size was " + table.size(), 11, table.size());
      final List<Map<String, StreamEvent>> eventsTable = table.getEvents(Collectors.toMap(
            event -> event.getChannel(),
            Function.identity(),
            LoggingMapMergeFunction.getInstance(),
            LinkedHashMap::new))
            .collect(Collectors.toList());
      assertEquals("Size was " + eventsTable.size(), 11, eventsTable.size());
      final Map<String, StreamEvent> eventsRow = eventsTable.get(0);
      assertEquals("Size was " + eventsRow.size(), 2, eventsRow.size());
      assertTrue(eventsRow.containsKey(scalarChannel));
      assertTrue(eventsRow.containsKey(waveformChannel));

      StreamEvent event = eventsRow.get(scalarChannel);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(scalarChannel, event.getChannel());
      assertEquals(1, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(1, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {1}, event.getShape());
      assertEquals(1, event.getStatistics().getCount());
      assertEquals(Type.Float64.getKey(), event.getType());
      assertEquals("9.430576E-5", event.getValueAsString());

      event = eventsRow.get(waveformChannel);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(waveformChannel, event.getChannel());
      assertEquals(2048, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(1, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {2048}, event.getShape());
      assertEquals(2048, event.getStatistics().getCount());
      assertEquals(Type.Int32.getKey(), event.getType());
      assertEquals(
            "[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 3, 4, 6, 6, 8, 6, 5, 3, 3, 3, 8, 12, 14... shape [2048]]",
            event.getValueAsString(30));
   }

   @Test
   public void testEventQueryTableAggregated_01() throws Exception {
      assertNotNull(queryBackend);
      assertNotNull(queryServer);

      String scalarChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT-AVG";
      String waveformChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT";
      DAQQuery query = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.parse("2016-10-12T14:00"),
                  TimeUtils.parse("2016-10-12T16:00")),
            new ChannelName(scalarChannel, queryBackend),
            new ChannelName(waveformChannel, queryBackend));
      query.setAggregation(new AggregationDescriptor(AggregationType.value)
            .setAggregations(Arrays.asList(Aggregation.typed)));
      query.setMapping(new Mapping());

      final ChannelEventTable table = RestHelper.queryEventsTable(context, queryServer, query);
      assertEquals("Size was " + table.size(), 11, table.size());
      final List<Map<String, StreamEvent>> eventsTable = table.getEvents(Collectors.toMap(
            event -> event.getChannel(),
            Function.identity(),
            LoggingMapMergeFunction.getInstance(),
            LinkedHashMap::new))
            .collect(Collectors.toList());
      assertEquals("Size was " + eventsTable.size(), 11, eventsTable.size());
      final Map<String, StreamEvent> eventsRow = eventsTable.get(0);
      assertEquals("Size was " + eventsRow.size(), 2, eventsRow.size());
      assertTrue(eventsRow.containsKey(scalarChannel));
      assertTrue(eventsRow.containsKey(waveformChannel));

      StreamEvent event = eventsRow.get(scalarChannel);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(scalarChannel, event.getChannel());
      // assertEquals(1, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(1, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {1}, event.getShape());
      assertEquals(1, event.getStatistics().getCount());
      assertNull(event.getType());
      assertTrue(event.getValueAsString(30).startsWith("StorelessStatistics"));

      event = eventsRow.get(waveformChannel);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(waveformChannel, event.getChannel());
      // assertEquals(1, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(1, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {1}, event.getShape());
      assertEquals(2048, event.getStatistics().getCount());
      assertNull(event.getType());
      assertTrue(event.getValueAsString(30).startsWith("StorelessStatistics"));
   }

   @Test
   public void testEventQueryTableAggregatedBin_01() throws Exception {
      assertNotNull(queryBackend);
      assertNotNull(queryServer);

      String scalarChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT-AVG";
      String waveformChannel = "SINEG01-RCIR-PUP10:SIG-AMPLT";
      DAQQuery query = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.parse("2016-10-12T14:00"),
                  TimeUtils.parse("2016-10-12T16:00")),
            new ChannelName(scalarChannel, queryBackend),
            new ChannelName(waveformChannel, queryBackend));
      query.setAggregation(new AggregationDescriptor(AggregationType.value)
            .setNrOfBins(2)
            .setAggregations(Arrays.asList(Aggregation.typed)));
      query.setMapping(new Mapping());

      int firstBinEventCount = 6; // 5;
      final ChannelEventTable table = RestHelper.queryEventsTable(context, queryServer, query);
      assertEquals("Size was " + table.size(), 2, table.size());
      final List<Map<String, StreamEvent>> eventsTable = table.getEvents(Collectors.toMap(
            event -> event.getChannel(),
            Function.identity(),
            LoggingMapMergeFunction.getInstance(),
            LinkedHashMap::new))
            .collect(Collectors.toList());
      assertEquals("Size was " + eventsTable.size(), 2, eventsTable.size());
      final Map<String, StreamEvent> eventsRow = eventsTable.get(0);
      assertEquals("Size was " + eventsRow.size(), 2, eventsRow.size());
      assertTrue(eventsRow.containsKey(scalarChannel));
      assertTrue(eventsRow.containsKey(waveformChannel));

      StreamEvent event = eventsRow.get(scalarChannel);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(scalarChannel, event.getChannel());
      // assertEquals(1, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(firstBinEventCount, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {1}, event.getShape());
      assertEquals(firstBinEventCount, event.getStatistics().getCount());
      assertNull(event.getType());
      assertTrue(event.getValueAsString(30).startsWith("StorelessStatistics"));

      event = eventsRow.get(waveformChannel);
      assertEquals(Event.class, event.getClass());
      assertEquals(queryBackend, event.getBackend());
      assertEquals(waveformChannel, event.getChannel());
      // assertEquals(1, event.getDoubleStream(ByteConverter.PARALLELISM_FALSE).count());
      assertEquals(firstBinEventCount, event.getEventCount());
      assertEquals("2016-10-12T14:07:09.914760000+02:00", event.getGlobalDate());
      assertEquals("1970-01-01T01:00:00.000000000+01:00", event.getIocDate());
      assertEquals(638760000, event.getPulseId());
      assertArrayEquals(new int[] {1}, event.getShape());
      assertEquals(firstBinEventCount * 2048, event.getStatistics().getCount());
      assertNull(event.getType());
      assertTrue(event.getValueAsString(30).startsWith("StorelessStatistics"));
   }

   @Test
   public void testQueryManager_01() throws Exception {
      final List<Backend> backends = queryManager.getBackends().collect(Collectors.toList());
      assertEquals(7, backends.size());

      final List<String> channels = queryManager.getChannels(new ChannelsRequest())
            .flatMap(response -> response.getChannels().stream())
            .collect(Collectors.toList());
      assertTrue("Size was " + channels.size(), channels.size() > 400000);
   }
}
